/*
 * LibV abstraction layer.
 * Version 4-alpha.
 *
 * Copyright (C) 1998 - 2011 Vladimir A. Smirnov (kkatarn@sleepgate.ru)
 *
 * <ru_koi8-r>    ,   </ru_koi8-r>
 * <ru_win1251>    ,   </ru_win1251>
 *
 * Last update: 15 nov 2006
 *
 * Important note.
 * The only question with OpenGL is an extension management.
 * Some useful stuff former available as ARB extensions now promoted to
 * core functionality; it is clear that using core functions is the
 * _right_ way to do things.
 *
 * But some stuff still aren't in the core (under Win32), and at the
 * moment of writing not all vendors (actually, almost _none_ of them)
 * provides _solid_ GL 2.0 - capable drivers (NVDA is indeed _really_
 * close to the point, though; don't know much about others).
 *
 * So, we have to implement _portable_ ways to:
 *  - determine (_after_ first call to vglInit()) the GL version
 *    and availability of _all_ functions required by that version;
 *  - determine if the specific extension is available;
 *  - instantly call any available core or extension function without
 *    that Win32-specific madness with getting pointers;
 *  - be able to fall back to different path if the specific extension
 *    isn't supported (i.e. we don't want to get 'unresolved external'
 *    errors when loading the executable).
 *
 * Moreover, under Win32 we sometimes have to call requred functions from
 * the program linked with LibV dynamically (even if the program
 * written with other language - ObjectPascal, for example).
 *
 * The best way to call GL function is via pointer stored in DLL; for C
 * program it can be done as follows:
 *
 * - in LibV's vglext.c:
 *	__declspec(dllexport) (__stdcall *proc_ptr)(...);
 *	...
 *	proc_ptr = load_gl_proc(...);
 *
 * - in program dynamically linked with LibV.dll:
 *	extern __declspec(dllimport) (__stdcall *proc_ptr)(...);
 *	... // init_extensions
 *	proc_ptr(...);
 *
 * But at least one of us - either me or dcc32.exe - is stupid. The
 * above method just don't works with Delphi. Even such a hack:
 *
 *	procedure dead_ptr(...);stdcall;external 'LibV.dll' name 'proc_ptr';
 *	type proc_t = procedure(...);stdcall;
 *	var
 *		proc_ptr : proc_t;
 *	begin
 *		asm
 *			mov eax, offset dead_ptr
 *			mov proc_ptr, [eax]
 *		end;
 *		proc_ptr(...);
 *	end.
 *
 * ... is useless. It looks (after endless debugging sessions with wdw.exe
 * and td32.exe) like Delphi uses some kind of jump table; trying to pass
 * through it _is_ insane.
 *
 * So, the (ugly...) way as it seems to me:
 *
 *	#ifdef __WIN32__
 *		#ifdef	BUILD_LIBV_DLL
 *			// we are compiling the DLL
 *			#define VGLEXTAPI __declspec(dllexport)
 *			#ifdef	DLL_TARGET_DCC
 *				// for Delphi
 *				in GL.h declare VGLEXTAPI extern gate functions;
 *				in vglext.c declare local pointers to real functions,
 *				initialize pointers and call them from gate functions.
 *			#else // !DLL_TARGET_DCC
 *				in GL.h declare VGLEXTAPI extern _pointers_;
 *				in vglext.c use wglGetProcAddress() and initialize pointers.
 *			#endif // !DLL_TARGET_DCC
 *		#else // !BUILD_LIBV_DLL
 *			// Client C program or static library
 *			#ifdef DYNAMIC_LIBV_LINK
 *				// Client have to be linked with libv.dll later
 *				#define VGLEXTAPI GLAPI - __declspec(dllimport)
 *			#else // !DYNAMIC_LIBV_LINK
 *				// Statically linked client or static library
 *				#define VGLEXTAPI
 *			#endif // !DYNAMIC_LIBV_LINK
 *			in GL.h declare extern _pointers_;
 *			in vglext.c use wglGetProcAddress() and initialize pointers.
 *		#endif // !BUILD_LIBV_DLL
 *	#else // !__WIN32__ -- thanks God !
 *		in GL.h declare extern _pointers_;
 *		in vglext.c use glXGetProcAddress() (_right_ way) or
 *		dlopen("libGL.so") (hack) and initialize pointers.
 *	#else // !__WIN32__
 *
 * Another Win32 note: according to Silicon Graphics, "there is an
 * unavoidable inconsistency with wglGetProcAddress, which returns
 * context-dependent pointers."
 *
 * Actually, it is not that bad: for the same GL implementation
 * wglGetProcAddress() always returns same pointers. So, it has to be said
 * the pointers are implementation-specific, not context-specific (at
 * least, this is true for my machine with NV 7X.xxx GL driver). Because
 * of this, it is a bad idea to protect pointers in the VGL context (call
 * via protected pointer requires the hold of global lock). We just either
 * do _not_ allow to switch to another implementation (for example, to
 * Microsoft software GL after HW GL) after first call to vglCreateWindow,
 * or do _not_ allow any extended stuff after such a switch.
 *
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 *
 * Availability of extension functions must be detected at run-time
 * _after_ successful call to vglCreateWindow() in such a way:
 *
 *	if(!vglHave(glMultiTexCoord2fARB) || !vglHave(glActiveTextureARB)))
 *		die("brain-dead renderer does not support GL_ARB_multitexture.");
 *	else
 *		trace("GL_ARB_multitexture usable.");
 *	where
 *
 *	#if	defined(BUILD_LIBV_DLL) && defined(DLL_TARGET_DCC)
 *		vboolean vglHave(proc) { return GLPTR(proc) ? true : false; }
 *	#else
 *		#define	vglHave(proc) (GLPTR(proc) ? true : false)
 *	#endif
 */

/*
 * Mesa 3-D graphics library
 * Version:  6.2
 *
 * Copyright (C) 1999-2004  Brian Paul   All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef	__Vgl_h_
#define	__Vgl_h_

#ifdef	__gl_h_
#error	"System-specific OpenGL stuff can't be used in LibV program."
#error	"All you need are probably already in this LibV header;"
#error	"please, locate and remove all \"#include \"GL/gl.h\" or"
#error	"replace them with \"#include \"v_lib.h\"."
#else	/* !__gl_h_ */
/*
 * Prevent later inclusion of system-specific stuff
 */
#define	__gl_h_
#endif	/* !__gl_h_ */

#ifdef	__WIN32__

#ifdef	_MSC_VER
#define GLAPI extern __declspec(dllimport)
#else	/* !_MSC_VER */
#define GLAPI extern
#endif	/* !_MSC_VER */

#ifdef	DLL_TARGET_DCC

#ifdef	_MSC_VER
#define VGLEXTAPI extern __declspec(dllexport)
#else	/* !_MSC_VER */
#define VGLEXTAPI extern
#endif	/* !_MSC_VER */

#else	/* !DLL_TARGET_DCC */

#define VGLEXTAPI extern

#endif	/* !DLL_TARGET_DCC */

#define GLAPIENTRY __stdcall
#define	VGLEXTAPIENTRY	__stdcall

#else	/* !__WIN32__ */

#define GLAPI extern
#define VGLEXTAPI extern
#define GLAPIENTRY
#define	VGLEXTAPIENTRY

#endif /* !__WIN32__ */

#if	defined(DLL_TARGET_DCC) && !defined(BUILD_LIBV_DLL)
#error "DLL_TARGET_DCC can not be defined without BUILD_LIBV_DLL"
#endif

#include <stddef.h> /* for ptrdiff_t */

typedef unsigned int GLenum;
typedef unsigned char GLboolean;
typedef unsigned int GLbitfield;
typedef signed char GLbyte;
typedef short GLshort;
typedef int GLint;
typedef int GLsizei;
typedef unsigned char GLubyte;
typedef unsigned short GLushort;
typedef unsigned int GLuint;
typedef float GLfloat;
typedef float GLclampf;
typedef double GLdouble;
typedef double GLclampd;
typedef void GLvoid;

#define GL_VERSION_1_1                      1

/* AttribMask */
#define GL_CURRENT_BIT                      0x00000001
#define GL_POINT_BIT                        0x00000002
#define GL_LINE_BIT                         0x00000004
#define GL_POLYGON_BIT                      0x00000008
#define GL_POLYGON_STIPPLE_BIT              0x00000010
#define GL_PIXEL_MODE_BIT                   0x00000020
#define GL_LIGHTING_BIT                     0x00000040
#define GL_FOG_BIT                          0x00000080
#define GL_DEPTH_BUFFER_BIT                 0x00000100
#define GL_ACCUM_BUFFER_BIT                 0x00000200
#define GL_STENCIL_BUFFER_BIT               0x00000400
#define GL_VIEWPORT_BIT                     0x00000800
#define GL_TRANSFORM_BIT                    0x00001000
#define GL_ENABLE_BIT                       0x00002000
#define GL_COLOR_BUFFER_BIT                 0x00004000
#define GL_HINT_BIT                         0x00008000
#define GL_EVAL_BIT                         0x00010000
#define GL_LIST_BIT                         0x00020000
#define GL_TEXTURE_BIT                      0x00040000
#define GL_SCISSOR_BIT                      0x00080000
#define GL_ALL_ATTRIB_BITS                  0x000fffff

/* ClientAttribMask */
#define GL_CLIENT_PIXEL_STORE_BIT           0x00000001
#define GL_CLIENT_VERTEX_ARRAY_BIT          0x00000002
#define GL_CLIENT_ALL_ATTRIB_BITS           0xFFFFFFFF

/* Boolean */
#define GL_FALSE                            0
#define GL_TRUE                             1

/* BeginMode */
#define GL_POINTS                           0x0000
#define GL_LINES                            0x0001
#define GL_LINE_LOOP                        0x0002
#define GL_LINE_STRIP                       0x0003
#define GL_TRIANGLES                        0x0004
#define GL_TRIANGLE_STRIP                   0x0005
#define GL_TRIANGLE_FAN                     0x0006
#define GL_QUADS                            0x0007
#define GL_QUAD_STRIP                       0x0008
#define GL_POLYGON                          0x0009

/* AccumOp */
#define GL_ACCUM                            0x0100
#define GL_LOAD                             0x0101
#define GL_RETURN                           0x0102
#define GL_MULT                             0x0103
#define GL_ADD                              0x0104

/* AlphaFunction */
#define GL_NEVER                            0x0200
#define GL_LESS                             0x0201
#define GL_EQUAL                            0x0202
#define GL_LEQUAL                           0x0203
#define GL_GREATER                          0x0204
#define GL_NOTEQUAL                         0x0205
#define GL_GEQUAL                           0x0206
#define GL_ALWAYS                           0x0207

/* BlendingFactor */
#define GL_ZERO                             0
#define GL_ONE                              1
#define GL_SRC_COLOR                        0x0300
#define GL_ONE_MINUS_SRC_COLOR              0x0301
#define GL_SRC_ALPHA                        0x0302
#define GL_ONE_MINUS_SRC_ALPHA              0x0303
#define GL_DST_ALPHA                        0x0304
#define GL_ONE_MINUS_DST_ALPHA              0x0305
#define GL_DST_COLOR                        0x0306
#define GL_ONE_MINUS_DST_COLOR              0x0307
#define GL_SRC_ALPHA_SATURATE               0x0308

/* DrawBufferMode */
#define GL_NONE                             0
#define GL_FRONT_LEFT                       0x0400
#define GL_FRONT_RIGHT                      0x0401
#define GL_BACK_LEFT                        0x0402
#define GL_BACK_RIGHT                       0x0403
#define GL_FRONT                            0x0404
#define GL_BACK                             0x0405
#define GL_LEFT                             0x0406
#define GL_RIGHT                            0x0407
#define GL_FRONT_AND_BACK                   0x0408
#define GL_AUX0                             0x0409
#define GL_AUX1                             0x040A
#define GL_AUX2                             0x040B
#define GL_AUX3                             0x040C

/* ErrorCode */
#define GL_NO_ERROR                         0
#define GL_INVALID_ENUM                     0x0500
#define GL_INVALID_VALUE                    0x0501
#define GL_INVALID_OPERATION                0x0502
#define GL_STACK_OVERFLOW                   0x0503
#define GL_STACK_UNDERFLOW                  0x0504
#define GL_OUT_OF_MEMORY                    0x0505

/* FeedbackType */
#define GL_2D                               0x0600
#define GL_3D                               0x0601
#define GL_3D_COLOR                         0x0602
#define GL_3D_COLOR_TEXTURE                 0x0603
#define GL_4D_COLOR_TEXTURE                 0x0604

/* FeedBackToken */
#define GL_PASS_THROUGH_TOKEN               0x0700
#define GL_POINT_TOKEN                      0x0701
#define GL_LINE_TOKEN                       0x0702
#define GL_POLYGON_TOKEN                    0x0703
#define GL_BITMAP_TOKEN                     0x0704
#define GL_DRAW_PIXEL_TOKEN                 0x0705
#define GL_COPY_PIXEL_TOKEN                 0x0706
#define GL_LINE_RESET_TOKEN                 0x0707

/* FogMode */
#define GL_EXP                              0x0800
#define GL_EXP2                             0x0801

/* FrontFaceDirection */
#define GL_CW                               0x0900
#define GL_CCW                              0x0901

/* GetMapQuery */
#define GL_COEFF                            0x0A00
#define GL_ORDER                            0x0A01
#define GL_DOMAIN                           0x0A02

/* GetPixelMap */
#define GL_PIXEL_MAP_I_TO_I                 0x0C70
#define GL_PIXEL_MAP_S_TO_S                 0x0C71
#define GL_PIXEL_MAP_I_TO_R                 0x0C72
#define GL_PIXEL_MAP_I_TO_G                 0x0C73
#define GL_PIXEL_MAP_I_TO_B                 0x0C74
#define GL_PIXEL_MAP_I_TO_A                 0x0C75
#define GL_PIXEL_MAP_R_TO_R                 0x0C76
#define GL_PIXEL_MAP_G_TO_G                 0x0C77
#define GL_PIXEL_MAP_B_TO_B                 0x0C78
#define GL_PIXEL_MAP_A_TO_A                 0x0C79

/* GetPointervPName */
#define GL_VERTEX_ARRAY_POINTER             0x808E
#define GL_NORMAL_ARRAY_POINTER             0x808F
#define GL_COLOR_ARRAY_POINTER              0x8090
#define GL_INDEX_ARRAY_POINTER              0x8091
#define GL_TEXTURE_COORD_ARRAY_POINTER      0x8092
#define GL_EDGE_FLAG_ARRAY_POINTER          0x8093

/* GetPName */
#define GL_CURRENT_COLOR                    0x0B00
#define GL_CURRENT_INDEX                    0x0B01
#define GL_CURRENT_NORMAL                   0x0B02
#define GL_CURRENT_TEXTURE_COORDS           0x0B03
#define GL_CURRENT_RASTER_COLOR             0x0B04
#define GL_CURRENT_RASTER_INDEX             0x0B05
#define GL_CURRENT_RASTER_TEXTURE_COORDS    0x0B06
#define GL_CURRENT_RASTER_POSITION          0x0B07
#define GL_CURRENT_RASTER_POSITION_VALID    0x0B08
#define GL_CURRENT_RASTER_DISTANCE          0x0B09
#define GL_POINT_SMOOTH                     0x0B10
#define GL_POINT_SIZE                       0x0B11
#define GL_POINT_SIZE_RANGE                 0x0B12
#define GL_POINT_SIZE_GRANULARITY           0x0B13
#define GL_LINE_SMOOTH                      0x0B20
#define GL_LINE_WIDTH                       0x0B21
#define GL_LINE_WIDTH_RANGE                 0x0B22
#define GL_LINE_WIDTH_GRANULARITY           0x0B23
#define GL_LINE_STIPPLE                     0x0B24
#define GL_LINE_STIPPLE_PATTERN             0x0B25
#define GL_LINE_STIPPLE_REPEAT              0x0B26
#define GL_LIST_MODE                        0x0B30
#define GL_MAX_LIST_NESTING                 0x0B31
#define GL_LIST_BASE                        0x0B32
#define GL_LIST_INDEX                       0x0B33
#define GL_POLYGON_MODE                     0x0B40
#define GL_POLYGON_SMOOTH                   0x0B41
#define GL_POLYGON_STIPPLE                  0x0B42
#define GL_EDGE_FLAG                        0x0B43
#define GL_CULL_FACE                        0x0B44
#define GL_CULL_FACE_MODE                   0x0B45
#define GL_FRONT_FACE                       0x0B46
#define GL_LIGHTING                         0x0B50
#define GL_LIGHT_MODEL_LOCAL_VIEWER         0x0B51
#define GL_LIGHT_MODEL_TWO_SIDE             0x0B52
#define GL_LIGHT_MODEL_AMBIENT              0x0B53
#define GL_SHADE_MODEL                      0x0B54
#define GL_COLOR_MATERIAL_FACE              0x0B55
#define GL_COLOR_MATERIAL_PARAMETER         0x0B56
#define GL_COLOR_MATERIAL                   0x0B57
#define GL_FOG                              0x0B60
#define GL_FOG_INDEX                        0x0B61
#define GL_FOG_DENSITY                      0x0B62
#define GL_FOG_START                        0x0B63
#define GL_FOG_END                          0x0B64
#define GL_FOG_MODE                         0x0B65
#define GL_FOG_COLOR                        0x0B66
#define GL_DEPTH_RANGE                      0x0B70
#define GL_DEPTH_TEST                       0x0B71
#define GL_DEPTH_WRITEMASK                  0x0B72
#define GL_DEPTH_CLEAR_VALUE                0x0B73
#define GL_DEPTH_FUNC                       0x0B74
#define GL_ACCUM_CLEAR_VALUE                0x0B80
#define GL_STENCIL_TEST                     0x0B90
#define GL_STENCIL_CLEAR_VALUE              0x0B91
#define GL_STENCIL_FUNC                     0x0B92
#define GL_STENCIL_VALUE_MASK               0x0B93
#define GL_STENCIL_FAIL                     0x0B94
#define GL_STENCIL_PASS_DEPTH_FAIL          0x0B95
#define GL_STENCIL_PASS_DEPTH_PASS          0x0B96
#define GL_STENCIL_REF                      0x0B97
#define GL_STENCIL_WRITEMASK                0x0B98
#define GL_MATRIX_MODE                      0x0BA0
#define GL_NORMALIZE                        0x0BA1
#define GL_VIEWPORT                         0x0BA2
#define GL_MODELVIEW_STACK_DEPTH            0x0BA3
#define GL_PROJECTION_STACK_DEPTH           0x0BA4
#define GL_TEXTURE_STACK_DEPTH              0x0BA5
#define GL_MODELVIEW_MATRIX                 0x0BA6
#define GL_PROJECTION_MATRIX                0x0BA7
#define GL_TEXTURE_MATRIX                   0x0BA8
#define GL_ATTRIB_STACK_DEPTH               0x0BB0
#define GL_CLIENT_ATTRIB_STACK_DEPTH        0x0BB1
#define GL_ALPHA_TEST                       0x0BC0
#define GL_ALPHA_TEST_FUNC                  0x0BC1
#define GL_ALPHA_TEST_REF                   0x0BC2
#define GL_DITHER                           0x0BD0
#define GL_BLEND_DST                        0x0BE0
#define GL_BLEND_SRC                        0x0BE1
#define GL_BLEND                            0x0BE2
#define GL_LOGIC_OP_MODE                    0x0BF0
#define GL_INDEX_LOGIC_OP                   0x0BF1
#define GL_LOGIC_OP                         GL_INDEX_LOGIC_OP
#define GL_COLOR_LOGIC_OP                   0x0BF2
#define GL_AUX_BUFFERS                      0x0C00
#define GL_DRAW_BUFFER                      0x0C01
#define GL_READ_BUFFER                      0x0C02
#define GL_SCISSOR_BOX                      0x0C10
#define GL_SCISSOR_TEST                     0x0C11
#define GL_INDEX_CLEAR_VALUE                0x0C20
#define GL_INDEX_WRITEMASK                  0x0C21
#define GL_COLOR_CLEAR_VALUE                0x0C22
#define GL_COLOR_WRITEMASK                  0x0C23
#define GL_INDEX_MODE                       0x0C30
#define GL_RGBA_MODE                        0x0C31
#define GL_DOUBLEBUFFER                     0x0C32
#define GL_STEREO                           0x0C33
#define GL_RENDER_MODE                      0x0C40
#define GL_PERSPECTIVE_CORRECTION_HINT      0x0C50
#define GL_POINT_SMOOTH_HINT                0x0C51
#define GL_LINE_SMOOTH_HINT                 0x0C52
#define GL_POLYGON_SMOOTH_HINT              0x0C53
#define GL_FOG_HINT                         0x0C54
#define GL_TEXTURE_GEN_S                    0x0C60
#define GL_TEXTURE_GEN_T                    0x0C61
#define GL_TEXTURE_GEN_R                    0x0C62
#define GL_TEXTURE_GEN_Q                    0x0C63
#define GL_PIXEL_MAP_I_TO_I_SIZE            0x0CB0
#define GL_PIXEL_MAP_S_TO_S_SIZE            0x0CB1
#define GL_PIXEL_MAP_I_TO_R_SIZE            0x0CB2
#define GL_PIXEL_MAP_I_TO_G_SIZE            0x0CB3
#define GL_PIXEL_MAP_I_TO_B_SIZE            0x0CB4
#define GL_PIXEL_MAP_I_TO_A_SIZE            0x0CB5
#define GL_PIXEL_MAP_R_TO_R_SIZE            0x0CB6
#define GL_PIXEL_MAP_G_TO_G_SIZE            0x0CB7
#define GL_PIXEL_MAP_B_TO_B_SIZE            0x0CB8
#define GL_PIXEL_MAP_A_TO_A_SIZE            0x0CB9
#define GL_UNPACK_SWAP_BYTES                0x0CF0
#define GL_UNPACK_LSB_FIRST                 0x0CF1
#define GL_UNPACK_ROW_LENGTH                0x0CF2
#define GL_UNPACK_SKIP_ROWS                 0x0CF3
#define GL_UNPACK_SKIP_PIXELS               0x0CF4
#define GL_UNPACK_ALIGNMENT                 0x0CF5
#define GL_PACK_SWAP_BYTES                  0x0D00
#define GL_PACK_LSB_FIRST                   0x0D01
#define GL_PACK_ROW_LENGTH                  0x0D02
#define GL_PACK_SKIP_ROWS                   0x0D03
#define GL_PACK_SKIP_PIXELS                 0x0D04
#define GL_PACK_ALIGNMENT                   0x0D05
#define GL_MAP_COLOR                        0x0D10
#define GL_MAP_STENCIL                      0x0D11
#define GL_INDEX_SHIFT                      0x0D12
#define GL_INDEX_OFFSET                     0x0D13
#define GL_RED_SCALE                        0x0D14
#define GL_RED_BIAS                         0x0D15
#define GL_ZOOM_X                           0x0D16
#define GL_ZOOM_Y                           0x0D17
#define GL_GREEN_SCALE                      0x0D18
#define GL_GREEN_BIAS                       0x0D19
#define GL_BLUE_SCALE                       0x0D1A
#define GL_BLUE_BIAS                        0x0D1B
#define GL_ALPHA_SCALE                      0x0D1C
#define GL_ALPHA_BIAS                       0x0D1D
#define GL_DEPTH_SCALE                      0x0D1E
#define GL_DEPTH_BIAS                       0x0D1F
#define GL_MAX_EVAL_ORDER                   0x0D30
#define GL_MAX_LIGHTS                       0x0D31
#define GL_MAX_CLIP_PLANES                  0x0D32
#define GL_MAX_TEXTURE_SIZE                 0x0D33
#define GL_MAX_PIXEL_MAP_TABLE              0x0D34
#define GL_MAX_ATTRIB_STACK_DEPTH           0x0D35
#define GL_MAX_MODELVIEW_STACK_DEPTH        0x0D36
#define GL_MAX_NAME_STACK_DEPTH             0x0D37
#define GL_MAX_PROJECTION_STACK_DEPTH       0x0D38
#define GL_MAX_TEXTURE_STACK_DEPTH          0x0D39
#define GL_MAX_VIEWPORT_DIMS                0x0D3A
#define GL_MAX_CLIENT_ATTRIB_STACK_DEPTH    0x0D3B
#define GL_SUBPIXEL_BITS                    0x0D50
#define GL_INDEX_BITS                       0x0D51
#define GL_RED_BITS                         0x0D52
#define GL_GREEN_BITS                       0x0D53
#define GL_BLUE_BITS                        0x0D54
#define GL_ALPHA_BITS                       0x0D55
#define GL_DEPTH_BITS                       0x0D56
#define GL_STENCIL_BITS                     0x0D57
#define GL_ACCUM_RED_BITS                   0x0D58
#define GL_ACCUM_GREEN_BITS                 0x0D59
#define GL_ACCUM_BLUE_BITS                  0x0D5A
#define GL_ACCUM_ALPHA_BITS                 0x0D5B
#define GL_NAME_STACK_DEPTH                 0x0D70
#define GL_AUTO_NORMAL                      0x0D80
#define GL_MAP1_COLOR_4                     0x0D90
#define GL_MAP1_INDEX                       0x0D91
#define GL_MAP1_NORMAL                      0x0D92
#define GL_MAP1_TEXTURE_COORD_1             0x0D93
#define GL_MAP1_TEXTURE_COORD_2             0x0D94
#define GL_MAP1_TEXTURE_COORD_3             0x0D95
#define GL_MAP1_TEXTURE_COORD_4             0x0D96
#define GL_MAP1_VERTEX_3                    0x0D97
#define GL_MAP1_VERTEX_4                    0x0D98
#define GL_MAP2_COLOR_4                     0x0DB0
#define GL_MAP2_INDEX                       0x0DB1
#define GL_MAP2_NORMAL                      0x0DB2
#define GL_MAP2_TEXTURE_COORD_1             0x0DB3
#define GL_MAP2_TEXTURE_COORD_2             0x0DB4
#define GL_MAP2_TEXTURE_COORD_3             0x0DB5
#define GL_MAP2_TEXTURE_COORD_4             0x0DB6
#define GL_MAP2_VERTEX_3                    0x0DB7
#define GL_MAP2_VERTEX_4                    0x0DB8
#define GL_MAP1_GRID_DOMAIN                 0x0DD0
#define GL_MAP1_GRID_SEGMENTS               0x0DD1
#define GL_MAP2_GRID_DOMAIN                 0x0DD2
#define GL_MAP2_GRID_SEGMENTS               0x0DD3
#define GL_TEXTURE_1D                       0x0DE0
#define GL_TEXTURE_2D                       0x0DE1
#define GL_FEEDBACK_BUFFER_POINTER          0x0DF0
#define GL_FEEDBACK_BUFFER_SIZE             0x0DF1
#define GL_FEEDBACK_BUFFER_TYPE             0x0DF2
#define GL_SELECTION_BUFFER_POINTER         0x0DF3
#define GL_SELECTION_BUFFER_SIZE            0x0DF4
#define GL_POLYGON_OFFSET_UNITS             0x2A00
#define GL_POLYGON_OFFSET_POINT             0x2A01
#define GL_POLYGON_OFFSET_LINE              0x2A02
#define GL_POLYGON_OFFSET_FILL              0x8037
#define GL_POLYGON_OFFSET_FACTOR            0x8038
#define GL_TEXTURE_BINDING_1D               0x8068
#define GL_TEXTURE_BINDING_2D               0x8069
#define GL_VERTEX_ARRAY                     0x8074
#define GL_NORMAL_ARRAY                     0x8075
#define GL_COLOR_ARRAY                      0x8076
#define GL_INDEX_ARRAY                      0x8077
#define GL_TEXTURE_COORD_ARRAY              0x8078
#define GL_EDGE_FLAG_ARRAY                  0x8079
#define GL_VERTEX_ARRAY_SIZE                0x807A
#define GL_VERTEX_ARRAY_TYPE                0x807B
#define GL_VERTEX_ARRAY_STRIDE              0x807C
#define GL_NORMAL_ARRAY_TYPE                0x807E
#define GL_NORMAL_ARRAY_STRIDE              0x807F
#define GL_COLOR_ARRAY_SIZE                 0x8081
#define GL_COLOR_ARRAY_TYPE                 0x8082
#define GL_COLOR_ARRAY_STRIDE               0x8083
#define GL_INDEX_ARRAY_TYPE                 0x8085
#define GL_INDEX_ARRAY_STRIDE               0x8086
#define GL_TEXTURE_COORD_ARRAY_SIZE         0x8088
#define GL_TEXTURE_COORD_ARRAY_TYPE         0x8089
#define GL_TEXTURE_COORD_ARRAY_STRIDE       0x808A
#define GL_EDGE_FLAG_ARRAY_STRIDE           0x808C

/* GetTextureParameter */
#define GL_TEXTURE_WIDTH                    0x1000
#define GL_TEXTURE_HEIGHT                   0x1001
#define GL_TEXTURE_INTERNAL_FORMAT          0x1003
#define GL_TEXTURE_COMPONENTS               GL_TEXTURE_INTERNAL_FORMAT
#define GL_TEXTURE_BORDER_COLOR             0x1004
#define GL_TEXTURE_BORDER                   0x1005
#define GL_TEXTURE_RED_SIZE                 0x805C
#define GL_TEXTURE_GREEN_SIZE               0x805D
#define GL_TEXTURE_BLUE_SIZE                0x805E
#define GL_TEXTURE_ALPHA_SIZE               0x805F
#define GL_TEXTURE_LUMINANCE_SIZE           0x8060
#define GL_TEXTURE_INTENSITY_SIZE           0x8061
#define GL_TEXTURE_PRIORITY                 0x8066
#define GL_TEXTURE_RESIDENT                 0x8067

/* HintMode */
#define GL_DONT_CARE                        0x1100
#define GL_FASTEST                          0x1101
#define GL_NICEST                           0x1102

/* LightParameter */
#define GL_AMBIENT                          0x1200
#define GL_DIFFUSE                          0x1201
#define GL_SPECULAR                         0x1202
#define GL_POSITION                         0x1203
#define GL_SPOT_DIRECTION                   0x1204
#define GL_SPOT_EXPONENT                    0x1205
#define GL_SPOT_CUTOFF                      0x1206
#define GL_CONSTANT_ATTENUATION             0x1207
#define GL_LINEAR_ATTENUATION               0x1208
#define GL_QUADRATIC_ATTENUATION            0x1209

/* ListMode */
#define GL_COMPILE                          0x1300
#define GL_COMPILE_AND_EXECUTE              0x1301

/* DataType */
#define GL_BYTE                             0x1400
#define GL_UNSIGNED_BYTE                    0x1401
#define GL_SHORT                            0x1402
#define GL_UNSIGNED_SHORT                   0x1403
#define GL_INT                              0x1404
#define GL_UNSIGNED_INT                     0x1405
#define GL_FLOAT                            0x1406
#define GL_2_BYTES                          0x1407
#define GL_3_BYTES                          0x1408
#define GL_4_BYTES                          0x1409
#define GL_DOUBLE                           0x140A
#define GL_DOUBLE_EXT                       0x140A

/* LogicOp */
#define GL_CLEAR                            0x1500
#define GL_AND                              0x1501
#define GL_AND_REVERSE                      0x1502
#define GL_COPY                             0x1503
#define GL_AND_INVERTED                     0x1504
#define GL_NOOP                             0x1505
#define GL_XOR                              0x1506
#define GL_OR                               0x1507
#define GL_NOR                              0x1508
#define GL_EQUIV                            0x1509
#define GL_INVERT                           0x150A
#define GL_OR_REVERSE                       0x150B
#define GL_COPY_INVERTED                    0x150C
#define GL_OR_INVERTED                      0x150D
#define GL_NAND                             0x150E
#define GL_SET                              0x150F

/* MaterialParameter */
#define GL_EMISSION                         0x1600
#define GL_SHININESS                        0x1601
#define GL_AMBIENT_AND_DIFFUSE              0x1602
#define GL_COLOR_INDEXES                    0x1603

/* MatrixMode */
#define GL_MODELVIEW                        0x1700
#define GL_PROJECTION                       0x1701
#define GL_TEXTURE                          0x1702

/* PixelCopyType */
#define GL_COLOR                            0x1800
#define GL_DEPTH                            0x1801
#define GL_STENCIL                          0x1802

/* PixelFormat */
#define GL_COLOR_INDEX                      0x1900
#define GL_STENCIL_INDEX                    0x1901
#define GL_DEPTH_COMPONENT                  0x1902
#define GL_RED                              0x1903
#define GL_GREEN                            0x1904
#define GL_BLUE                             0x1905
#define GL_ALPHA                            0x1906
#define GL_RGB                              0x1907
#define GL_RGBA                             0x1908
#define GL_LUMINANCE                        0x1909
#define GL_LUMINANCE_ALPHA                  0x190A

/* PolygonMode */
#define GL_POINT                            0x1B00
#define GL_LINE                             0x1B01
#define GL_FILL                             0x1B02

/* RenderingMode */
#define GL_RENDER                           0x1C00
#define GL_FEEDBACK                         0x1C01
#define GL_SELECT                           0x1C02

/* ShadingModel */
#define GL_FLAT                             0x1D00
#define GL_SMOOTH                           0x1D01

/* StencilOp */
#define GL_KEEP                             0x1E00
#define GL_REPLACE                          0x1E01
#define GL_INCR                             0x1E02
#define GL_DECR                             0x1E03

/* StringName */
#define GL_VENDOR                           0x1F00
#define GL_RENDERER                         0x1F01
#define GL_VERSION                          0x1F02
#define GL_EXTENSIONS                       0x1F03

/* TextureCoordName */
#define GL_S                                0x2000
#define GL_T                                0x2001
#define GL_R                                0x2002
#define GL_Q                                0x2003

/* TextureEnvMode */
#define GL_MODULATE                         0x2100
#define GL_DECAL                            0x2101

/* TextureEnvParameter */
#define GL_TEXTURE_ENV_MODE                 0x2200
#define GL_TEXTURE_ENV_COLOR                0x2201

/* TextureEnvTarget */
#define GL_TEXTURE_ENV                      0x2300

/* TextureGenMode */
#define GL_EYE_LINEAR                       0x2400
#define GL_OBJECT_LINEAR                    0x2401
#define GL_SPHERE_MAP                       0x2402

/* TextureGenParameter */
#define GL_TEXTURE_GEN_MODE                 0x2500
#define GL_OBJECT_PLANE                     0x2501
#define GL_EYE_PLANE                        0x2502

/* TextureMagFilter */
#define GL_NEAREST                          0x2600
#define GL_LINEAR                           0x2601

/* TextureMinFilter */
#define GL_NEAREST_MIPMAP_NEAREST           0x2700
#define GL_LINEAR_MIPMAP_NEAREST            0x2701
#define GL_NEAREST_MIPMAP_LINEAR            0x2702
#define GL_LINEAR_MIPMAP_LINEAR             0x2703

/* TextureParameterName */
#define GL_TEXTURE_MAG_FILTER               0x2800
#define GL_TEXTURE_MIN_FILTER               0x2801
#define GL_TEXTURE_WRAP_S                   0x2802
#define GL_TEXTURE_WRAP_T                   0x2803

/* TextureTarget */
#define GL_PROXY_TEXTURE_1D                 0x8063
#define GL_PROXY_TEXTURE_2D                 0x8064

/* TextureWrapMode */
#define GL_CLAMP                            0x2900
#define GL_REPEAT                           0x2901

/* PixelInternalFormat */
#define GL_R3_G3_B2                         0x2A10
#define GL_ALPHA4                           0x803B
#define GL_ALPHA8                           0x803C
#define GL_ALPHA12                          0x803D
#define GL_ALPHA16                          0x803E
#define GL_LUMINANCE4                       0x803F
#define GL_LUMINANCE8                       0x8040
#define GL_LUMINANCE12                      0x8041
#define GL_LUMINANCE16                      0x8042
#define GL_LUMINANCE4_ALPHA4                0x8043
#define GL_LUMINANCE6_ALPHA2                0x8044
#define GL_LUMINANCE8_ALPHA8                0x8045
#define GL_LUMINANCE12_ALPHA4               0x8046
#define GL_LUMINANCE12_ALPHA12              0x8047
#define GL_LUMINANCE16_ALPHA16              0x8048
#define GL_INTENSITY                        0x8049
#define GL_INTENSITY4                       0x804A
#define GL_INTENSITY8                       0x804B
#define GL_INTENSITY12                      0x804C
#define GL_INTENSITY16                      0x804D
#define GL_RGB4                             0x804F
#define GL_RGB5                             0x8050
#define GL_RGB8                             0x8051
#define GL_RGB10                            0x8052
#define GL_RGB12                            0x8053
#define GL_RGB16                            0x8054
#define GL_RGBA2                            0x8055
#define GL_RGBA4                            0x8056
#define GL_RGB5_A1                          0x8057
#define GL_RGBA8                            0x8058
#define GL_RGB10_A2                         0x8059
#define GL_RGBA12                           0x805A
#define GL_RGBA16                           0x805B

/* InterleavedArrayFormat */
#define GL_V2F                              0x2A20
#define GL_V3F                              0x2A21
#define GL_C4UB_V2F                         0x2A22
#define GL_C4UB_V3F                         0x2A23
#define GL_C3F_V3F                          0x2A24
#define GL_N3F_V3F                          0x2A25
#define GL_C4F_N3F_V3F                      0x2A26
#define GL_T2F_V3F                          0x2A27
#define GL_T4F_V4F                          0x2A28
#define GL_T2F_C4UB_V3F                     0x2A29
#define GL_T2F_C3F_V3F                      0x2A2A
#define GL_T2F_N3F_V3F                      0x2A2B
#define GL_T2F_C4F_N3F_V3F                  0x2A2C
#define GL_T4F_C4F_N3F_V4F                  0x2A2D

/* ClipPlaneName */
#define GL_CLIP_PLANE0                      0x3000
#define GL_CLIP_PLANE1                      0x3001
#define GL_CLIP_PLANE2                      0x3002
#define GL_CLIP_PLANE3                      0x3003
#define GL_CLIP_PLANE4                      0x3004
#define GL_CLIP_PLANE5                      0x3005

/* LightName */
#define GL_LIGHT0                           0x4000
#define GL_LIGHT1                           0x4001
#define GL_LIGHT2                           0x4002
#define GL_LIGHT3                           0x4003
#define GL_LIGHT4                           0x4004
#define GL_LIGHT5                           0x4005
#define GL_LIGHT6                           0x4006
#define GL_LIGHT7                           0x4007

/*
 * Base OpenGL 1.1 functionality
 */

#ifdef	BASEGL11_DYNAMIC

/*
 * TODO: still not realized; look at vglext.c
 * FIXME: the only way to load basegl is GetProcAddress() / dlsym().
 */

#ifdef	DLL_TARGET_DCC

#ifdef	__cplusplus
extern "C" {
#endif /* __cplusplus */

/*
 * Prototypes
 */
VGLEXTAPI void VGLEXTAPIENTRY glAccum (GLenum op, GLfloat value);
VGLEXTAPI void VGLEXTAPIENTRY glAlphaFunc (GLenum func, GLclampf ref);
VGLEXTAPI GLboolean VGLEXTAPIENTRY glAreTexturesResident (GLsizei n, const GLuint *textures, GLboolean *residences);
VGLEXTAPI void VGLEXTAPIENTRY glArrayElement (GLint i);
VGLEXTAPI void VGLEXTAPIENTRY glBegin (GLenum mode);
VGLEXTAPI void VGLEXTAPIENTRY glBindTexture (GLenum target, GLuint texture);
VGLEXTAPI void VGLEXTAPIENTRY glBitmap (GLsizei width, GLsizei height, GLfloat xorig, GLfloat yorig, GLfloat xmove, GLfloat ymove, const GLubyte *bitmap);
VGLEXTAPI void VGLEXTAPIENTRY glBlendFunc (GLenum sfactor, GLenum dfactor);
VGLEXTAPI void VGLEXTAPIENTRY glCallList (GLuint list);
VGLEXTAPI void VGLEXTAPIENTRY glCallLists (GLsizei n, GLenum type, const GLvoid *lists);
VGLEXTAPI void VGLEXTAPIENTRY glClear (GLbitfield mask);
VGLEXTAPI void VGLEXTAPIENTRY glClearAccum (GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha);
VGLEXTAPI void VGLEXTAPIENTRY glClearColor (GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha);
VGLEXTAPI void VGLEXTAPIENTRY glClearDepth (GLclampd depth);
VGLEXTAPI void VGLEXTAPIENTRY glClearIndex (GLfloat c);
VGLEXTAPI void VGLEXTAPIENTRY glClearStencil (GLint s);
VGLEXTAPI void VGLEXTAPIENTRY glClipPlane (GLenum plane, const GLdouble *equation);
VGLEXTAPI void VGLEXTAPIENTRY glColor3b (GLbyte red, GLbyte green, GLbyte blue);
VGLEXTAPI void VGLEXTAPIENTRY glColor3bv (const GLbyte *v);
VGLEXTAPI void VGLEXTAPIENTRY glColor3d (GLdouble red, GLdouble green, GLdouble blue);
VGLEXTAPI void VGLEXTAPIENTRY glColor3dv (const GLdouble *v);
VGLEXTAPI void VGLEXTAPIENTRY glColor3f (GLfloat red, GLfloat green, GLfloat blue);
VGLEXTAPI void VGLEXTAPIENTRY glColor3fv (const GLfloat *v);
VGLEXTAPI void VGLEXTAPIENTRY glColor3i (GLint red, GLint green, GLint blue);
VGLEXTAPI void VGLEXTAPIENTRY glColor3iv (const GLint *v);
VGLEXTAPI void VGLEXTAPIENTRY glColor3s (GLshort red, GLshort green, GLshort blue);
VGLEXTAPI void VGLEXTAPIENTRY glColor3sv (const GLshort *v);
VGLEXTAPI void VGLEXTAPIENTRY glColor3ub (GLubyte red, GLubyte green, GLubyte blue);
VGLEXTAPI void VGLEXTAPIENTRY glColor3ubv (const GLubyte *v);
VGLEXTAPI void VGLEXTAPIENTRY glColor3ui (GLuint red, GLuint green, GLuint blue);
VGLEXTAPI void VGLEXTAPIENTRY glColor3uiv (const GLuint *v);
VGLEXTAPI void VGLEXTAPIENTRY glColor3us (GLushort red, GLushort green, GLushort blue);
VGLEXTAPI void VGLEXTAPIENTRY glColor3usv (const GLushort *v);
VGLEXTAPI void VGLEXTAPIENTRY glColor4b (GLbyte red, GLbyte green, GLbyte blue, GLbyte alpha);
VGLEXTAPI void VGLEXTAPIENTRY glColor4bv (const GLbyte *v);
VGLEXTAPI void VGLEXTAPIENTRY glColor4d (GLdouble red, GLdouble green, GLdouble blue, GLdouble alpha);
VGLEXTAPI void VGLEXTAPIENTRY glColor4dv (const GLdouble *v);
VGLEXTAPI void VGLEXTAPIENTRY glColor4f (GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha);
VGLEXTAPI void VGLEXTAPIENTRY glColor4fv (const GLfloat *v);
VGLEXTAPI void VGLEXTAPIENTRY glColor4i (GLint red, GLint green, GLint blue, GLint alpha);
VGLEXTAPI void VGLEXTAPIENTRY glColor4iv (const GLint *v);
VGLEXTAPI void VGLEXTAPIENTRY glColor4s (GLshort red, GLshort green, GLshort blue, GLshort alpha);
VGLEXTAPI void VGLEXTAPIENTRY glColor4sv (const GLshort *v);
VGLEXTAPI void VGLEXTAPIENTRY glColor4ub (GLubyte red, GLubyte green, GLubyte blue, GLubyte alpha);
VGLEXTAPI void VGLEXTAPIENTRY glColor4ubv (const GLubyte *v);
VGLEXTAPI void VGLEXTAPIENTRY glColor4ui (GLuint red, GLuint green, GLuint blue, GLuint alpha);
VGLEXTAPI void VGLEXTAPIENTRY glColor4uiv (const GLuint *v);
VGLEXTAPI void VGLEXTAPIENTRY glColor4us (GLushort red, GLushort green, GLushort blue, GLushort alpha);
VGLEXTAPI void VGLEXTAPIENTRY glColor4usv (const GLushort *v);
VGLEXTAPI void VGLEXTAPIENTRY glColorMask (GLboolean red, GLboolean green, GLboolean blue, GLboolean alpha);
VGLEXTAPI void VGLEXTAPIENTRY glColorMaterial (GLenum face, GLenum mode);
VGLEXTAPI void VGLEXTAPIENTRY glColorPointer (GLint size, GLenum type, GLsizei stride, const GLvoid *pointer);
VGLEXTAPI void VGLEXTAPIENTRY glCopyPixels (GLint x, GLint y, GLsizei width, GLsizei height, GLenum type);
VGLEXTAPI void VGLEXTAPIENTRY glCopyTexImage1D (GLenum target, GLint level, GLenum internalformat, GLint x, GLint y, GLsizei width, GLint border);
VGLEXTAPI void VGLEXTAPIENTRY glCopyTexImage2D (GLenum target, GLint level, GLenum internalformat, GLint x, GLint y, GLsizei width, GLsizei height, GLint border);
VGLEXTAPI void VGLEXTAPIENTRY glCopyTexSubImage1D (GLenum target, GLint level, GLint xoffset, GLint x, GLint y, GLsizei width);
VGLEXTAPI void VGLEXTAPIENTRY glCopyTexSubImage2D (GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint x, GLint y, GLsizei width, GLsizei height);
VGLEXTAPI void VGLEXTAPIENTRY glCullFace (GLenum mode);
VGLEXTAPI void VGLEXTAPIENTRY glDeleteLists (GLuint list, GLsizei range);
VGLEXTAPI void VGLEXTAPIENTRY glDeleteTextures (GLsizei n, const GLuint *textures);
VGLEXTAPI void VGLEXTAPIENTRY glDepthFunc (GLenum func);
VGLEXTAPI void VGLEXTAPIENTRY glDepthMask (GLboolean flag);
VGLEXTAPI void VGLEXTAPIENTRY glDepthRange (GLclampd zNear, GLclampd zFar);
VGLEXTAPI void VGLEXTAPIENTRY glDisable (GLenum cap);
VGLEXTAPI void VGLEXTAPIENTRY glDisableClientState (GLenum array);
VGLEXTAPI void VGLEXTAPIENTRY glDrawArrays (GLenum mode, GLint first, GLsizei count);
VGLEXTAPI void VGLEXTAPIENTRY glDrawBuffer (GLenum mode);
VGLEXTAPI void VGLEXTAPIENTRY glDrawElements (GLenum mode, GLsizei count, GLenum type, const GLvoid *indices);
VGLEXTAPI void VGLEXTAPIENTRY glDrawPixels (GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels);
VGLEXTAPI void VGLEXTAPIENTRY glEdgeFlag (GLboolean flag);
VGLEXTAPI void VGLEXTAPIENTRY glEdgeFlagPointer (GLsizei stride, const GLboolean *pointer);
VGLEXTAPI void VGLEXTAPIENTRY glEdgeFlagv (const GLboolean *flag);
VGLEXTAPI void VGLEXTAPIENTRY glEnable (GLenum cap);
VGLEXTAPI void VGLEXTAPIENTRY glEnableClientState (GLenum array);
VGLEXTAPI void VGLEXTAPIENTRY glEnd (void);
VGLEXTAPI void VGLEXTAPIENTRY glEndList (void);
VGLEXTAPI void VGLEXTAPIENTRY glEvalCoord1d (GLdouble u);
VGLEXTAPI void VGLEXTAPIENTRY glEvalCoord1dv (const GLdouble *u);
VGLEXTAPI void VGLEXTAPIENTRY glEvalCoord1f (GLfloat u);
VGLEXTAPI void VGLEXTAPIENTRY glEvalCoord1fv (const GLfloat *u);
VGLEXTAPI void VGLEXTAPIENTRY glEvalCoord2d (GLdouble u, GLdouble v);
VGLEXTAPI void VGLEXTAPIENTRY glEvalCoord2dv (const GLdouble *u);
VGLEXTAPI void VGLEXTAPIENTRY glEvalCoord2f (GLfloat u, GLfloat v);
VGLEXTAPI void VGLEXTAPIENTRY glEvalCoord2fv (const GLfloat *u);
VGLEXTAPI void VGLEXTAPIENTRY glEvalMesh1 (GLenum mode, GLint i1, GLint i2);
VGLEXTAPI void VGLEXTAPIENTRY glEvalMesh2 (GLenum mode, GLint i1, GLint i2, GLint j1, GLint j2);
VGLEXTAPI void VGLEXTAPIENTRY glEvalPoint1 (GLint i);
VGLEXTAPI void VGLEXTAPIENTRY glEvalPoint2 (GLint i, GLint j);
VGLEXTAPI void VGLEXTAPIENTRY glFeedbackBuffer (GLsizei size, GLenum type, GLfloat *buffer);
VGLEXTAPI void VGLEXTAPIENTRY glFinish (void);
VGLEXTAPI void VGLEXTAPIENTRY glFlush (void);
VGLEXTAPI void VGLEXTAPIENTRY glFogf (GLenum pname, GLfloat param);
VGLEXTAPI void VGLEXTAPIENTRY glFogfv (GLenum pname, const GLfloat *params);
VGLEXTAPI void VGLEXTAPIENTRY glFogi (GLenum pname, GLint param);
VGLEXTAPI void VGLEXTAPIENTRY glFogiv (GLenum pname, const GLint *params);
VGLEXTAPI void VGLEXTAPIENTRY glFrontFace (GLenum mode);
VGLEXTAPI void VGLEXTAPIENTRY glFrustum (GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble zNear, GLdouble zFar);
VGLEXTAPI GLuint VGLEXTAPIENTRY glGenLists (GLsizei range);
VGLEXTAPI void VGLEXTAPIENTRY glGenTextures (GLsizei n, GLuint *textures);
VGLEXTAPI void VGLEXTAPIENTRY glGetBooleanv (GLenum pname, GLboolean *params);
VGLEXTAPI void VGLEXTAPIENTRY glGetClipPlane (GLenum plane, GLdouble *equation);
VGLEXTAPI void VGLEXTAPIENTRY glGetDoublev (GLenum pname, GLdouble *params);
VGLEXTAPI GLenum VGLEXTAPIENTRY glGetError (void);
VGLEXTAPI void VGLEXTAPIENTRY glGetFloatv (GLenum pname, GLfloat *params);
VGLEXTAPI void VGLEXTAPIENTRY glGetIntegerv (GLenum pname, GLint *params);
VGLEXTAPI void VGLEXTAPIENTRY glGetLightfv (GLenum light, GLenum pname, GLfloat *params);
VGLEXTAPI void VGLEXTAPIENTRY glGetLightiv (GLenum light, GLenum pname, GLint *params);
VGLEXTAPI void VGLEXTAPIENTRY glGetMapdv (GLenum target, GLenum query, GLdouble *v);
VGLEXTAPI void VGLEXTAPIENTRY glGetMapfv (GLenum target, GLenum query, GLfloat *v);
VGLEXTAPI void VGLEXTAPIENTRY glGetMapiv (GLenum target, GLenum query, GLint *v);
VGLEXTAPI void VGLEXTAPIENTRY glGetMaterialfv (GLenum face, GLenum pname, GLfloat *params);
VGLEXTAPI void VGLEXTAPIENTRY glGetMaterialiv (GLenum face, GLenum pname, GLint *params);
VGLEXTAPI void VGLEXTAPIENTRY glGetPixelMapfv (GLenum map, GLfloat *values);
VGLEXTAPI void VGLEXTAPIENTRY glGetPixelMapuiv (GLenum map, GLuint *values);
VGLEXTAPI void VGLEXTAPIENTRY glGetPixelMapusv (GLenum map, GLushort *values);
VGLEXTAPI void VGLEXTAPIENTRY glGetPointerv (GLenum pname, GLvoid* *params);
VGLEXTAPI void VGLEXTAPIENTRY glGetPolygonStipple (GLubyte *mask);
VGLEXTAPI const GLubyte * VGLEXTAPIENTRY glGetString (GLenum name);
VGLEXTAPI void VGLEXTAPIENTRY glGetTexEnvfv (GLenum target, GLenum pname, GLfloat *params);
VGLEXTAPI void VGLEXTAPIENTRY glGetTexEnviv (GLenum target, GLenum pname, GLint *params);
VGLEXTAPI void VGLEXTAPIENTRY glGetTexGendv (GLenum coord, GLenum pname, GLdouble *params);
VGLEXTAPI void VGLEXTAPIENTRY glGetTexGenfv (GLenum coord, GLenum pname, GLfloat *params);
VGLEXTAPI void VGLEXTAPIENTRY glGetTexGeniv (GLenum coord, GLenum pname, GLint *params);
VGLEXTAPI void VGLEXTAPIENTRY glGetTexImage (GLenum target, GLint level, GLenum format, GLenum type, GLvoid *pixels);
VGLEXTAPI void VGLEXTAPIENTRY glGetTexLevelParameterfv (GLenum target, GLint level, GLenum pname, GLfloat *params);
VGLEXTAPI void VGLEXTAPIENTRY glGetTexLevelParameteriv (GLenum target, GLint level, GLenum pname, GLint *params);
VGLEXTAPI void VGLEXTAPIENTRY glGetTexParameterfv (GLenum target, GLenum pname, GLfloat *params);
VGLEXTAPI void VGLEXTAPIENTRY glGetTexParameteriv (GLenum target, GLenum pname, GLint *params);
VGLEXTAPI void VGLEXTAPIENTRY glHint (GLenum target, GLenum mode);
VGLEXTAPI void VGLEXTAPIENTRY glIndexMask (GLuint mask);
VGLEXTAPI void VGLEXTAPIENTRY glIndexPointer (GLenum type, GLsizei stride, const GLvoid *pointer);
VGLEXTAPI void VGLEXTAPIENTRY glIndexd (GLdouble c);
VGLEXTAPI void VGLEXTAPIENTRY glIndexdv (const GLdouble *c);
VGLEXTAPI void VGLEXTAPIENTRY glIndexf (GLfloat c);
VGLEXTAPI void VGLEXTAPIENTRY glIndexfv (const GLfloat *c);
VGLEXTAPI void VGLEXTAPIENTRY glIndexi (GLint c);
VGLEXTAPI void VGLEXTAPIENTRY glIndexiv (const GLint *c);
VGLEXTAPI void VGLEXTAPIENTRY glIndexs (GLshort c);
VGLEXTAPI void VGLEXTAPIENTRY glIndexsv (const GLshort *c);
VGLEXTAPI void VGLEXTAPIENTRY glIndexub (GLubyte c);
VGLEXTAPI void VGLEXTAPIENTRY glIndexubv (const GLubyte *c);
VGLEXTAPI void VGLEXTAPIENTRY glInitNames (void);
VGLEXTAPI void VGLEXTAPIENTRY glInterleavedArrays (GLenum format, GLsizei stride, const GLvoid *pointer);
VGLEXTAPI GLboolean VGLEXTAPIENTRY glIsEnabled (GLenum cap);
VGLEXTAPI GLboolean VGLEXTAPIENTRY glIsList (GLuint list);
VGLEXTAPI GLboolean VGLEXTAPIENTRY glIsTexture (GLuint texture);
VGLEXTAPI void VGLEXTAPIENTRY glLightModelf (GLenum pname, GLfloat param);
VGLEXTAPI void VGLEXTAPIENTRY glLightModelfv (GLenum pname, const GLfloat *params);
VGLEXTAPI void VGLEXTAPIENTRY glLightModeli (GLenum pname, GLint param);
VGLEXTAPI void VGLEXTAPIENTRY glLightModeliv (GLenum pname, const GLint *params);
VGLEXTAPI void VGLEXTAPIENTRY glLightf (GLenum light, GLenum pname, GLfloat param);
VGLEXTAPI void VGLEXTAPIENTRY glLightfv (GLenum light, GLenum pname, const GLfloat *params);
VGLEXTAPI void VGLEXTAPIENTRY glLighti (GLenum light, GLenum pname, GLint param);
VGLEXTAPI void VGLEXTAPIENTRY glLightiv (GLenum light, GLenum pname, const GLint *params);
VGLEXTAPI void VGLEXTAPIENTRY glLineStipple (GLint factor, GLushort pattern);
VGLEXTAPI void VGLEXTAPIENTRY glLineWidth (GLfloat width);
VGLEXTAPI void VGLEXTAPIENTRY glListBase (GLuint base);
VGLEXTAPI void VGLEXTAPIENTRY glLoadIdentity (void);
VGLEXTAPI void VGLEXTAPIENTRY glLoadMatrixd (const GLdouble *m);
VGLEXTAPI void VGLEXTAPIENTRY glLoadMatrixf (const GLfloat *m);
VGLEXTAPI void VGLEXTAPIENTRY glLoadName (GLuint name);
VGLEXTAPI void VGLEXTAPIENTRY glLogicOp (GLenum opcode);
VGLEXTAPI void VGLEXTAPIENTRY glMap1d (GLenum target, GLdouble u1, GLdouble u2, GLint stride, GLint order, const GLdouble *points);
VGLEXTAPI void VGLEXTAPIENTRY glMap1f (GLenum target, GLfloat u1, GLfloat u2, GLint stride, GLint order, const GLfloat *points);
VGLEXTAPI void VGLEXTAPIENTRY glMap2d (GLenum target, GLdouble u1, GLdouble u2, GLint ustride, GLint uorder, GLdouble v1, GLdouble v2, GLint vstride, GLint vorder, const GLdouble *points);
VGLEXTAPI void VGLEXTAPIENTRY glMap2f (GLenum target, GLfloat u1, GLfloat u2, GLint ustride, GLint uorder, GLfloat v1, GLfloat v2, GLint vstride, GLint vorder, const GLfloat *points);
VGLEXTAPI void VGLEXTAPIENTRY glMapGrid1d (GLint un, GLdouble u1, GLdouble u2);
VGLEXTAPI void VGLEXTAPIENTRY glMapGrid1f (GLint un, GLfloat u1, GLfloat u2);
VGLEXTAPI void VGLEXTAPIENTRY glMapGrid2d (GLint un, GLdouble u1, GLdouble u2, GLint vn, GLdouble v1, GLdouble v2);
VGLEXTAPI void VGLEXTAPIENTRY glMapGrid2f (GLint un, GLfloat u1, GLfloat u2, GLint vn, GLfloat v1, GLfloat v2);
VGLEXTAPI void VGLEXTAPIENTRY glMaterialf (GLenum face, GLenum pname, GLfloat param);
VGLEXTAPI void VGLEXTAPIENTRY glMaterialfv (GLenum face, GLenum pname, const GLfloat *params);
VGLEXTAPI void VGLEXTAPIENTRY glMateriali (GLenum face, GLenum pname, GLint param);
VGLEXTAPI void VGLEXTAPIENTRY glMaterialiv (GLenum face, GLenum pname, const GLint *params);
VGLEXTAPI void VGLEXTAPIENTRY glMatrixMode (GLenum mode);
VGLEXTAPI void VGLEXTAPIENTRY glMultMatrixd (const GLdouble *m);
VGLEXTAPI void VGLEXTAPIENTRY glMultMatrixf (const GLfloat *m);
VGLEXTAPI void VGLEXTAPIENTRY glNewList (GLuint list, GLenum mode);
VGLEXTAPI void VGLEXTAPIENTRY glNormal3b (GLbyte nx, GLbyte ny, GLbyte nz);
VGLEXTAPI void VGLEXTAPIENTRY glNormal3bv (const GLbyte *v);
VGLEXTAPI void VGLEXTAPIENTRY glNormal3d (GLdouble nx, GLdouble ny, GLdouble nz);
VGLEXTAPI void VGLEXTAPIENTRY glNormal3dv (const GLdouble *v);
VGLEXTAPI void VGLEXTAPIENTRY glNormal3f (GLfloat nx, GLfloat ny, GLfloat nz);
VGLEXTAPI void VGLEXTAPIENTRY glNormal3fv (const GLfloat *v);
VGLEXTAPI void VGLEXTAPIENTRY glNormal3i (GLint nx, GLint ny, GLint nz);
VGLEXTAPI void VGLEXTAPIENTRY glNormal3iv (const GLint *v);
VGLEXTAPI void VGLEXTAPIENTRY glNormal3s (GLshort nx, GLshort ny, GLshort nz);
VGLEXTAPI void VGLEXTAPIENTRY glNormal3sv (const GLshort *v);
VGLEXTAPI void VGLEXTAPIENTRY glNormalPointer (GLenum type, GLsizei stride, const GLvoid *pointer);
VGLEXTAPI void VGLEXTAPIENTRY glOrtho (GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble zNear, GLdouble zFar);
VGLEXTAPI void VGLEXTAPIENTRY glPassThrough (GLfloat token);
VGLEXTAPI void VGLEXTAPIENTRY glPixelMapfv (GLenum map, GLint mapsize, const GLfloat *values);
VGLEXTAPI void VGLEXTAPIENTRY glPixelMapuiv (GLenum map, GLint mapsize, const GLuint *values);
VGLEXTAPI void VGLEXTAPIENTRY glPixelMapusv (GLenum map, GLint mapsize, const GLushort *values);
VGLEXTAPI void VGLEXTAPIENTRY glPixelStoref (GLenum pname, GLfloat param);
VGLEXTAPI void VGLEXTAPIENTRY glPixelStorei (GLenum pname, GLint param);
VGLEXTAPI void VGLEXTAPIENTRY glPixelTransferf (GLenum pname, GLfloat param);
VGLEXTAPI void VGLEXTAPIENTRY glPixelTransferi (GLenum pname, GLint param);
VGLEXTAPI void VGLEXTAPIENTRY glPixelZoom (GLfloat xfactor, GLfloat yfactor);
VGLEXTAPI void VGLEXTAPIENTRY glPointSize (GLfloat size);
VGLEXTAPI void VGLEXTAPIENTRY glPolygonMode (GLenum face, GLenum mode);
VGLEXTAPI void VGLEXTAPIENTRY glPolygonOffset (GLfloat factor, GLfloat units);
VGLEXTAPI void VGLEXTAPIENTRY glPolygonStipple (const GLubyte *mask);
VGLEXTAPI void VGLEXTAPIENTRY glPopAttrib (void);
VGLEXTAPI void VGLEXTAPIENTRY glPopClientAttrib (void);
VGLEXTAPI void VGLEXTAPIENTRY glPopMatrix (void);
VGLEXTAPI void VGLEXTAPIENTRY glPopName (void);
VGLEXTAPI void VGLEXTAPIENTRY glPrioritizeTextures (GLsizei n, const GLuint *textures, const GLclampf *priorities);
VGLEXTAPI void VGLEXTAPIENTRY glPushAttrib (GLbitfield mask);
VGLEXTAPI void VGLEXTAPIENTRY glPushClientAttrib (GLbitfield mask);
VGLEXTAPI void VGLEXTAPIENTRY glPushMatrix (void);
VGLEXTAPI void VGLEXTAPIENTRY glPushName (GLuint name);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos2d (GLdouble x, GLdouble y);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos2dv (const GLdouble *v);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos2f (GLfloat x, GLfloat y);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos2fv (const GLfloat *v);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos2i (GLint x, GLint y);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos2iv (const GLint *v);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos2s (GLshort x, GLshort y);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos2sv (const GLshort *v);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos3d (GLdouble x, GLdouble y, GLdouble z);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos3dv (const GLdouble *v);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos3f (GLfloat x, GLfloat y, GLfloat z);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos3fv (const GLfloat *v);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos3i (GLint x, GLint y, GLint z);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos3iv (const GLint *v);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos3s (GLshort x, GLshort y, GLshort z);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos3sv (const GLshort *v);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos4d (GLdouble x, GLdouble y, GLdouble z, GLdouble w);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos4dv (const GLdouble *v);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos4f (GLfloat x, GLfloat y, GLfloat z, GLfloat w);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos4fv (const GLfloat *v);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos4i (GLint x, GLint y, GLint z, GLint w);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos4iv (const GLint *v);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos4s (GLshort x, GLshort y, GLshort z, GLshort w);
VGLEXTAPI void VGLEXTAPIENTRY glRasterPos4sv (const GLshort *v);
VGLEXTAPI void VGLEXTAPIENTRY glReadBuffer (GLenum mode);
VGLEXTAPI void VGLEXTAPIENTRY glReadPixels (GLint x, GLint y, GLsizei width, GLsizei height, GLenum format, GLenum type, GLvoid *pixels);
VGLEXTAPI void VGLEXTAPIENTRY glRectd (GLdouble x1, GLdouble y1, GLdouble x2, GLdouble y2);
VGLEXTAPI void VGLEXTAPIENTRY glRectdv (const GLdouble *v1, const GLdouble *v2);
VGLEXTAPI void VGLEXTAPIENTRY glRectf (GLfloat x1, GLfloat y1, GLfloat x2, GLfloat y2);
VGLEXTAPI void VGLEXTAPIENTRY glRectfv (const GLfloat *v1, const GLfloat *v2);
VGLEXTAPI void VGLEXTAPIENTRY glRecti (GLint x1, GLint y1, GLint x2, GLint y2);
VGLEXTAPI void VGLEXTAPIENTRY glRectiv (const GLint *v1, const GLint *v2);
VGLEXTAPI void VGLEXTAPIENTRY glRects (GLshort x1, GLshort y1, GLshort x2, GLshort y2);
VGLEXTAPI void VGLEXTAPIENTRY glRectsv (const GLshort *v1, const GLshort *v2);
VGLEXTAPI GLint VGLEXTAPIENTRY glRenderMode (GLenum mode);
VGLEXTAPI void VGLEXTAPIENTRY glRotated (GLdouble angle, GLdouble x, GLdouble y, GLdouble z);
VGLEXTAPI void VGLEXTAPIENTRY glRotatef (GLfloat angle, GLfloat x, GLfloat y, GLfloat z);
VGLEXTAPI void VGLEXTAPIENTRY glScaled (GLdouble x, GLdouble y, GLdouble z);
VGLEXTAPI void VGLEXTAPIENTRY glScalef (GLfloat x, GLfloat y, GLfloat z);
VGLEXTAPI void VGLEXTAPIENTRY glScissor (GLint x, GLint y, GLsizei width, GLsizei height);
VGLEXTAPI void VGLEXTAPIENTRY glSelectBuffer (GLsizei size, GLuint *buffer);
VGLEXTAPI void VGLEXTAPIENTRY glShadeModel (GLenum mode);
VGLEXTAPI void VGLEXTAPIENTRY glStencilFunc (GLenum func, GLint ref, GLuint mask);
VGLEXTAPI void VGLEXTAPIENTRY glStencilMask (GLuint mask);
VGLEXTAPI void VGLEXTAPIENTRY glStencilOp (GLenum fail, GLenum zfail, GLenum zpass);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord1d (GLdouble s);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord1dv (const GLdouble *v);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord1f (GLfloat s);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord1fv (const GLfloat *v);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord1i (GLint s);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord1iv (const GLint *v);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord1s (GLshort s);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord1sv (const GLshort *v);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord2d (GLdouble s, GLdouble t);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord2dv (const GLdouble *v);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord2f (GLfloat s, GLfloat t);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord2fv (const GLfloat *v);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord2i (GLint s, GLint t);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord2iv (const GLint *v);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord2s (GLshort s, GLshort t);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord2sv (const GLshort *v);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord3d (GLdouble s, GLdouble t, GLdouble r);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord3dv (const GLdouble *v);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord3f (GLfloat s, GLfloat t, GLfloat r);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord3fv (const GLfloat *v);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord3i (GLint s, GLint t, GLint r);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord3iv (const GLint *v);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord3s (GLshort s, GLshort t, GLshort r);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord3sv (const GLshort *v);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord4d (GLdouble s, GLdouble t, GLdouble r, GLdouble q);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord4dv (const GLdouble *v);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord4f (GLfloat s, GLfloat t, GLfloat r, GLfloat q);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord4fv (const GLfloat *v);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord4i (GLint s, GLint t, GLint r, GLint q);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord4iv (const GLint *v);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord4s (GLshort s, GLshort t, GLshort r, GLshort q);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoord4sv (const GLshort *v);
VGLEXTAPI void VGLEXTAPIENTRY glTexCoordPointer (GLint size, GLenum type, GLsizei stride, const GLvoid *pointer);
VGLEXTAPI void VGLEXTAPIENTRY glTexEnvf (GLenum target, GLenum pname, GLfloat param);
VGLEXTAPI void VGLEXTAPIENTRY glTexEnvfv (GLenum target, GLenum pname, const GLfloat *params);
VGLEXTAPI void VGLEXTAPIENTRY glTexEnvi (GLenum target, GLenum pname, GLint param);
VGLEXTAPI void VGLEXTAPIENTRY glTexEnviv (GLenum target, GLenum pname, const GLint *params);
VGLEXTAPI void VGLEXTAPIENTRY glTexGend (GLenum coord, GLenum pname, GLdouble param);
VGLEXTAPI void VGLEXTAPIENTRY glTexGendv (GLenum coord, GLenum pname, const GLdouble *params);
VGLEXTAPI void VGLEXTAPIENTRY glTexGenf (GLenum coord, GLenum pname, GLfloat param);
VGLEXTAPI void VGLEXTAPIENTRY glTexGenfv (GLenum coord, GLenum pname, const GLfloat *params);
VGLEXTAPI void VGLEXTAPIENTRY glTexGeni (GLenum coord, GLenum pname, GLint param);
VGLEXTAPI void VGLEXTAPIENTRY glTexGeniv (GLenum coord, GLenum pname, const GLint *params);
VGLEXTAPI void VGLEXTAPIENTRY glTexImage1D (GLenum target, GLint level, GLint components, GLsizei width, GLint border, GLenum format, GLenum type, const GLvoid *pixels);
VGLEXTAPI void VGLEXTAPIENTRY glTexImage2D (GLenum target, GLint level, GLint components, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, const GLvoid *pixels);
VGLEXTAPI void VGLEXTAPIENTRY glTexParameterf (GLenum target, GLenum pname, GLfloat param);
VGLEXTAPI void VGLEXTAPIENTRY glTexParameterfv (GLenum target, GLenum pname, const GLfloat *params);
VGLEXTAPI void VGLEXTAPIENTRY glTexParameteri (GLenum target, GLenum pname, GLint param);
VGLEXTAPI void VGLEXTAPIENTRY glTexParameteriv (GLenum target, GLenum pname, const GLint *params);
VGLEXTAPI void VGLEXTAPIENTRY glTexSubImage1D (GLenum target, GLint level, GLint xoffset, GLsizei width, GLenum format, GLenum type, const GLvoid *pixels);
VGLEXTAPI void VGLEXTAPIENTRY glTexSubImage2D (GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels);
VGLEXTAPI void VGLEXTAPIENTRY glTranslated (GLdouble x, GLdouble y, GLdouble z);
VGLEXTAPI void VGLEXTAPIENTRY glTranslatef (GLfloat x, GLfloat y, GLfloat z);
VGLEXTAPI void VGLEXTAPIENTRY glVertex2d (GLdouble x, GLdouble y);
VGLEXTAPI void VGLEXTAPIENTRY glVertex2dv (const GLdouble *v);
VGLEXTAPI void VGLEXTAPIENTRY glVertex2f (GLfloat x, GLfloat y);
VGLEXTAPI void VGLEXTAPIENTRY glVertex2fv (const GLfloat *v);
VGLEXTAPI void VGLEXTAPIENTRY glVertex2i (GLint x, GLint y);
VGLEXTAPI void VGLEXTAPIENTRY glVertex2iv (const GLint *v);
VGLEXTAPI void VGLEXTAPIENTRY glVertex2s (GLshort x, GLshort y);
VGLEXTAPI void VGLEXTAPIENTRY glVertex2sv (const GLshort *v);
VGLEXTAPI void VGLEXTAPIENTRY glVertex3d (GLdouble x, GLdouble y, GLdouble z);
VGLEXTAPI void VGLEXTAPIENTRY glVertex3dv (const GLdouble *v);
VGLEXTAPI void VGLEXTAPIENTRY glVertex3f (GLfloat x, GLfloat y, GLfloat z);
VGLEXTAPI void VGLEXTAPIENTRY glVertex3fv (const GLfloat *v);
VGLEXTAPI void VGLEXTAPIENTRY glVertex3i (GLint x, GLint y, GLint z);
VGLEXTAPI void VGLEXTAPIENTRY glVertex3iv (const GLint *v);
VGLEXTAPI void VGLEXTAPIENTRY glVertex3s (GLshort x, GLshort y, GLshort z);
VGLEXTAPI void VGLEXTAPIENTRY glVertex3sv (const GLshort *v);
VGLEXTAPI void VGLEXTAPIENTRY glVertex4d (GLdouble x, GLdouble y, GLdouble z, GLdouble w);
VGLEXTAPI void VGLEXTAPIENTRY glVertex4dv (const GLdouble *v);
VGLEXTAPI void VGLEXTAPIENTRY glVertex4f (GLfloat x, GLfloat y, GLfloat z, GLfloat w);
VGLEXTAPI void VGLEXTAPIENTRY glVertex4fv (const GLfloat *v);
VGLEXTAPI void VGLEXTAPIENTRY glVertex4i (GLint x, GLint y, GLint z, GLint w);
VGLEXTAPI void VGLEXTAPIENTRY glVertex4iv (const GLint *v);
VGLEXTAPI void VGLEXTAPIENTRY glVertex4s (GLshort x, GLshort y, GLshort z, GLshort w);
VGLEXTAPI void VGLEXTAPIENTRY glVertex4sv (const GLshort *v);
VGLEXTAPI void VGLEXTAPIENTRY glVertexPointer (GLint size, GLenum type, GLsizei stride, const GLvoid *pointer);
VGLEXTAPI void VGLEXTAPIENTRY glViewport (GLint x, GLint y, GLsizei width, GLsizei height);

#ifdef	__cplusplus
}
#endif /* __cplusplus */

#else	/* !DLL_TARGET_DCC */

/*
 * Pointers
 */
VGLEXTAPI void (VGLEXTAPIENTRY *glAccum) (GLenum op, GLfloat value);
VGLEXTAPI void (VGLEXTAPIENTRY *glAlphaFunc) (GLenum func, GLclampf ref);
VGLEXTAPI GLboolean (VGLEXTAPIENTRY *glAreTexturesResident) (GLsizei n, const GLuint *textures, GLboolean *residences);
VGLEXTAPI void (VGLEXTAPIENTRY *glArrayElement) (GLint i);
VGLEXTAPI void (VGLEXTAPIENTRY *glBegin) (GLenum mode);
VGLEXTAPI void (VGLEXTAPIENTRY *glBindTexture) (GLenum target, GLuint texture);
VGLEXTAPI void (VGLEXTAPIENTRY *glBitmap) (GLsizei width, GLsizei height, GLfloat xorig, GLfloat yorig, GLfloat xmove, GLfloat ymove, const GLubyte *bitmap);
VGLEXTAPI void (VGLEXTAPIENTRY *glBlendFunc) (GLenum sfactor, GLenum dfactor);
VGLEXTAPI void (VGLEXTAPIENTRY *glCallList) (GLuint list);
VGLEXTAPI void (VGLEXTAPIENTRY *glCallLists) (GLsizei n, GLenum type, const GLvoid *lists);
VGLEXTAPI void (VGLEXTAPIENTRY *glClear) (GLbitfield mask);
VGLEXTAPI void (VGLEXTAPIENTRY *glClearAccum) (GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha);
VGLEXTAPI void (VGLEXTAPIENTRY *glClearColor) (GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha);
VGLEXTAPI void (VGLEXTAPIENTRY *glClearDepth) (GLclampd depth);
VGLEXTAPI void (VGLEXTAPIENTRY *glClearIndex) (GLfloat c);
VGLEXTAPI void (VGLEXTAPIENTRY *glClearStencil) (GLint s);
VGLEXTAPI void (VGLEXTAPIENTRY *glClipPlane) (GLenum plane, const GLdouble *equation);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor3b) (GLbyte red, GLbyte green, GLbyte blue);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor3bv) (const GLbyte *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor3d) (GLdouble red, GLdouble green, GLdouble blue);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor3dv) (const GLdouble *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor3f) (GLfloat red, GLfloat green, GLfloat blue);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor3fv) (const GLfloat *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor3i) (GLint red, GLint green, GLint blue);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor3iv) (const GLint *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor3s) (GLshort red, GLshort green, GLshort blue);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor3sv) (const GLshort *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor3ub) (GLubyte red, GLubyte green, GLubyte blue);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor3ubv) (const GLubyte *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor3ui) (GLuint red, GLuint green, GLuint blue);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor3uiv) (const GLuint *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor3us) (GLushort red, GLushort green, GLushort blue);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor3usv) (const GLushort *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor4b) (GLbyte red, GLbyte green, GLbyte blue, GLbyte alpha);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor4bv) (const GLbyte *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor4d) (GLdouble red, GLdouble green, GLdouble blue, GLdouble alpha);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor4dv) (const GLdouble *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor4f) (GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor4fv) (const GLfloat *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor4i) (GLint red, GLint green, GLint blue, GLint alpha);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor4iv) (const GLint *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor4s) (GLshort red, GLshort green, GLshort blue, GLshort alpha);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor4sv) (const GLshort *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor4ub) (GLubyte red, GLubyte green, GLubyte blue, GLubyte alpha);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor4ubv) (const GLubyte *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor4ui) (GLuint red, GLuint green, GLuint blue, GLuint alpha);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor4uiv) (const GLuint *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor4us) (GLushort red, GLushort green, GLushort blue, GLushort alpha);
VGLEXTAPI void (VGLEXTAPIENTRY *glColor4usv) (const GLushort *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glColorMask) (GLboolean red, GLboolean green, GLboolean blue, GLboolean alpha);
VGLEXTAPI void (VGLEXTAPIENTRY *glColorMaterial) (GLenum face, GLenum mode);
VGLEXTAPI void (VGLEXTAPIENTRY *glColorPointer) (GLint size, GLenum type, GLsizei stride, const GLvoid *pointer);
VGLEXTAPI void (VGLEXTAPIENTRY *glCopyPixels) (GLint x, GLint y, GLsizei width, GLsizei height, GLenum type);
VGLEXTAPI void (VGLEXTAPIENTRY *glCopyTexImage1D) (GLenum target, GLint level, GLenum internalformat, GLint x, GLint y, GLsizei width, GLint border);
VGLEXTAPI void (VGLEXTAPIENTRY *glCopyTexImage2D) (GLenum target, GLint level, GLenum internalformat, GLint x, GLint y, GLsizei width, GLsizei height, GLint border);
VGLEXTAPI void (VGLEXTAPIENTRY *glCopyTexSubImage1D) (GLenum target, GLint level, GLint xoffset, GLint x, GLint y, GLsizei width);
VGLEXTAPI void (VGLEXTAPIENTRY *glCopyTexSubImage2D) (GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint x, GLint y, GLsizei width, GLsizei height);
VGLEXTAPI void (VGLEXTAPIENTRY *glCullFace) (GLenum mode);
VGLEXTAPI void (VGLEXTAPIENTRY *glDeleteLists) (GLuint list, GLsizei range);
VGLEXTAPI void (VGLEXTAPIENTRY *glDeleteTextures) (GLsizei n, const GLuint *textures);
VGLEXTAPI void (VGLEXTAPIENTRY *glDepthFunc) (GLenum func);
VGLEXTAPI void (VGLEXTAPIENTRY *glDepthMask) (GLboolean flag);
VGLEXTAPI void (VGLEXTAPIENTRY *glDepthRange) (GLclampd zNear, GLclampd zFar);
VGLEXTAPI void (VGLEXTAPIENTRY *glDisable) (GLenum cap);
VGLEXTAPI void (VGLEXTAPIENTRY *glDisableClientState) (GLenum array);
VGLEXTAPI void (VGLEXTAPIENTRY *glDrawArrays) (GLenum mode, GLint first, GLsizei count);
VGLEXTAPI void (VGLEXTAPIENTRY *glDrawBuffer) (GLenum mode);
VGLEXTAPI void (VGLEXTAPIENTRY *glDrawElements) (GLenum mode, GLsizei count, GLenum type, const GLvoid *indices);
VGLEXTAPI void (VGLEXTAPIENTRY *glDrawPixels) (GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels);
VGLEXTAPI void (VGLEXTAPIENTRY *glEdgeFlag) (GLboolean flag);
VGLEXTAPI void (VGLEXTAPIENTRY *glEdgeFlagPointer) (GLsizei stride, const GLboolean *pointer);
VGLEXTAPI void (VGLEXTAPIENTRY *glEdgeFlagv) (const GLboolean *flag);
VGLEXTAPI void (VGLEXTAPIENTRY *glEnable) (GLenum cap);
VGLEXTAPI void (VGLEXTAPIENTRY *glEnableClientState) (GLenum array);
VGLEXTAPI void (VGLEXTAPIENTRY *glEnd) (void);
VGLEXTAPI void (VGLEXTAPIENTRY *glEndList) (void);
VGLEXTAPI void (VGLEXTAPIENTRY *glEvalCoord1d) (GLdouble u);
VGLEXTAPI void (VGLEXTAPIENTRY *glEvalCoord1dv) (const GLdouble *u);
VGLEXTAPI void (VGLEXTAPIENTRY *glEvalCoord1f) (GLfloat u);
VGLEXTAPI void (VGLEXTAPIENTRY *glEvalCoord1fv) (const GLfloat *u);
VGLEXTAPI void (VGLEXTAPIENTRY *glEvalCoord2d) (GLdouble u, GLdouble v);
VGLEXTAPI void (VGLEXTAPIENTRY *glEvalCoord2dv) (const GLdouble *u);
VGLEXTAPI void (VGLEXTAPIENTRY *glEvalCoord2f) (GLfloat u, GLfloat v);
VGLEXTAPI void (VGLEXTAPIENTRY *glEvalCoord2fv) (const GLfloat *u);
VGLEXTAPI void (VGLEXTAPIENTRY *glEvalMesh1) (GLenum mode, GLint i1, GLint i2);
VGLEXTAPI void (VGLEXTAPIENTRY *glEvalMesh2) (GLenum mode, GLint i1, GLint i2, GLint j1, GLint j2);
VGLEXTAPI void (VGLEXTAPIENTRY *glEvalPoint1) (GLint i);
VGLEXTAPI void (VGLEXTAPIENTRY *glEvalPoint2) (GLint i, GLint j);
VGLEXTAPI void (VGLEXTAPIENTRY *glFeedbackBuffer) (GLsizei size, GLenum type, GLfloat *buffer);
VGLEXTAPI void (VGLEXTAPIENTRY *glFinish) (void);
VGLEXTAPI void (VGLEXTAPIENTRY *glFlush) (void);
VGLEXTAPI void (VGLEXTAPIENTRY *glFogf) (GLenum pname, GLfloat param);
VGLEXTAPI void (VGLEXTAPIENTRY *glFogfv) (GLenum pname, const GLfloat *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glFogi) (GLenum pname, GLint param);
VGLEXTAPI void (VGLEXTAPIENTRY *glFogiv) (GLenum pname, const GLint *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glFrontFace) (GLenum mode);
VGLEXTAPI void (VGLEXTAPIENTRY *glFrustum) (GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble zNear, GLdouble zFar);
VGLEXTAPI GLuint (VGLEXTAPIENTRY *glGenLists) (GLsizei range);
VGLEXTAPI void (VGLEXTAPIENTRY *glGenTextures) (GLsizei n, GLuint *textures);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetBooleanv) (GLenum pname, GLboolean *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetClipPlane) (GLenum plane, GLdouble *equation);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetDoublev) (GLenum pname, GLdouble *params);
VGLEXTAPI GLenum (VGLEXTAPIENTRY *glGetError) (void);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetFloatv) (GLenum pname, GLfloat *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetIntegerv) (GLenum pname, GLint *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetLightfv) (GLenum light, GLenum pname, GLfloat *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetLightiv) (GLenum light, GLenum pname, GLint *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetMapdv) (GLenum target, GLenum query, GLdouble *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetMapfv) (GLenum target, GLenum query, GLfloat *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetMapiv) (GLenum target, GLenum query, GLint *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetMaterialfv) (GLenum face, GLenum pname, GLfloat *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetMaterialiv) (GLenum face, GLenum pname, GLint *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetPixelMapfv) (GLenum map, GLfloat *values);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetPixelMapuiv) (GLenum map, GLuint *values);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetPixelMapusv) (GLenum map, GLushort *values);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetPointerv) (GLenum pname, GLvoid* *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetPolygonStipple) (GLubyte *mask);
VGLEXTAPI const GLubyte * (VGLEXTAPIENTRY *glGetString) (GLenum name);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetTexEnvfv) (GLenum target, GLenum pname, GLfloat *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetTexEnviv) (GLenum target, GLenum pname, GLint *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetTexGendv) (GLenum coord, GLenum pname, GLdouble *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetTexGenfv) (GLenum coord, GLenum pname, GLfloat *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetTexGeniv) (GLenum coord, GLenum pname, GLint *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetTexImage) (GLenum target, GLint level, GLenum format, GLenum type, GLvoid *pixels);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetTexLevelParameterfv) (GLenum target, GLint level, GLenum pname, GLfloat *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetTexLevelParameteriv) (GLenum target, GLint level, GLenum pname, GLint *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetTexParameterfv) (GLenum target, GLenum pname, GLfloat *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetTexParameteriv) (GLenum target, GLenum pname, GLint *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glHint) (GLenum target, GLenum mode);
VGLEXTAPI void (VGLEXTAPIENTRY *glIndexMask) (GLuint mask);
VGLEXTAPI void (VGLEXTAPIENTRY *glIndexPointer) (GLenum type, GLsizei stride, const GLvoid *pointer);
VGLEXTAPI void (VGLEXTAPIENTRY *glIndexd) (GLdouble c);
VGLEXTAPI void (VGLEXTAPIENTRY *glIndexdv) (const GLdouble *c);
VGLEXTAPI void (VGLEXTAPIENTRY *glIndexf) (GLfloat c);
VGLEXTAPI void (VGLEXTAPIENTRY *glIndexfv) (const GLfloat *c);
VGLEXTAPI void (VGLEXTAPIENTRY *glIndexi) (GLint c);
VGLEXTAPI void (VGLEXTAPIENTRY *glIndexiv) (const GLint *c);
VGLEXTAPI void (VGLEXTAPIENTRY *glIndexs) (GLshort c);
VGLEXTAPI void (VGLEXTAPIENTRY *glIndexsv) (const GLshort *c);
VGLEXTAPI void (VGLEXTAPIENTRY *glIndexub) (GLubyte c);
VGLEXTAPI void (VGLEXTAPIENTRY *glIndexubv) (const GLubyte *c);
VGLEXTAPI void (VGLEXTAPIENTRY *glInitNames) (void);
VGLEXTAPI void (VGLEXTAPIENTRY *glInterleavedArrays) (GLenum format, GLsizei stride, const GLvoid *pointer);
VGLEXTAPI GLboolean (VGLEXTAPIENTRY *glIsEnabled) (GLenum cap);
VGLEXTAPI GLboolean (VGLEXTAPIENTRY *glIsList) (GLuint list);
VGLEXTAPI GLboolean (VGLEXTAPIENTRY *glIsTexture) (GLuint texture);
VGLEXTAPI void (VGLEXTAPIENTRY *glLightModelf) (GLenum pname, GLfloat param);
VGLEXTAPI void (VGLEXTAPIENTRY *glLightModelfv) (GLenum pname, const GLfloat *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glLightModeli) (GLenum pname, GLint param);
VGLEXTAPI void (VGLEXTAPIENTRY *glLightModeliv) (GLenum pname, const GLint *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glLightf) (GLenum light, GLenum pname, GLfloat param);
VGLEXTAPI void (VGLEXTAPIENTRY *glLightfv) (GLenum light, GLenum pname, const GLfloat *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glLighti) (GLenum light, GLenum pname, GLint param);
VGLEXTAPI void (VGLEXTAPIENTRY *glLightiv) (GLenum light, GLenum pname, const GLint *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glLineStipple) (GLint factor, GLushort pattern);
VGLEXTAPI void (VGLEXTAPIENTRY *glLineWidth) (GLfloat width);
VGLEXTAPI void (VGLEXTAPIENTRY *glListBase) (GLuint base);
VGLEXTAPI void (VGLEXTAPIENTRY *glLoadIdentity) (void);
VGLEXTAPI void (VGLEXTAPIENTRY *glLoadMatrixd) (const GLdouble *m);
VGLEXTAPI void (VGLEXTAPIENTRY *glLoadMatrixf) (const GLfloat *m);
VGLEXTAPI void (VGLEXTAPIENTRY *glLoadName) (GLuint name);
VGLEXTAPI void (VGLEXTAPIENTRY *glLogicOp) (GLenum opcode);
VGLEXTAPI void (VGLEXTAPIENTRY *glMap1d) (GLenum target, GLdouble u1, GLdouble u2, GLint stride, GLint order, const GLdouble *points);
VGLEXTAPI void (VGLEXTAPIENTRY *glMap1f) (GLenum target, GLfloat u1, GLfloat u2, GLint stride, GLint order, const GLfloat *points);
VGLEXTAPI void (VGLEXTAPIENTRY *glMap2d) (GLenum target, GLdouble u1, GLdouble u2, GLint ustride, GLint uorder, GLdouble v1, GLdouble v2, GLint vstride, GLint vorder, const GLdouble *points);
VGLEXTAPI void (VGLEXTAPIENTRY *glMap2f) (GLenum target, GLfloat u1, GLfloat u2, GLint ustride, GLint uorder, GLfloat v1, GLfloat v2, GLint vstride, GLint vorder, const GLfloat *points);
VGLEXTAPI void (VGLEXTAPIENTRY *glMapGrid1d) (GLint un, GLdouble u1, GLdouble u2);
VGLEXTAPI void (VGLEXTAPIENTRY *glMapGrid1f) (GLint un, GLfloat u1, GLfloat u2);
VGLEXTAPI void (VGLEXTAPIENTRY *glMapGrid2d) (GLint un, GLdouble u1, GLdouble u2, GLint vn, GLdouble v1, GLdouble v2);
VGLEXTAPI void (VGLEXTAPIENTRY *glMapGrid2f) (GLint un, GLfloat u1, GLfloat u2, GLint vn, GLfloat v1, GLfloat v2);
VGLEXTAPI void (VGLEXTAPIENTRY *glMaterialf) (GLenum face, GLenum pname, GLfloat param);
VGLEXTAPI void (VGLEXTAPIENTRY *glMaterialfv) (GLenum face, GLenum pname, const GLfloat *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glMateriali) (GLenum face, GLenum pname, GLint param);
VGLEXTAPI void (VGLEXTAPIENTRY *glMaterialiv) (GLenum face, GLenum pname, const GLint *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glMatrixMode) (GLenum mode);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultMatrixd) (const GLdouble *m);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultMatrixf) (const GLfloat *m);
VGLEXTAPI void (VGLEXTAPIENTRY *glNewList) (GLuint list, GLenum mode);
VGLEXTAPI void (VGLEXTAPIENTRY *glNormal3b) (GLbyte nx, GLbyte ny, GLbyte nz);
VGLEXTAPI void (VGLEXTAPIENTRY *glNormal3bv) (const GLbyte *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glNormal3d) (GLdouble nx, GLdouble ny, GLdouble nz);
VGLEXTAPI void (VGLEXTAPIENTRY *glNormal3dv) (const GLdouble *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glNormal3f) (GLfloat nx, GLfloat ny, GLfloat nz);
VGLEXTAPI void (VGLEXTAPIENTRY *glNormal3fv) (const GLfloat *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glNormal3i) (GLint nx, GLint ny, GLint nz);
VGLEXTAPI void (VGLEXTAPIENTRY *glNormal3iv) (const GLint *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glNormal3s) (GLshort nx, GLshort ny, GLshort nz);
VGLEXTAPI void (VGLEXTAPIENTRY *glNormal3sv) (const GLshort *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glNormalPointer) (GLenum type, GLsizei stride, const GLvoid *pointer);
VGLEXTAPI void (VGLEXTAPIENTRY *glOrtho) (GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble zNear, GLdouble zFar);
VGLEXTAPI void (VGLEXTAPIENTRY *glPassThrough) (GLfloat token);
VGLEXTAPI void (VGLEXTAPIENTRY *glPixelMapfv) (GLenum map, GLint mapsize, const GLfloat *values);
VGLEXTAPI void (VGLEXTAPIENTRY *glPixelMapuiv) (GLenum map, GLint mapsize, const GLuint *values);
VGLEXTAPI void (VGLEXTAPIENTRY *glPixelMapusv) (GLenum map, GLint mapsize, const GLushort *values);
VGLEXTAPI void (VGLEXTAPIENTRY *glPixelStoref) (GLenum pname, GLfloat param);
VGLEXTAPI void (VGLEXTAPIENTRY *glPixelStorei) (GLenum pname, GLint param);
VGLEXTAPI void (VGLEXTAPIENTRY *glPixelTransferf) (GLenum pname, GLfloat param);
VGLEXTAPI void (VGLEXTAPIENTRY *glPixelTransferi) (GLenum pname, GLint param);
VGLEXTAPI void (VGLEXTAPIENTRY *glPixelZoom) (GLfloat xfactor, GLfloat yfactor);
VGLEXTAPI void (VGLEXTAPIENTRY *glPointSize) (GLfloat size);
VGLEXTAPI void (VGLEXTAPIENTRY *glPolygonMode) (GLenum face, GLenum mode);
VGLEXTAPI void (VGLEXTAPIENTRY *glPolygonOffset) (GLfloat factor, GLfloat units);
VGLEXTAPI void (VGLEXTAPIENTRY *glPolygonStipple) (const GLubyte *mask);
VGLEXTAPI void (VGLEXTAPIENTRY *glPopAttrib) (void);
VGLEXTAPI void (VGLEXTAPIENTRY *glPopClientAttrib) (void);
VGLEXTAPI void (VGLEXTAPIENTRY *glPopMatrix) (void);
VGLEXTAPI void (VGLEXTAPIENTRY *glPopName) (void);
VGLEXTAPI void (VGLEXTAPIENTRY *glPrioritizeTextures) (GLsizei n, const GLuint *textures, const GLclampf *priorities);
VGLEXTAPI void (VGLEXTAPIENTRY *glPushAttrib) (GLbitfield mask);
VGLEXTAPI void (VGLEXTAPIENTRY *glPushClientAttrib) (GLbitfield mask);
VGLEXTAPI void (VGLEXTAPIENTRY *glPushMatrix) (void);
VGLEXTAPI void (VGLEXTAPIENTRY *glPushName) (GLuint name);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos2d) (GLdouble x, GLdouble y);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos2dv) (const GLdouble *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos2f) (GLfloat x, GLfloat y);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos2fv) (const GLfloat *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos2i) (GLint x, GLint y);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos2iv) (const GLint *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos2s) (GLshort x, GLshort y);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos2sv) (const GLshort *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos3d) (GLdouble x, GLdouble y, GLdouble z);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos3dv) (const GLdouble *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos3f) (GLfloat x, GLfloat y, GLfloat z);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos3fv) (const GLfloat *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos3i) (GLint x, GLint y, GLint z);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos3iv) (const GLint *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos3s) (GLshort x, GLshort y, GLshort z);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos3sv) (const GLshort *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos4d) (GLdouble x, GLdouble y, GLdouble z, GLdouble w);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos4dv) (const GLdouble *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos4f) (GLfloat x, GLfloat y, GLfloat z, GLfloat w);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos4fv) (const GLfloat *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos4i) (GLint x, GLint y, GLint z, GLint w);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos4iv) (const GLint *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos4s) (GLshort x, GLshort y, GLshort z, GLshort w);
VGLEXTAPI void (VGLEXTAPIENTRY *glRasterPos4sv) (const GLshort *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glReadBuffer) (GLenum mode);
VGLEXTAPI void (VGLEXTAPIENTRY *glReadPixels) (GLint x, GLint y, GLsizei width, GLsizei height, GLenum format, GLenum type, GLvoid *pixels);
VGLEXTAPI void (VGLEXTAPIENTRY *glRectd) (GLdouble x1, GLdouble y1, GLdouble x2, GLdouble y2);
VGLEXTAPI void (VGLEXTAPIENTRY *glRectdv) (const GLdouble *v1, const GLdouble *v2);
VGLEXTAPI void (VGLEXTAPIENTRY *glRectf) (GLfloat x1, GLfloat y1, GLfloat x2, GLfloat y2);
VGLEXTAPI void (VGLEXTAPIENTRY *glRectfv) (const GLfloat *v1, const GLfloat *v2);
VGLEXTAPI void (VGLEXTAPIENTRY *glRecti) (GLint x1, GLint y1, GLint x2, GLint y2);
VGLEXTAPI void (VGLEXTAPIENTRY *glRectiv) (const GLint *v1, const GLint *v2);
VGLEXTAPI void (VGLEXTAPIENTRY *glRects) (GLshort x1, GLshort y1, GLshort x2, GLshort y2);
VGLEXTAPI void (VGLEXTAPIENTRY *glRectsv) (const GLshort *v1, const GLshort *v2);
VGLEXTAPI GLint (VGLEXTAPIENTRY *glRenderMode) (GLenum mode);
VGLEXTAPI void (VGLEXTAPIENTRY *glRotated) (GLdouble angle, GLdouble x, GLdouble y, GLdouble z);
VGLEXTAPI void (VGLEXTAPIENTRY *glRotatef) (GLfloat angle, GLfloat x, GLfloat y, GLfloat z);
VGLEXTAPI void (VGLEXTAPIENTRY *glScaled) (GLdouble x, GLdouble y, GLdouble z);
VGLEXTAPI void (VGLEXTAPIENTRY *glScalef) (GLfloat x, GLfloat y, GLfloat z);
VGLEXTAPI void (VGLEXTAPIENTRY *glScissor) (GLint x, GLint y, GLsizei width, GLsizei height);
VGLEXTAPI void (VGLEXTAPIENTRY *glSelectBuffer) (GLsizei size, GLuint *buffer);
VGLEXTAPI void (VGLEXTAPIENTRY *glShadeModel) (GLenum mode);
VGLEXTAPI void (VGLEXTAPIENTRY *glStencilFunc) (GLenum func, GLint ref, GLuint mask);
VGLEXTAPI void (VGLEXTAPIENTRY *glStencilMask) (GLuint mask);
VGLEXTAPI void (VGLEXTAPIENTRY *glStencilOp) (GLenum fail, GLenum zfail, GLenum zpass);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord1d) (GLdouble s);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord1dv) (const GLdouble *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord1f) (GLfloat s);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord1fv) (const GLfloat *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord1i) (GLint s);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord1iv) (const GLint *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord1s) (GLshort s);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord1sv) (const GLshort *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord2d) (GLdouble s, GLdouble t);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord2dv) (const GLdouble *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord2f) (GLfloat s, GLfloat t);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord2fv) (const GLfloat *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord2i) (GLint s, GLint t);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord2iv) (const GLint *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord2s) (GLshort s, GLshort t);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord2sv) (const GLshort *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord3d) (GLdouble s, GLdouble t, GLdouble r);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord3dv) (const GLdouble *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord3f) (GLfloat s, GLfloat t, GLfloat r);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord3fv) (const GLfloat *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord3i) (GLint s, GLint t, GLint r);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord3iv) (const GLint *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord3s) (GLshort s, GLshort t, GLshort r);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord3sv) (const GLshort *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord4d) (GLdouble s, GLdouble t, GLdouble r, GLdouble q);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord4dv) (const GLdouble *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord4f) (GLfloat s, GLfloat t, GLfloat r, GLfloat q);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord4fv) (const GLfloat *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord4i) (GLint s, GLint t, GLint r, GLint q);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord4iv) (const GLint *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord4s) (GLshort s, GLshort t, GLshort r, GLshort q);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoord4sv) (const GLshort *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexCoordPointer) (GLint size, GLenum type, GLsizei stride, const GLvoid *pointer);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexEnvf) (GLenum target, GLenum pname, GLfloat param);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexEnvfv) (GLenum target, GLenum pname, const GLfloat *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexEnvi) (GLenum target, GLenum pname, GLint param);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexEnviv) (GLenum target, GLenum pname, const GLint *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexGend) (GLenum coord, GLenum pname, GLdouble param);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexGendv) (GLenum coord, GLenum pname, const GLdouble *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexGenf) (GLenum coord, GLenum pname, GLfloat param);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexGenfv) (GLenum coord, GLenum pname, const GLfloat *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexGeni) (GLenum coord, GLenum pname, GLint param);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexGeniv) (GLenum coord, GLenum pname, const GLint *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexImage1D) (GLenum target, GLint level, GLint components, GLsizei width, GLint border, GLenum format, GLenum type, const GLvoid *pixels);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexImage2D) (GLenum target, GLint level, GLint components, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, const GLvoid *pixels);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexParameterf) (GLenum target, GLenum pname, GLfloat param);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexParameterfv) (GLenum target, GLenum pname, const GLfloat *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexParameteri) (GLenum target, GLenum pname, GLint param);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexParameteriv) (GLenum target, GLenum pname, const GLint *params);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexSubImage1D) (GLenum target, GLint level, GLint xoffset, GLsizei width, GLenum format, GLenum type, const GLvoid *pixels);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexSubImage2D) (GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels);
VGLEXTAPI void (VGLEXTAPIENTRY *glTranslated) (GLdouble x, GLdouble y, GLdouble z);
VGLEXTAPI void (VGLEXTAPIENTRY *glTranslatef) (GLfloat x, GLfloat y, GLfloat z);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex2d) (GLdouble x, GLdouble y);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex2dv) (const GLdouble *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex2f) (GLfloat x, GLfloat y);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex2fv) (const GLfloat *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex2i) (GLint x, GLint y);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex2iv) (const GLint *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex2s) (GLshort x, GLshort y);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex2sv) (const GLshort *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex3d) (GLdouble x, GLdouble y, GLdouble z);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex3dv) (const GLdouble *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex3f) (GLfloat x, GLfloat y, GLfloat z);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex3fv) (const GLfloat *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex3i) (GLint x, GLint y, GLint z);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex3iv) (const GLint *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex3s) (GLshort x, GLshort y, GLshort z);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex3sv) (const GLshort *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex4d) (GLdouble x, GLdouble y, GLdouble z, GLdouble w);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex4dv) (const GLdouble *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex4f) (GLfloat x, GLfloat y, GLfloat z, GLfloat w);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex4fv) (const GLfloat *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex4i) (GLint x, GLint y, GLint z, GLint w);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex4iv) (const GLint *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex4s) (GLshort x, GLshort y, GLshort z, GLshort w);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertex4sv) (const GLshort *v);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexPointer) (GLint size, GLenum type, GLsizei stride, const GLvoid *pointer);
VGLEXTAPI void (VGLEXTAPIENTRY *glViewport) (GLint x, GLint y, GLsizei width, GLsizei height);

#endif	/* !DLL_TARGET_DCC */

#else	/* !BASEGL11_DYNAMIC */

/*
 * Static link: OpenGL 1.1 functions.
 */
#ifdef __cplusplus
extern "C" {
#endif

GLAPI void GLAPIENTRY glAccum (GLenum op, GLfloat value);
GLAPI void GLAPIENTRY glAlphaFunc (GLenum func, GLclampf ref);
GLAPI GLboolean GLAPIENTRY glAreTexturesResident (GLsizei n, const GLuint *textures, GLboolean *residences);
GLAPI void GLAPIENTRY glArrayElement (GLint i);
GLAPI void GLAPIENTRY glBegin (GLenum mode);
GLAPI void GLAPIENTRY glBindTexture (GLenum target, GLuint texture);
GLAPI void GLAPIENTRY glBitmap (GLsizei width, GLsizei height, GLfloat xorig, GLfloat yorig, GLfloat xmove, GLfloat ymove, const GLubyte *bitmap);
GLAPI void GLAPIENTRY glBlendFunc (GLenum sfactor, GLenum dfactor);
GLAPI void GLAPIENTRY glCallList (GLuint list);
GLAPI void GLAPIENTRY glCallLists (GLsizei n, GLenum type, const GLvoid *lists);
GLAPI void GLAPIENTRY glClear (GLbitfield mask);
GLAPI void GLAPIENTRY glClearAccum (GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha);
GLAPI void GLAPIENTRY glClearColor (GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha);
GLAPI void GLAPIENTRY glClearDepth (GLclampd depth);
GLAPI void GLAPIENTRY glClearIndex (GLfloat c);
GLAPI void GLAPIENTRY glClearStencil (GLint s);
GLAPI void GLAPIENTRY glClipPlane (GLenum plane, const GLdouble *equation);
GLAPI void GLAPIENTRY glColor3b (GLbyte red, GLbyte green, GLbyte blue);
GLAPI void GLAPIENTRY glColor3bv (const GLbyte *v);
GLAPI void GLAPIENTRY glColor3d (GLdouble red, GLdouble green, GLdouble blue);
GLAPI void GLAPIENTRY glColor3dv (const GLdouble *v);
GLAPI void GLAPIENTRY glColor3f (GLfloat red, GLfloat green, GLfloat blue);
GLAPI void GLAPIENTRY glColor3fv (const GLfloat *v);
GLAPI void GLAPIENTRY glColor3i (GLint red, GLint green, GLint blue);
GLAPI void GLAPIENTRY glColor3iv (const GLint *v);
GLAPI void GLAPIENTRY glColor3s (GLshort red, GLshort green, GLshort blue);
GLAPI void GLAPIENTRY glColor3sv (const GLshort *v);
GLAPI void GLAPIENTRY glColor3ub (GLubyte red, GLubyte green, GLubyte blue);
GLAPI void GLAPIENTRY glColor3ubv (const GLubyte *v);
GLAPI void GLAPIENTRY glColor3ui (GLuint red, GLuint green, GLuint blue);
GLAPI void GLAPIENTRY glColor3uiv (const GLuint *v);
GLAPI void GLAPIENTRY glColor3us (GLushort red, GLushort green, GLushort blue);
GLAPI void GLAPIENTRY glColor3usv (const GLushort *v);
GLAPI void GLAPIENTRY glColor4b (GLbyte red, GLbyte green, GLbyte blue, GLbyte alpha);
GLAPI void GLAPIENTRY glColor4bv (const GLbyte *v);
GLAPI void GLAPIENTRY glColor4d (GLdouble red, GLdouble green, GLdouble blue, GLdouble alpha);
GLAPI void GLAPIENTRY glColor4dv (const GLdouble *v);
GLAPI void GLAPIENTRY glColor4f (GLfloat red, GLfloat green, GLfloat blue, GLfloat alpha);
GLAPI void GLAPIENTRY glColor4fv (const GLfloat *v);
GLAPI void GLAPIENTRY glColor4i (GLint red, GLint green, GLint blue, GLint alpha);
GLAPI void GLAPIENTRY glColor4iv (const GLint *v);
GLAPI void GLAPIENTRY glColor4s (GLshort red, GLshort green, GLshort blue, GLshort alpha);
GLAPI void GLAPIENTRY glColor4sv (const GLshort *v);
GLAPI void GLAPIENTRY glColor4ub (GLubyte red, GLubyte green, GLubyte blue, GLubyte alpha);
GLAPI void GLAPIENTRY glColor4ubv (const GLubyte *v);
GLAPI void GLAPIENTRY glColor4ui (GLuint red, GLuint green, GLuint blue, GLuint alpha);
GLAPI void GLAPIENTRY glColor4uiv (const GLuint *v);
GLAPI void GLAPIENTRY glColor4us (GLushort red, GLushort green, GLushort blue, GLushort alpha);
GLAPI void GLAPIENTRY glColor4usv (const GLushort *v);
GLAPI void GLAPIENTRY glColorMask (GLboolean red, GLboolean green, GLboolean blue, GLboolean alpha);
GLAPI void GLAPIENTRY glColorMaterial (GLenum face, GLenum mode);
GLAPI void GLAPIENTRY glColorPointer (GLint size, GLenum type, GLsizei stride, const GLvoid *pointer);
GLAPI void GLAPIENTRY glCopyPixels (GLint x, GLint y, GLsizei width, GLsizei height, GLenum type);
GLAPI void GLAPIENTRY glCopyTexImage1D (GLenum target, GLint level, GLenum internalformat, GLint x, GLint y, GLsizei width, GLint border);
GLAPI void GLAPIENTRY glCopyTexImage2D (GLenum target, GLint level, GLenum internalformat, GLint x, GLint y, GLsizei width, GLsizei height, GLint border);
GLAPI void GLAPIENTRY glCopyTexSubImage1D (GLenum target, GLint level, GLint xoffset, GLint x, GLint y, GLsizei width);
GLAPI void GLAPIENTRY glCopyTexSubImage2D (GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint x, GLint y, GLsizei width, GLsizei height);
GLAPI void GLAPIENTRY glCullFace (GLenum mode);
GLAPI void GLAPIENTRY glDeleteLists (GLuint list, GLsizei range);
GLAPI void GLAPIENTRY glDeleteTextures (GLsizei n, const GLuint *textures);
GLAPI void GLAPIENTRY glDepthFunc (GLenum func);
GLAPI void GLAPIENTRY glDepthMask (GLboolean flag);
GLAPI void GLAPIENTRY glDepthRange (GLclampd zNear, GLclampd zFar);
GLAPI void GLAPIENTRY glDisable (GLenum cap);
GLAPI void GLAPIENTRY glDisableClientState (GLenum array);
GLAPI void GLAPIENTRY glDrawArrays (GLenum mode, GLint first, GLsizei count);
GLAPI void GLAPIENTRY glDrawBuffer (GLenum mode);
GLAPI void GLAPIENTRY glDrawElements (GLenum mode, GLsizei count, GLenum type, const GLvoid *indices);
GLAPI void GLAPIENTRY glDrawPixels (GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels);
GLAPI void GLAPIENTRY glEdgeFlag (GLboolean flag);
GLAPI void GLAPIENTRY glEdgeFlagPointer (GLsizei stride, const GLboolean *pointer);
GLAPI void GLAPIENTRY glEdgeFlagv (const GLboolean *flag);
GLAPI void GLAPIENTRY glEnable (GLenum cap);
GLAPI void GLAPIENTRY glEnableClientState (GLenum array);
GLAPI void GLAPIENTRY glEnd (void);
GLAPI void GLAPIENTRY glEndList (void);
GLAPI void GLAPIENTRY glEvalCoord1d (GLdouble u);
GLAPI void GLAPIENTRY glEvalCoord1dv (const GLdouble *u);
GLAPI void GLAPIENTRY glEvalCoord1f (GLfloat u);
GLAPI void GLAPIENTRY glEvalCoord1fv (const GLfloat *u);
GLAPI void GLAPIENTRY glEvalCoord2d (GLdouble u, GLdouble v);
GLAPI void GLAPIENTRY glEvalCoord2dv (const GLdouble *u);
GLAPI void GLAPIENTRY glEvalCoord2f (GLfloat u, GLfloat v);
GLAPI void GLAPIENTRY glEvalCoord2fv (const GLfloat *u);
GLAPI void GLAPIENTRY glEvalMesh1 (GLenum mode, GLint i1, GLint i2);
GLAPI void GLAPIENTRY glEvalMesh2 (GLenum mode, GLint i1, GLint i2, GLint j1, GLint j2);
GLAPI void GLAPIENTRY glEvalPoint1 (GLint i);
GLAPI void GLAPIENTRY glEvalPoint2 (GLint i, GLint j);
GLAPI void GLAPIENTRY glFeedbackBuffer (GLsizei size, GLenum type, GLfloat *buffer);
GLAPI void GLAPIENTRY glFinish (void);
GLAPI void GLAPIENTRY glFlush (void);
GLAPI void GLAPIENTRY glFogf (GLenum pname, GLfloat param);
GLAPI void GLAPIENTRY glFogfv (GLenum pname, const GLfloat *params);
GLAPI void GLAPIENTRY glFogi (GLenum pname, GLint param);
GLAPI void GLAPIENTRY glFogiv (GLenum pname, const GLint *params);
GLAPI void GLAPIENTRY glFrontFace (GLenum mode);
GLAPI void GLAPIENTRY glFrustum (GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble zNear, GLdouble zFar);
GLAPI GLuint GLAPIENTRY glGenLists (GLsizei range);
GLAPI void GLAPIENTRY glGenTextures (GLsizei n, GLuint *textures);
GLAPI void GLAPIENTRY glGetBooleanv (GLenum pname, GLboolean *params);
GLAPI void GLAPIENTRY glGetClipPlane (GLenum plane, GLdouble *equation);
GLAPI void GLAPIENTRY glGetDoublev (GLenum pname, GLdouble *params);
GLAPI GLenum GLAPIENTRY glGetError (void);
GLAPI void GLAPIENTRY glGetFloatv (GLenum pname, GLfloat *params);
GLAPI void GLAPIENTRY glGetIntegerv (GLenum pname, GLint *params);
GLAPI void GLAPIENTRY glGetLightfv (GLenum light, GLenum pname, GLfloat *params);
GLAPI void GLAPIENTRY glGetLightiv (GLenum light, GLenum pname, GLint *params);
GLAPI void GLAPIENTRY glGetMapdv (GLenum target, GLenum query, GLdouble *v);
GLAPI void GLAPIENTRY glGetMapfv (GLenum target, GLenum query, GLfloat *v);
GLAPI void GLAPIENTRY glGetMapiv (GLenum target, GLenum query, GLint *v);
GLAPI void GLAPIENTRY glGetMaterialfv (GLenum face, GLenum pname, GLfloat *params);
GLAPI void GLAPIENTRY glGetMaterialiv (GLenum face, GLenum pname, GLint *params);
GLAPI void GLAPIENTRY glGetPixelMapfv (GLenum map, GLfloat *values);
GLAPI void GLAPIENTRY glGetPixelMapuiv (GLenum map, GLuint *values);
GLAPI void GLAPIENTRY glGetPixelMapusv (GLenum map, GLushort *values);
GLAPI void GLAPIENTRY glGetPointerv (GLenum pname, GLvoid* *params);
GLAPI void GLAPIENTRY glGetPolygonStipple (GLubyte *mask);
GLAPI const GLubyte * GLAPIENTRY glGetString (GLenum name);
GLAPI void GLAPIENTRY glGetTexEnvfv (GLenum target, GLenum pname, GLfloat *params);
GLAPI void GLAPIENTRY glGetTexEnviv (GLenum target, GLenum pname, GLint *params);
GLAPI void GLAPIENTRY glGetTexGendv (GLenum coord, GLenum pname, GLdouble *params);
GLAPI void GLAPIENTRY glGetTexGenfv (GLenum coord, GLenum pname, GLfloat *params);
GLAPI void GLAPIENTRY glGetTexGeniv (GLenum coord, GLenum pname, GLint *params);
GLAPI void GLAPIENTRY glGetTexImage (GLenum target, GLint level, GLenum format, GLenum type, GLvoid *pixels);
GLAPI void GLAPIENTRY glGetTexLevelParameterfv (GLenum target, GLint level, GLenum pname, GLfloat *params);
GLAPI void GLAPIENTRY glGetTexLevelParameteriv (GLenum target, GLint level, GLenum pname, GLint *params);
GLAPI void GLAPIENTRY glGetTexParameterfv (GLenum target, GLenum pname, GLfloat *params);
GLAPI void GLAPIENTRY glGetTexParameteriv (GLenum target, GLenum pname, GLint *params);
GLAPI void GLAPIENTRY glHint (GLenum target, GLenum mode);
GLAPI void GLAPIENTRY glIndexMask (GLuint mask);
GLAPI void GLAPIENTRY glIndexPointer (GLenum type, GLsizei stride, const GLvoid *pointer);
GLAPI void GLAPIENTRY glIndexd (GLdouble c);
GLAPI void GLAPIENTRY glIndexdv (const GLdouble *c);
GLAPI void GLAPIENTRY glIndexf (GLfloat c);
GLAPI void GLAPIENTRY glIndexfv (const GLfloat *c);
GLAPI void GLAPIENTRY glIndexi (GLint c);
GLAPI void GLAPIENTRY glIndexiv (const GLint *c);
GLAPI void GLAPIENTRY glIndexs (GLshort c);
GLAPI void GLAPIENTRY glIndexsv (const GLshort *c);
GLAPI void GLAPIENTRY glIndexub (GLubyte c);
GLAPI void GLAPIENTRY glIndexubv (const GLubyte *c);
GLAPI void GLAPIENTRY glInitNames (void);
GLAPI void GLAPIENTRY glInterleavedArrays (GLenum format, GLsizei stride, const GLvoid *pointer);
GLAPI GLboolean GLAPIENTRY glIsEnabled (GLenum cap);
GLAPI GLboolean GLAPIENTRY glIsList (GLuint list);
GLAPI GLboolean GLAPIENTRY glIsTexture (GLuint texture);
GLAPI void GLAPIENTRY glLightModelf (GLenum pname, GLfloat param);
GLAPI void GLAPIENTRY glLightModelfv (GLenum pname, const GLfloat *params);
GLAPI void GLAPIENTRY glLightModeli (GLenum pname, GLint param);
GLAPI void GLAPIENTRY glLightModeliv (GLenum pname, const GLint *params);
GLAPI void GLAPIENTRY glLightf (GLenum light, GLenum pname, GLfloat param);
GLAPI void GLAPIENTRY glLightfv (GLenum light, GLenum pname, const GLfloat *params);
GLAPI void GLAPIENTRY glLighti (GLenum light, GLenum pname, GLint param);
GLAPI void GLAPIENTRY glLightiv (GLenum light, GLenum pname, const GLint *params);
GLAPI void GLAPIENTRY glLineStipple (GLint factor, GLushort pattern);
GLAPI void GLAPIENTRY glLineWidth (GLfloat width);
GLAPI void GLAPIENTRY glListBase (GLuint base);
GLAPI void GLAPIENTRY glLoadIdentity (void);
GLAPI void GLAPIENTRY glLoadMatrixd (const GLdouble *m);
GLAPI void GLAPIENTRY glLoadMatrixf (const GLfloat *m);
GLAPI void GLAPIENTRY glLoadName (GLuint name);
GLAPI void GLAPIENTRY glLogicOp (GLenum opcode);
GLAPI void GLAPIENTRY glMap1d (GLenum target, GLdouble u1, GLdouble u2, GLint stride, GLint order, const GLdouble *points);
GLAPI void GLAPIENTRY glMap1f (GLenum target, GLfloat u1, GLfloat u2, GLint stride, GLint order, const GLfloat *points);
GLAPI void GLAPIENTRY glMap2d (GLenum target, GLdouble u1, GLdouble u2, GLint ustride, GLint uorder, GLdouble v1, GLdouble v2, GLint vstride, GLint vorder, const GLdouble *points);
GLAPI void GLAPIENTRY glMap2f (GLenum target, GLfloat u1, GLfloat u2, GLint ustride, GLint uorder, GLfloat v1, GLfloat v2, GLint vstride, GLint vorder, const GLfloat *points);
GLAPI void GLAPIENTRY glMapGrid1d (GLint un, GLdouble u1, GLdouble u2);
GLAPI void GLAPIENTRY glMapGrid1f (GLint un, GLfloat u1, GLfloat u2);
GLAPI void GLAPIENTRY glMapGrid2d (GLint un, GLdouble u1, GLdouble u2, GLint vn, GLdouble v1, GLdouble v2);
GLAPI void GLAPIENTRY glMapGrid2f (GLint un, GLfloat u1, GLfloat u2, GLint vn, GLfloat v1, GLfloat v2);
GLAPI void GLAPIENTRY glMaterialf (GLenum face, GLenum pname, GLfloat param);
GLAPI void GLAPIENTRY glMaterialfv (GLenum face, GLenum pname, const GLfloat *params);
GLAPI void GLAPIENTRY glMateriali (GLenum face, GLenum pname, GLint param);
GLAPI void GLAPIENTRY glMaterialiv (GLenum face, GLenum pname, const GLint *params);
GLAPI void GLAPIENTRY glMatrixMode (GLenum mode);
GLAPI void GLAPIENTRY glMultMatrixd (const GLdouble *m);
GLAPI void GLAPIENTRY glMultMatrixf (const GLfloat *m);
GLAPI void GLAPIENTRY glNewList (GLuint list, GLenum mode);
GLAPI void GLAPIENTRY glNormal3b (GLbyte nx, GLbyte ny, GLbyte nz);
GLAPI void GLAPIENTRY glNormal3bv (const GLbyte *v);
GLAPI void GLAPIENTRY glNormal3d (GLdouble nx, GLdouble ny, GLdouble nz);
GLAPI void GLAPIENTRY glNormal3dv (const GLdouble *v);
GLAPI void GLAPIENTRY glNormal3f (GLfloat nx, GLfloat ny, GLfloat nz);
GLAPI void GLAPIENTRY glNormal3fv (const GLfloat *v);
GLAPI void GLAPIENTRY glNormal3i (GLint nx, GLint ny, GLint nz);
GLAPI void GLAPIENTRY glNormal3iv (const GLint *v);
GLAPI void GLAPIENTRY glNormal3s (GLshort nx, GLshort ny, GLshort nz);
GLAPI void GLAPIENTRY glNormal3sv (const GLshort *v);
GLAPI void GLAPIENTRY glNormalPointer (GLenum type, GLsizei stride, const GLvoid *pointer);
GLAPI void GLAPIENTRY glOrtho (GLdouble left, GLdouble right, GLdouble bottom, GLdouble top, GLdouble zNear, GLdouble zFar);
GLAPI void GLAPIENTRY glPassThrough (GLfloat token);
GLAPI void GLAPIENTRY glPixelMapfv (GLenum map, GLint mapsize, const GLfloat *values);
GLAPI void GLAPIENTRY glPixelMapuiv (GLenum map, GLint mapsize, const GLuint *values);
GLAPI void GLAPIENTRY glPixelMapusv (GLenum map, GLint mapsize, const GLushort *values);
GLAPI void GLAPIENTRY glPixelStoref (GLenum pname, GLfloat param);
GLAPI void GLAPIENTRY glPixelStorei (GLenum pname, GLint param);
GLAPI void GLAPIENTRY glPixelTransferf (GLenum pname, GLfloat param);
GLAPI void GLAPIENTRY glPixelTransferi (GLenum pname, GLint param);
GLAPI void GLAPIENTRY glPixelZoom (GLfloat xfactor, GLfloat yfactor);
GLAPI void GLAPIENTRY glPointSize (GLfloat size);
GLAPI void GLAPIENTRY glPolygonMode (GLenum face, GLenum mode);
GLAPI void GLAPIENTRY glPolygonOffset (GLfloat factor, GLfloat units);
GLAPI void GLAPIENTRY glPolygonStipple (const GLubyte *mask);
GLAPI void GLAPIENTRY glPopAttrib (void);
GLAPI void GLAPIENTRY glPopClientAttrib (void);
GLAPI void GLAPIENTRY glPopMatrix (void);
GLAPI void GLAPIENTRY glPopName (void);
GLAPI void GLAPIENTRY glPrioritizeTextures (GLsizei n, const GLuint *textures, const GLclampf *priorities);
GLAPI void GLAPIENTRY glPushAttrib (GLbitfield mask);
GLAPI void GLAPIENTRY glPushClientAttrib (GLbitfield mask);
GLAPI void GLAPIENTRY glPushMatrix (void);
GLAPI void GLAPIENTRY glPushName (GLuint name);
GLAPI void GLAPIENTRY glRasterPos2d (GLdouble x, GLdouble y);
GLAPI void GLAPIENTRY glRasterPos2dv (const GLdouble *v);
GLAPI void GLAPIENTRY glRasterPos2f (GLfloat x, GLfloat y);
GLAPI void GLAPIENTRY glRasterPos2fv (const GLfloat *v);
GLAPI void GLAPIENTRY glRasterPos2i (GLint x, GLint y);
GLAPI void GLAPIENTRY glRasterPos2iv (const GLint *v);
GLAPI void GLAPIENTRY glRasterPos2s (GLshort x, GLshort y);
GLAPI void GLAPIENTRY glRasterPos2sv (const GLshort *v);
GLAPI void GLAPIENTRY glRasterPos3d (GLdouble x, GLdouble y, GLdouble z);
GLAPI void GLAPIENTRY glRasterPos3dv (const GLdouble *v);
GLAPI void GLAPIENTRY glRasterPos3f (GLfloat x, GLfloat y, GLfloat z);
GLAPI void GLAPIENTRY glRasterPos3fv (const GLfloat *v);
GLAPI void GLAPIENTRY glRasterPos3i (GLint x, GLint y, GLint z);
GLAPI void GLAPIENTRY glRasterPos3iv (const GLint *v);
GLAPI void GLAPIENTRY glRasterPos3s (GLshort x, GLshort y, GLshort z);
GLAPI void GLAPIENTRY glRasterPos3sv (const GLshort *v);
GLAPI void GLAPIENTRY glRasterPos4d (GLdouble x, GLdouble y, GLdouble z, GLdouble w);
GLAPI void GLAPIENTRY glRasterPos4dv (const GLdouble *v);
GLAPI void GLAPIENTRY glRasterPos4f (GLfloat x, GLfloat y, GLfloat z, GLfloat w);
GLAPI void GLAPIENTRY glRasterPos4fv (const GLfloat *v);
GLAPI void GLAPIENTRY glRasterPos4i (GLint x, GLint y, GLint z, GLint w);
GLAPI void GLAPIENTRY glRasterPos4iv (const GLint *v);
GLAPI void GLAPIENTRY glRasterPos4s (GLshort x, GLshort y, GLshort z, GLshort w);
GLAPI void GLAPIENTRY glRasterPos4sv (const GLshort *v);
GLAPI void GLAPIENTRY glReadBuffer (GLenum mode);
GLAPI void GLAPIENTRY glReadPixels (GLint x, GLint y, GLsizei width, GLsizei height, GLenum format, GLenum type, GLvoid *pixels);
GLAPI void GLAPIENTRY glRectd (GLdouble x1, GLdouble y1, GLdouble x2, GLdouble y2);
GLAPI void GLAPIENTRY glRectdv (const GLdouble *v1, const GLdouble *v2);
GLAPI void GLAPIENTRY glRectf (GLfloat x1, GLfloat y1, GLfloat x2, GLfloat y2);
GLAPI void GLAPIENTRY glRectfv (const GLfloat *v1, const GLfloat *v2);
GLAPI void GLAPIENTRY glRecti (GLint x1, GLint y1, GLint x2, GLint y2);
GLAPI void GLAPIENTRY glRectiv (const GLint *v1, const GLint *v2);
GLAPI void GLAPIENTRY glRects (GLshort x1, GLshort y1, GLshort x2, GLshort y2);
GLAPI void GLAPIENTRY glRectsv (const GLshort *v1, const GLshort *v2);
GLAPI GLint GLAPIENTRY glRenderMode (GLenum mode);
GLAPI void GLAPIENTRY glRotated (GLdouble angle, GLdouble x, GLdouble y, GLdouble z);
GLAPI void GLAPIENTRY glRotatef (GLfloat angle, GLfloat x, GLfloat y, GLfloat z);
GLAPI void GLAPIENTRY glScaled (GLdouble x, GLdouble y, GLdouble z);
GLAPI void GLAPIENTRY glScalef (GLfloat x, GLfloat y, GLfloat z);
GLAPI void GLAPIENTRY glScissor (GLint x, GLint y, GLsizei width, GLsizei height);
GLAPI void GLAPIENTRY glSelectBuffer (GLsizei size, GLuint *buffer);
GLAPI void GLAPIENTRY glShadeModel (GLenum mode);
GLAPI void GLAPIENTRY glStencilFunc (GLenum func, GLint ref, GLuint mask);
GLAPI void GLAPIENTRY glStencilMask (GLuint mask);
GLAPI void GLAPIENTRY glStencilOp (GLenum fail, GLenum zfail, GLenum zpass);
GLAPI void GLAPIENTRY glTexCoord1d (GLdouble s);
GLAPI void GLAPIENTRY glTexCoord1dv (const GLdouble *v);
GLAPI void GLAPIENTRY glTexCoord1f (GLfloat s);
GLAPI void GLAPIENTRY glTexCoord1fv (const GLfloat *v);
GLAPI void GLAPIENTRY glTexCoord1i (GLint s);
GLAPI void GLAPIENTRY glTexCoord1iv (const GLint *v);
GLAPI void GLAPIENTRY glTexCoord1s (GLshort s);
GLAPI void GLAPIENTRY glTexCoord1sv (const GLshort *v);
GLAPI void GLAPIENTRY glTexCoord2d (GLdouble s, GLdouble t);
GLAPI void GLAPIENTRY glTexCoord2dv (const GLdouble *v);
GLAPI void GLAPIENTRY glTexCoord2f (GLfloat s, GLfloat t);
GLAPI void GLAPIENTRY glTexCoord2fv (const GLfloat *v);
GLAPI void GLAPIENTRY glTexCoord2i (GLint s, GLint t);
GLAPI void GLAPIENTRY glTexCoord2iv (const GLint *v);
GLAPI void GLAPIENTRY glTexCoord2s (GLshort s, GLshort t);
GLAPI void GLAPIENTRY glTexCoord2sv (const GLshort *v);
GLAPI void GLAPIENTRY glTexCoord3d (GLdouble s, GLdouble t, GLdouble r);
GLAPI void GLAPIENTRY glTexCoord3dv (const GLdouble *v);
GLAPI void GLAPIENTRY glTexCoord3f (GLfloat s, GLfloat t, GLfloat r);
GLAPI void GLAPIENTRY glTexCoord3fv (const GLfloat *v);
GLAPI void GLAPIENTRY glTexCoord3i (GLint s, GLint t, GLint r);
GLAPI void GLAPIENTRY glTexCoord3iv (const GLint *v);
GLAPI void GLAPIENTRY glTexCoord3s (GLshort s, GLshort t, GLshort r);
GLAPI void GLAPIENTRY glTexCoord3sv (const GLshort *v);
GLAPI void GLAPIENTRY glTexCoord4d (GLdouble s, GLdouble t, GLdouble r, GLdouble q);
GLAPI void GLAPIENTRY glTexCoord4dv (const GLdouble *v);
GLAPI void GLAPIENTRY glTexCoord4f (GLfloat s, GLfloat t, GLfloat r, GLfloat q);
GLAPI void GLAPIENTRY glTexCoord4fv (const GLfloat *v);
GLAPI void GLAPIENTRY glTexCoord4i (GLint s, GLint t, GLint r, GLint q);
GLAPI void GLAPIENTRY glTexCoord4iv (const GLint *v);
GLAPI void GLAPIENTRY glTexCoord4s (GLshort s, GLshort t, GLshort r, GLshort q);
GLAPI void GLAPIENTRY glTexCoord4sv (const GLshort *v);
GLAPI void GLAPIENTRY glTexCoordPointer (GLint size, GLenum type, GLsizei stride, const GLvoid *pointer);
GLAPI void GLAPIENTRY glTexEnvf (GLenum target, GLenum pname, GLfloat param);
GLAPI void GLAPIENTRY glTexEnvfv (GLenum target, GLenum pname, const GLfloat *params);
GLAPI void GLAPIENTRY glTexEnvi (GLenum target, GLenum pname, GLint param);
GLAPI void GLAPIENTRY glTexEnviv (GLenum target, GLenum pname, const GLint *params);
GLAPI void GLAPIENTRY glTexGend (GLenum coord, GLenum pname, GLdouble param);
GLAPI void GLAPIENTRY glTexGendv (GLenum coord, GLenum pname, const GLdouble *params);
GLAPI void GLAPIENTRY glTexGenf (GLenum coord, GLenum pname, GLfloat param);
GLAPI void GLAPIENTRY glTexGenfv (GLenum coord, GLenum pname, const GLfloat *params);
GLAPI void GLAPIENTRY glTexGeni (GLenum coord, GLenum pname, GLint param);
GLAPI void GLAPIENTRY glTexGeniv (GLenum coord, GLenum pname, const GLint *params);
GLAPI void GLAPIENTRY glTexImage1D (GLenum target, GLint level, GLint components, GLsizei width, GLint border, GLenum format, GLenum type, const GLvoid *pixels);
GLAPI void GLAPIENTRY glTexImage2D (GLenum target, GLint level, GLint components, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, const GLvoid *pixels);
GLAPI void GLAPIENTRY glTexParameterf (GLenum target, GLenum pname, GLfloat param);
GLAPI void GLAPIENTRY glTexParameterfv (GLenum target, GLenum pname, const GLfloat *params);
GLAPI void GLAPIENTRY glTexParameteri (GLenum target, GLenum pname, GLint param);
GLAPI void GLAPIENTRY glTexParameteriv (GLenum target, GLenum pname, const GLint *params);
GLAPI void GLAPIENTRY glTexSubImage1D (GLenum target, GLint level, GLint xoffset, GLsizei width, GLenum format, GLenum type, const GLvoid *pixels);
GLAPI void GLAPIENTRY glTexSubImage2D (GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels);
GLAPI void GLAPIENTRY glTranslated (GLdouble x, GLdouble y, GLdouble z);
GLAPI void GLAPIENTRY glTranslatef (GLfloat x, GLfloat y, GLfloat z);
GLAPI void GLAPIENTRY glVertex2d (GLdouble x, GLdouble y);
GLAPI void GLAPIENTRY glVertex2dv (const GLdouble *v);
GLAPI void GLAPIENTRY glVertex2f (GLfloat x, GLfloat y);
GLAPI void GLAPIENTRY glVertex2fv (const GLfloat *v);
GLAPI void GLAPIENTRY glVertex2i (GLint x, GLint y);
GLAPI void GLAPIENTRY glVertex2iv (const GLint *v);
GLAPI void GLAPIENTRY glVertex2s (GLshort x, GLshort y);
GLAPI void GLAPIENTRY glVertex2sv (const GLshort *v);
GLAPI void GLAPIENTRY glVertex3d (GLdouble x, GLdouble y, GLdouble z);
GLAPI void GLAPIENTRY glVertex3dv (const GLdouble *v);
GLAPI void GLAPIENTRY glVertex3f (GLfloat x, GLfloat y, GLfloat z);
GLAPI void GLAPIENTRY glVertex3fv (const GLfloat *v);
GLAPI void GLAPIENTRY glVertex3i (GLint x, GLint y, GLint z);
GLAPI void GLAPIENTRY glVertex3iv (const GLint *v);
GLAPI void GLAPIENTRY glVertex3s (GLshort x, GLshort y, GLshort z);
GLAPI void GLAPIENTRY glVertex3sv (const GLshort *v);
GLAPI void GLAPIENTRY glVertex4d (GLdouble x, GLdouble y, GLdouble z, GLdouble w);
GLAPI void GLAPIENTRY glVertex4dv (const GLdouble *v);
GLAPI void GLAPIENTRY glVertex4f (GLfloat x, GLfloat y, GLfloat z, GLfloat w);
GLAPI void GLAPIENTRY glVertex4fv (const GLfloat *v);
GLAPI void GLAPIENTRY glVertex4i (GLint x, GLint y, GLint z, GLint w);
GLAPI void GLAPIENTRY glVertex4iv (const GLint *v);
GLAPI void GLAPIENTRY glVertex4s (GLshort x, GLshort y, GLshort z, GLshort w);
GLAPI void GLAPIENTRY glVertex4sv (const GLshort *v);
GLAPI void GLAPIENTRY glVertexPointer (GLint size, GLenum type, GLsizei stride, const GLvoid *pointer);
GLAPI void GLAPIENTRY glViewport (GLint x, GLint y, GLsizei width, GLsizei height);

#ifdef __cplusplus
}
#endif

#endif	/* !BASE_DYNAMIC */

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * ARB extensions.
 * Independent from GL version (or GL 1.1 is required):
 *  - ARB_multitexture;
 *  - ARB_texture_compression;
 *  - ARB_texture_cube_map;
 *  - ARB_multisample;
 *  - ARB_texture_env_add;
 *  - ARB_texture_env_combine;
 *  - ARB_texture_env_dot3;
 *  - ARB_texture_border_clamp;
 *  - ARB_transpose_matrix;
 *  - ARB_depth_texture;
 *	- ARB_shadow;
 *	- ARB_shadow_ambient;
 *  - ARB_point_parameters;
 *  - ARB_texture_env_crossbar;
 *  - ARB_texture_mirrored_repeat;
 *  - ARB_window_pos;
 *  - ARB_shader_objects;
 *  - ARB_vertex_shader;
 *  - ARB_fragment_shader;
 *  - ARB_shading_language_100;
 *	- ARB_imaging.
 * GL 1.3 is required:
 *  - ARB_vertex_program;
 *  - ARB_fragment_program;
 *  - ARB_fragment_program_shadow.
 * GL 1.4 is required:
 *  - ARB_vertex_buffer_object;
 *  - ARB_occlusion_query;
 *  - ARB_draw_buffers;
 *  - ARB_texture_non_power_of_two;
 *  - ARB_point_sprite;
 *  - ARB_texture_rectangle.
 *
 * Extension to add to vglCreateWindowEx(): !!! PRELIMINARY
 *
 * +---------------------------+-----------------------------------------+
 * |        pname              |                parameters               |
 * +---------------------------+-----------------------------------------+
 * | VGL_MULTISAMPLE                                ???                  |
 * +---------------------------------------------------------------------+
 */

/*
 * Tokens
 */

/* GL_ARB_multitexture */
#define GL_TEXTURE0_ARB                   0x84C0
#define GL_TEXTURE1_ARB                   0x84C1
#define GL_TEXTURE2_ARB                   0x84C2
#define GL_TEXTURE3_ARB                   0x84C3
#define GL_TEXTURE4_ARB                   0x84C4
#define GL_TEXTURE5_ARB                   0x84C5
#define GL_TEXTURE6_ARB                   0x84C6
#define GL_TEXTURE7_ARB                   0x84C7
#define GL_TEXTURE8_ARB                   0x84C8
#define GL_TEXTURE9_ARB                   0x84C9
#define GL_TEXTURE10_ARB                  0x84CA
#define GL_TEXTURE11_ARB                  0x84CB
#define GL_TEXTURE12_ARB                  0x84CC
#define GL_TEXTURE13_ARB                  0x84CD
#define GL_TEXTURE14_ARB                  0x84CE
#define GL_TEXTURE15_ARB                  0x84CF
#define GL_TEXTURE16_ARB                  0x84D0
#define GL_TEXTURE17_ARB                  0x84D1
#define GL_TEXTURE18_ARB                  0x84D2
#define GL_TEXTURE19_ARB                  0x84D3
#define GL_TEXTURE20_ARB                  0x84D4
#define GL_TEXTURE21_ARB                  0x84D5
#define GL_TEXTURE22_ARB                  0x84D6
#define GL_TEXTURE23_ARB                  0x84D7
#define GL_TEXTURE24_ARB                  0x84D8
#define GL_TEXTURE25_ARB                  0x84D9
#define GL_TEXTURE26_ARB                  0x84DA
#define GL_TEXTURE27_ARB                  0x84DB
#define GL_TEXTURE28_ARB                  0x84DC
#define GL_TEXTURE29_ARB                  0x84DD
#define GL_TEXTURE30_ARB                  0x84DE
#define GL_TEXTURE31_ARB                  0x84DF
#define GL_ACTIVE_TEXTURE_ARB             0x84E0
#define GL_CLIENT_ACTIVE_TEXTURE_ARB      0x84E1
#define GL_MAX_TEXTURE_UNITS_ARB          0x84E2
/* ARB_texture_compression */
#define GL_COMPRESSED_ALPHA_ARB           0x84E9
#define GL_COMPRESSED_LUMINANCE_ARB       0x84EA
#define GL_COMPRESSED_LUMINANCE_ALPHA_ARB 0x84EB
#define GL_COMPRESSED_INTENSITY_ARB       0x84EC
#define GL_COMPRESSED_RGB_ARB             0x84ED
#define GL_COMPRESSED_RGBA_ARB            0x84EE
#define GL_TEXTURE_COMPRESSION_HINT_ARB   0x84EF
#define GL_TEXTURE_COMPRESSED_IMAGE_SIZE_ARB 0x86A0
#define GL_TEXTURE_COMPRESSED_ARB         0x86A1
#define GL_NUM_COMPRESSED_TEXTURE_FORMATS_ARB 0x86A2
#define GL_COMPRESSED_TEXTURE_FORMATS_ARB 0x86A3
/* ARB_texture_cube_map */
#define GL_NORMAL_MAP_ARB                 0x8511
#define GL_REFLECTION_MAP_ARB             0x8512
#define GL_TEXTURE_CUBE_MAP_ARB           0x8513
#define GL_TEXTURE_BINDING_CUBE_MAP_ARB   0x8514
#define GL_TEXTURE_CUBE_MAP_POSITIVE_X_ARB 0x8515
#define GL_TEXTURE_CUBE_MAP_NEGATIVE_X_ARB 0x8516
#define GL_TEXTURE_CUBE_MAP_POSITIVE_Y_ARB 0x8517
#define GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_ARB 0x8518
#define GL_TEXTURE_CUBE_MAP_POSITIVE_Z_ARB 0x8519
#define GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_ARB 0x851A
#define GL_PROXY_TEXTURE_CUBE_MAP_ARB     0x851B
#define GL_MAX_CUBE_MAP_TEXTURE_SIZE_ARB  0x851C
/* ARB_multisample */
#define GL_MULTISAMPLE_ARB                0x809D
#define GL_SAMPLE_ALPHA_TO_COVERAGE_ARB   0x809E
#define GL_SAMPLE_ALPHA_TO_ONE_ARB        0x809F
#define GL_SAMPLE_COVERAGE_ARB            0x80A0
#define GL_SAMPLE_BUFFERS_ARB             0x80A8
#define GL_SAMPLES_ARB                    0x80A9
#define GL_SAMPLE_COVERAGE_VALUE_ARB      0x80AA
#define GL_SAMPLE_COVERAGE_INVERT_ARB     0x80AB
#define GL_MULTISAMPLE_BIT_ARB            0x20000000
/* ARB_texture_env_add */
/* ARB_texture_env_combine */
#define GL_COMBINE_ARB                    0x8570
#define GL_COMBINE_RGB_ARB                0x8571
#define GL_COMBINE_ALPHA_ARB              0x8572
#define GL_SOURCE0_RGB_ARB                0x8580
#define GL_SOURCE1_RGB_ARB                0x8581
#define GL_SOURCE2_RGB_ARB                0x8582
#define GL_SOURCE0_ALPHA_ARB              0x8588
#define GL_SOURCE1_ALPHA_ARB              0x8589
#define GL_SOURCE2_ALPHA_ARB              0x858A
#define GL_OPERAND0_RGB_ARB               0x8590
#define GL_OPERAND1_RGB_ARB               0x8591
#define GL_OPERAND2_RGB_ARB               0x8592
#define GL_OPERAND0_ALPHA_ARB             0x8598
#define GL_OPERAND1_ALPHA_ARB             0x8599
#define GL_OPERAND2_ALPHA_ARB             0x859A
#define GL_RGB_SCALE_ARB                  0x8573
#define GL_ADD_SIGNED_ARB                 0x8574
#define GL_INTERPOLATE_ARB                0x8575
#define GL_SUBTRACT_ARB                   0x84E7
#define GL_CONSTANT_ARB                   0x8576
#define GL_PRIMARY_COLOR_ARB              0x8577
#define GL_PREVIOUS_ARB                   0x8578
/* ARB_texture_env_dot3 */
#define GL_DOT3_RGB_ARB                   0x86AE
#define GL_DOT3_RGBA_ARB                  0x86AF
/* ARB_texture_border_clamp */
#define GL_CLAMP_TO_BORDER_ARB            0x812D
/* ARB_transpose_matrix */
#define GL_TRANSPOSE_MODELVIEW_MATRIX_ARB 0x84E3
#define GL_TRANSPOSE_PROJECTION_MATRIX_ARB 0x84E4
#define GL_TRANSPOSE_TEXTURE_MATRIX_ARB   0x84E5
#define GL_TRANSPOSE_COLOR_MATRIX_ARB     0x84E6
/* ARB_depth_texture */
#define GL_DEPTH_COMPONENT16_ARB          0x81A5
#define GL_DEPTH_COMPONENT24_ARB          0x81A6
#define GL_DEPTH_COMPONENT32_ARB          0x81A7
#define GL_TEXTURE_DEPTH_SIZE_ARB         0x884A
#define GL_DEPTH_TEXTURE_MODE_ARB         0x884B
/* ARB_shadow */
#define GL_TEXTURE_COMPARE_MODE_ARB       0x884C
#define GL_TEXTURE_COMPARE_FUNC_ARB       0x884D
#define GL_COMPARE_R_TO_TEXTURE_ARB       0x884E
/* ARB_shadow_ambient */
#define GL_TEXTURE_COMPARE_FAIL_VALUE_ARB 0x80BF
/* ARB_point_parameters */
#define GL_POINT_SIZE_MIN_ARB             0x8126
#define GL_POINT_SIZE_MAX_ARB             0x8127
#define GL_POINT_FADE_THRESHOLD_SIZE_ARB  0x8128
#define GL_POINT_DISTANCE_ATTENUATION_ARB 0x8129
/* ARB_texture_env_crossbar */
/* ARB_texture_mirrored_repeat */
#define GL_MIRRORED_REPEAT_ARB            0x8370
/* ARB_window_pos */
/* ARB_shader_objects */
#define GL_PROGRAM_OBJECT_ARB             0x8B40
#define GL_SHADER_OBJECT_ARB              0x8B48
#define GL_OBJECT_TYPE_ARB                0x8B4E
#define GL_OBJECT_SUBTYPE_ARB             0x8B4F
#define GL_FLOAT_VEC2_ARB                 0x8B50
#define GL_FLOAT_VEC3_ARB                 0x8B51
#define GL_FLOAT_VEC4_ARB                 0x8B52
#define GL_INT_VEC2_ARB                   0x8B53
#define GL_INT_VEC3_ARB                   0x8B54
#define GL_INT_VEC4_ARB                   0x8B55
#define GL_BOOL_ARB                       0x8B56
#define GL_BOOL_VEC2_ARB                  0x8B57
#define GL_BOOL_VEC3_ARB                  0x8B58
#define GL_BOOL_VEC4_ARB                  0x8B59
#define GL_FLOAT_MAT2_ARB                 0x8B5A
#define GL_FLOAT_MAT3_ARB                 0x8B5B
#define GL_FLOAT_MAT4_ARB                 0x8B5C
#define GL_SAMPLER_1D_ARB                 0x8B5D
#define GL_SAMPLER_2D_ARB                 0x8B5E
#define GL_SAMPLER_3D_ARB                 0x8B5F
#define GL_SAMPLER_CUBE_ARB               0x8B60
#define GL_SAMPLER_1D_SHADOW_ARB          0x8B61
#define GL_SAMPLER_2D_SHADOW_ARB          0x8B62
#define GL_SAMPLER_2D_RECT_ARB            0x8B63
#define GL_SAMPLER_2D_RECT_SHADOW_ARB     0x8B64
#define GL_OBJECT_DELETE_STATUS_ARB       0x8B80
#define GL_OBJECT_COMPILE_STATUS_ARB      0x8B81
#define GL_OBJECT_LINK_STATUS_ARB         0x8B82
#define GL_OBJECT_VALIDATE_STATUS_ARB     0x8B83
#define GL_OBJECT_INFO_LOG_LENGTH_ARB     0x8B84
#define GL_OBJECT_ATTACHED_OBJECTS_ARB    0x8B85
#define GL_OBJECT_ACTIVE_UNIFORMS_ARB     0x8B86
#define GL_OBJECT_ACTIVE_UNIFORM_MAX_LENGTH_ARB 0x8B87
#define GL_OBJECT_SHADER_SOURCE_LENGTH_ARB 0x8B88
typedef char GLcharARB; /* native character */
typedef unsigned int GLhandleARB; /* shader object handle */
/* ARB_vertex_shader */
#define GL_VERTEX_SHADER_ARB              0x8B31
#define GL_MAX_VERTEX_UNIFORM_COMPONENTS_ARB 0x8B4A
#define GL_MAX_VARYING_FLOATS_ARB         0x8B4B
#define GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS_ARB 0x8B4C
#define GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS_ARB 0x8B4D
#define GL_OBJECT_ACTIVE_ATTRIBUTES_ARB   0x8B89
#define GL_OBJECT_ACTIVE_ATTRIBUTE_MAX_LENGTH_ARB 0x8B8A
/* ARB_fragment_shader */
#define GL_FRAGMENT_SHADER_ARB            0x8B30
#define GL_MAX_FRAGMENT_UNIFORM_COMPONENTS_ARB 0x8B49
#define GL_FRAGMENT_SHADER_DERIVATIVE_HINT_ARB 0x8B8B
/* ARB_shading_language_100 */
#define GL_SHADING_LANGUAGE_VERSION_ARB   0x8B8C
/* ARB_vertex_program */
#define GL_COLOR_SUM_ARB                  0x8458
#define GL_VERTEX_PROGRAM_ARB             0x8620
#define GL_VERTEX_ATTRIB_ARRAY_ENABLED_ARB 0x8622
#define GL_VERTEX_ATTRIB_ARRAY_SIZE_ARB   0x8623
#define GL_VERTEX_ATTRIB_ARRAY_STRIDE_ARB 0x8624
#define GL_VERTEX_ATTRIB_ARRAY_TYPE_ARB   0x8625
#define GL_CURRENT_VERTEX_ATTRIB_ARB      0x8626
#define GL_PROGRAM_LENGTH_ARB             0x8627
#define GL_PROGRAM_STRING_ARB             0x8628
#define GL_MAX_PROGRAM_MATRIX_STACK_DEPTH_ARB 0x862E
#define GL_MAX_PROGRAM_MATRICES_ARB       0x862F
#define GL_CURRENT_MATRIX_STACK_DEPTH_ARB 0x8640
#define GL_CURRENT_MATRIX_ARB             0x8641
#define GL_VERTEX_PROGRAM_POINT_SIZE_ARB  0x8642
#define GL_VERTEX_PROGRAM_TWO_SIDE_ARB    0x8643
#define GL_VERTEX_ATTRIB_ARRAY_POINTER_ARB 0x8645
#define GL_PROGRAM_ERROR_POSITION_ARB     0x864B
#define GL_PROGRAM_BINDING_ARB            0x8677
#define GL_MAX_VERTEX_ATTRIBS_ARB         0x8869
#define GL_VERTEX_ATTRIB_ARRAY_NORMALIZED_ARB 0x886A
#define GL_PROGRAM_ERROR_STRING_ARB       0x8874
#define GL_PROGRAM_FORMAT_ASCII_ARB       0x8875
#define GL_PROGRAM_FORMAT_ARB             0x8876
#define GL_PROGRAM_INSTRUCTIONS_ARB       0x88A0
#define GL_MAX_PROGRAM_INSTRUCTIONS_ARB   0x88A1
#define GL_PROGRAM_NATIVE_INSTRUCTIONS_ARB 0x88A2
#define GL_MAX_PROGRAM_NATIVE_INSTRUCTIONS_ARB 0x88A3
#define GL_PROGRAM_TEMPORARIES_ARB        0x88A4
#define GL_MAX_PROGRAM_TEMPORARIES_ARB    0x88A5
#define GL_PROGRAM_NATIVE_TEMPORARIES_ARB 0x88A6
#define GL_MAX_PROGRAM_NATIVE_TEMPORARIES_ARB 0x88A7
#define GL_PROGRAM_PARAMETERS_ARB         0x88A8
#define GL_MAX_PROGRAM_PARAMETERS_ARB     0x88A9
#define GL_PROGRAM_NATIVE_PARAMETERS_ARB  0x88AA
#define GL_MAX_PROGRAM_NATIVE_PARAMETERS_ARB 0x88AB
#define GL_PROGRAM_ATTRIBS_ARB            0x88AC
#define GL_MAX_PROGRAM_ATTRIBS_ARB        0x88AD
#define GL_PROGRAM_NATIVE_ATTRIBS_ARB     0x88AE
#define GL_MAX_PROGRAM_NATIVE_ATTRIBS_ARB 0x88AF
#define GL_PROGRAM_ADDRESS_REGISTERS_ARB  0x88B0
#define GL_MAX_PROGRAM_ADDRESS_REGISTERS_ARB 0x88B1
#define GL_PROGRAM_NATIVE_ADDRESS_REGISTERS_ARB 0x88B2
#define GL_MAX_PROGRAM_NATIVE_ADDRESS_REGISTERS_ARB 0x88B3
#define GL_MAX_PROGRAM_LOCAL_PARAMETERS_ARB 0x88B4
#define GL_MAX_PROGRAM_ENV_PARAMETERS_ARB 0x88B5
#define GL_PROGRAM_UNDER_NATIVE_LIMITS_ARB 0x88B6
#define GL_TRANSPOSE_CURRENT_MATRIX_ARB   0x88B7
#define GL_MATRIX0_ARB                    0x88C0
#define GL_MATRIX1_ARB                    0x88C1
#define GL_MATRIX2_ARB                    0x88C2
#define GL_MATRIX3_ARB                    0x88C3
#define GL_MATRIX4_ARB                    0x88C4
#define GL_MATRIX5_ARB                    0x88C5
#define GL_MATRIX6_ARB                    0x88C6
#define GL_MATRIX7_ARB                    0x88C7
#define GL_MATRIX8_ARB                    0x88C8
#define GL_MATRIX9_ARB                    0x88C9
#define GL_MATRIX10_ARB                   0x88CA
#define GL_MATRIX11_ARB                   0x88CB
#define GL_MATRIX12_ARB                   0x88CC
#define GL_MATRIX13_ARB                   0x88CD
#define GL_MATRIX14_ARB                   0x88CE
#define GL_MATRIX15_ARB                   0x88CF
#define GL_MATRIX16_ARB                   0x88D0
#define GL_MATRIX17_ARB                   0x88D1
#define GL_MATRIX18_ARB                   0x88D2
#define GL_MATRIX19_ARB                   0x88D3
#define GL_MATRIX20_ARB                   0x88D4
#define GL_MATRIX21_ARB                   0x88D5
#define GL_MATRIX22_ARB                   0x88D6
#define GL_MATRIX23_ARB                   0x88D7
#define GL_MATRIX24_ARB                   0x88D8
#define GL_MATRIX25_ARB                   0x88D9
#define GL_MATRIX26_ARB                   0x88DA
#define GL_MATRIX27_ARB                   0x88DB
#define GL_MATRIX28_ARB                   0x88DC
#define GL_MATRIX29_ARB                   0x88DD
#define GL_MATRIX30_ARB                   0x88DE
#define GL_MATRIX31_ARB                   0x88DF
/* ARB_fragment_program */
#define GL_FRAGMENT_PROGRAM_ARB           0x8804
#define GL_PROGRAM_ALU_INSTRUCTIONS_ARB   0x8805
#define GL_PROGRAM_TEX_INSTRUCTIONS_ARB   0x8806
#define GL_PROGRAM_TEX_INDIRECTIONS_ARB   0x8807
#define GL_PROGRAM_NATIVE_ALU_INSTRUCTIONS_ARB 0x8808
#define GL_PROGRAM_NATIVE_TEX_INSTRUCTIONS_ARB 0x8809
#define GL_PROGRAM_NATIVE_TEX_INDIRECTIONS_ARB 0x880A
#define GL_MAX_PROGRAM_ALU_INSTRUCTIONS_ARB 0x880B
#define GL_MAX_PROGRAM_TEX_INSTRUCTIONS_ARB 0x880C
#define GL_MAX_PROGRAM_TEX_INDIRECTIONS_ARB 0x880D
#define GL_MAX_PROGRAM_NATIVE_ALU_INSTRUCTIONS_ARB 0x880E
#define GL_MAX_PROGRAM_NATIVE_TEX_INSTRUCTIONS_ARB 0x880F
#define GL_MAX_PROGRAM_NATIVE_TEX_INDIRECTIONS_ARB 0x8810
#define GL_MAX_TEXTURE_COORDS_ARB         0x8871
#define GL_MAX_TEXTURE_IMAGE_UNITS_ARB    0x8872
/* ARB_fragment_program_shadow */
/* ARB_vertex_buffer_object */
#define GL_BUFFER_SIZE_ARB                0x8764
#define GL_BUFFER_USAGE_ARB               0x8765
#define GL_ARRAY_BUFFER_ARB               0x8892
#define GL_ELEMENT_ARRAY_BUFFER_ARB       0x8893
#define GL_ARRAY_BUFFER_BINDING_ARB       0x8894
#define GL_ELEMENT_ARRAY_BUFFER_BINDING_ARB 0x8895
#define GL_VERTEX_ARRAY_BUFFER_BINDING_ARB 0x8896
#define GL_NORMAL_ARRAY_BUFFER_BINDING_ARB 0x8897
#define GL_COLOR_ARRAY_BUFFER_BINDING_ARB 0x8898
#define GL_INDEX_ARRAY_BUFFER_BINDING_ARB 0x8899
#define GL_TEXTURE_COORD_ARRAY_BUFFER_BINDING_ARB 0x889A
#define GL_EDGE_FLAG_ARRAY_BUFFER_BINDING_ARB 0x889B
#define GL_SECONDARY_COLOR_ARRAY_BUFFER_BINDING_ARB 0x889C
#define GL_FOG_COORDINATE_ARRAY_BUFFER_BINDING_ARB 0x889D
#define GL_WEIGHT_ARRAY_BUFFER_BINDING_ARB 0x889E
#define GL_VERTEX_ATTRIB_ARRAY_BUFFER_BINDING_ARB 0x889F
#define GL_READ_ONLY_ARB                  0x88B8
#define GL_WRITE_ONLY_ARB                 0x88B9
#define GL_READ_WRITE_ARB                 0x88BA
#define GL_BUFFER_ACCESS_ARB              0x88BB
#define GL_BUFFER_MAPPED_ARB              0x88BC
#define GL_BUFFER_MAP_POINTER_ARB         0x88BD
#define GL_STREAM_DRAW_ARB                0x88E0
#define GL_STREAM_READ_ARB                0x88E1
#define GL_STREAM_COPY_ARB                0x88E2
#define GL_STATIC_DRAW_ARB                0x88E4
#define GL_STATIC_READ_ARB                0x88E5
#define GL_STATIC_COPY_ARB                0x88E6
#define GL_DYNAMIC_DRAW_ARB               0x88E8
#define GL_DYNAMIC_READ_ARB               0x88E9
#define GL_DYNAMIC_COPY_ARB               0x88EA
typedef ptrdiff_t GLintptrARB;
typedef ptrdiff_t GLsizeiptrARB;
/* ARB_occlusion_query */
#define GL_QUERY_COUNTER_BITS_ARB         0x8864
#define GL_CURRENT_QUERY_ARB              0x8865
#define GL_QUERY_RESULT_ARB               0x8866
#define GL_QUERY_RESULT_AVAILABLE_ARB     0x8867
#define GL_SAMPLES_PASSED_ARB             0x8914
/* ARB_draw_buffers */
#define GL_MAX_DRAW_BUFFERS_ARB           0x8824
#define GL_DRAW_BUFFER0_ARB               0x8825
#define GL_DRAW_BUFFER1_ARB               0x8826
#define GL_DRAW_BUFFER2_ARB               0x8827
#define GL_DRAW_BUFFER3_ARB               0x8828
#define GL_DRAW_BUFFER4_ARB               0x8829
#define GL_DRAW_BUFFER5_ARB               0x882A
#define GL_DRAW_BUFFER6_ARB               0x882B
#define GL_DRAW_BUFFER7_ARB               0x882C
#define GL_DRAW_BUFFER8_ARB               0x882D
#define GL_DRAW_BUFFER9_ARB               0x882E
#define GL_DRAW_BUFFER10_ARB              0x882F
#define GL_DRAW_BUFFER11_ARB              0x8830
#define GL_DRAW_BUFFER12_ARB              0x8831
#define GL_DRAW_BUFFER13_ARB              0x8832
#define GL_DRAW_BUFFER14_ARB              0x8833
#define GL_DRAW_BUFFER15_ARB              0x8834
/* ARB_texture_non_power_of_two */
/* ARB_point_sprite */
#define GL_POINT_SPRITE_ARB               0x8861
#define GL_COORD_REPLACE_ARB              0x8862
/* ARB_texture_rectangle */
#define GL_TEXTURE_RECTANGLE_ARB          0x84F5
#define GL_TEXTURE_BINDING_RECTANGLE_ARB  0x84F6
#define GL_PROXY_TEXTURE_RECTANGLE_ARB    0x84F7
#define GL_MAX_RECTANGLE_TEXTURE_SIZE_ARB 0x84F8
/* ARB_imaging */
#define GL_CONSTANT_COLOR                 0x8001
#define GL_ONE_MINUS_CONSTANT_COLOR       0x8002
#define GL_CONSTANT_ALPHA                 0x8003
#define GL_ONE_MINUS_CONSTANT_ALPHA       0x8004
#define GL_BLEND_COLOR                    0x8005
#define GL_FUNC_ADD                       0x8006
#define GL_MIN                            0x8007
#define GL_MAX                            0x8008
#define GL_BLEND_EQUATION                 0x8009
#define GL_FUNC_SUBTRACT                  0x800A
#define GL_FUNC_REVERSE_SUBTRACT          0x800B
#define GL_CONVOLUTION_1D                 0x8010
#define GL_CONVOLUTION_2D                 0x8011
#define GL_SEPARABLE_2D                   0x8012
#define GL_CONVOLUTION_BORDER_MODE        0x8013
#define GL_CONVOLUTION_FILTER_SCALE       0x8014
#define GL_CONVOLUTION_FILTER_BIAS        0x8015
#define GL_REDUCE                         0x8016
#define GL_CONVOLUTION_FORMAT             0x8017
#define GL_CONVOLUTION_WIDTH              0x8018
#define GL_CONVOLUTION_HEIGHT             0x8019
#define GL_MAX_CONVOLUTION_WIDTH          0x801A
#define GL_MAX_CONVOLUTION_HEIGHT         0x801B
#define GL_POST_CONVOLUTION_RED_SCALE     0x801C
#define GL_POST_CONVOLUTION_GREEN_SCALE   0x801D
#define GL_POST_CONVOLUTION_BLUE_SCALE    0x801E
#define GL_POST_CONVOLUTION_ALPHA_SCALE   0x801F
#define GL_POST_CONVOLUTION_RED_BIAS      0x8020
#define GL_POST_CONVOLUTION_GREEN_BIAS    0x8021
#define GL_POST_CONVOLUTION_BLUE_BIAS     0x8022
#define GL_POST_CONVOLUTION_ALPHA_BIAS    0x8023
#define GL_HISTOGRAM                      0x8024
#define GL_PROXY_HISTOGRAM                0x8025
#define GL_HISTOGRAM_WIDTH                0x8026
#define GL_HISTOGRAM_FORMAT               0x8027
#define GL_HISTOGRAM_RED_SIZE             0x8028
#define GL_HISTOGRAM_GREEN_SIZE           0x8029
#define GL_HISTOGRAM_BLUE_SIZE            0x802A
#define GL_HISTOGRAM_ALPHA_SIZE           0x802B
#define GL_HISTOGRAM_LUMINANCE_SIZE       0x802C
#define GL_HISTOGRAM_SINK                 0x802D
#define GL_MINMAX                         0x802E
#define GL_MINMAX_FORMAT                  0x802F
#define GL_MINMAX_SINK                    0x8030
#define GL_TABLE_TOO_LARGE                0x8031
#define GL_COLOR_MATRIX                   0x80B1
#define GL_COLOR_MATRIX_STACK_DEPTH       0x80B2
#define GL_MAX_COLOR_MATRIX_STACK_DEPTH   0x80B3
#define GL_POST_COLOR_MATRIX_RED_SCALE    0x80B4
#define GL_POST_COLOR_MATRIX_GREEN_SCALE  0x80B5
#define GL_POST_COLOR_MATRIX_BLUE_SCALE   0x80B6
#define GL_POST_COLOR_MATRIX_ALPHA_SCALE  0x80B7
#define GL_POST_COLOR_MATRIX_RED_BIAS     0x80B8
#define GL_POST_COLOR_MATRIX_GREEN_BIAS   0x80B9
#define GL_POST_COLOR_MATRIX_BLUE_BIAS    0x80BA
#define GL_POST_COLOR_MATRIX_ALPHA_BIAS   0x80BB
#define GL_COLOR_TABLE                    0x80D0
#define GL_POST_CONVOLUTION_COLOR_TABLE   0x80D1
#define GL_POST_COLOR_MATRIX_COLOR_TABLE  0x80D2
#define GL_PROXY_COLOR_TABLE              0x80D3
#define GL_PROXY_POST_CONVOLUTION_COLOR_TABLE 0x80D4
#define GL_PROXY_POST_COLOR_MATRIX_COLOR_TABLE 0x80D5
#define GL_COLOR_TABLE_SCALE              0x80D6
#define GL_COLOR_TABLE_BIAS               0x80D7
#define GL_COLOR_TABLE_FORMAT             0x80D8
#define GL_COLOR_TABLE_WIDTH              0x80D9
#define GL_COLOR_TABLE_RED_SIZE           0x80DA
#define GL_COLOR_TABLE_GREEN_SIZE         0x80DB
#define GL_COLOR_TABLE_BLUE_SIZE          0x80DC
#define GL_COLOR_TABLE_ALPHA_SIZE         0x80DD
#define GL_COLOR_TABLE_LUMINANCE_SIZE     0x80DE
#define GL_COLOR_TABLE_INTENSITY_SIZE     0x80DF
#define GL_CONSTANT_BORDER                0x8151
#define GL_REPLICATE_BORDER               0x8153
#define GL_CONVOLUTION_BORDER_COLOR       0x8154

#ifdef	DLL_TARGET_DCC

#ifdef	__cplusplus
extern "C" {
#endif /* __cplusplus */

/*
 * Prototypes
 */

/* GL_ARB_multitexture */
VGLEXTAPI void VGLEXTAPIENTRY glActiveTextureARB (GLenum);
VGLEXTAPI void VGLEXTAPIENTRY glClientActiveTextureARB (GLenum);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord1dARB (GLenum, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord1dvARB (GLenum, const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord1fARB (GLenum, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord1fvARB (GLenum, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord1iARB (GLenum, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord1ivARB (GLenum, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord1sARB (GLenum, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord1svARB (GLenum, const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord2dARB (GLenum, GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord2dvARB (GLenum, const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord2fARB (GLenum, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord2fvARB (GLenum, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord2iARB (GLenum, GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord2ivARB (GLenum, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord2sARB (GLenum, GLshort, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord2svARB (GLenum, const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord3dARB (GLenum, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord3dvARB (GLenum, const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord3fARB (GLenum, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord3fvARB (GLenum, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord3iARB (GLenum, GLint, GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord3ivARB (GLenum, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord3sARB (GLenum, GLshort, GLshort, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord3svARB (GLenum, const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord4dARB (GLenum, GLdouble, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord4dvARB (GLenum, const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord4fARB (GLenum, GLfloat, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord4fvARB (GLenum, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord4iARB (GLenum, GLint, GLint, GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord4ivARB (GLenum, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord4sARB (GLenum, GLshort, GLshort, GLshort, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord4svARB (GLenum, const GLshort *);
/* ARB_texture_compression */
VGLEXTAPI void VGLEXTAPIENTRY glCompressedTexImage3DARB (GLenum, GLint, GLenum, GLsizei, GLsizei, GLsizei, GLint, GLsizei, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glCompressedTexImage2DARB (GLenum, GLint, GLenum, GLsizei, GLsizei, GLint, GLsizei, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glCompressedTexImage1DARB (GLenum, GLint, GLenum, GLsizei, GLint, GLsizei, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glCompressedTexSubImage3DARB (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glCompressedTexSubImage2DARB (GLenum, GLint, GLint, GLint, GLsizei, GLsizei, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glCompressedTexSubImage1DARB (GLenum, GLint, GLint, GLsizei, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glGetCompressedTexImageARB (GLenum, GLint, GLvoid *);
/* ARB_multisample */
VGLEXTAPI void VGLEXTAPIENTRY glSampleCoverageARB (GLclampf, GLboolean);
/* ARB_transpose_matrix */
VGLEXTAPI void VGLEXTAPIENTRY glLoadTransposeMatrixfARB (const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glLoadTransposeMatrixdARB (const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glMultTransposeMatrixfARB (const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glMultTransposeMatrixdARB (const GLdouble *);
/* ARB_point_parameters */
VGLEXTAPI void VGLEXTAPIENTRY glPointParameterfARB (GLenum, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glPointParameterfvARB (GLenum, const GLfloat *);
/* ARB_window_pos */
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos2dARB (GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos2dvARB (const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos2fARB (GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos2fvARB (const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos2iARB (GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos2ivARB (const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos2sARB (GLshort, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos2svARB (const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos3dARB (GLdouble, GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos3dvARB (const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos3fARB (GLfloat, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos3fvARB (const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos3iARB (GLint, GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos3ivARB (const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos3sARB (GLshort, GLshort, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos3svARB (const GLshort *);
/* ARB_shader_objects */
VGLEXTAPI void VGLEXTAPIENTRY glDeleteObjectARB (GLhandleARB);
VGLEXTAPI GLhandleARB VGLEXTAPIENTRY glGetHandleARB (GLenum);
VGLEXTAPI void VGLEXTAPIENTRY glDetachObjectARB (GLhandleARB, GLhandleARB);
VGLEXTAPI GLhandleARB VGLEXTAPIENTRY glCreateShaderObjectARB (GLenum);
VGLEXTAPI void VGLEXTAPIENTRY glShaderSourceARB (GLhandleARB, GLsizei, const GLcharARB* *, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glCompileShaderARB (GLhandleARB);
VGLEXTAPI GLhandleARB VGLEXTAPIENTRY glCreateProgramObjectARB (void);
VGLEXTAPI void VGLEXTAPIENTRY glAttachObjectARB (GLhandleARB, GLhandleARB);
VGLEXTAPI void VGLEXTAPIENTRY glLinkProgramARB (GLhandleARB);
VGLEXTAPI void VGLEXTAPIENTRY glUseProgramObjectARB (GLhandleARB);
VGLEXTAPI void VGLEXTAPIENTRY glValidateProgramARB (GLhandleARB);
VGLEXTAPI void VGLEXTAPIENTRY glUniform1fARB (GLint, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glUniform2fARB (GLint, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glUniform3fARB (GLint, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glUniform4fARB (GLint, GLfloat, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glUniform1iARB (GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glUniform2iARB (GLint, GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glUniform3iARB (GLint, GLint, GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glUniform4iARB (GLint, GLint, GLint, GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glUniform1fvARB (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glUniform2fvARB (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glUniform3fvARB (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glUniform4fvARB (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glUniform1ivARB (GLint, GLsizei, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glUniform2ivARB (GLint, GLsizei, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glUniform3ivARB (GLint, GLsizei, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glUniform4ivARB (GLint, GLsizei, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glUniformMatrix2fvARB (GLint, GLsizei, GLboolean, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glUniformMatrix3fvARB (GLint, GLsizei, GLboolean, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glUniformMatrix4fvARB (GLint, GLsizei, GLboolean, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glGetObjectParameterfvARB (GLhandleARB, GLenum, GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glGetObjectParameterivARB (GLhandleARB, GLenum, GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glGetInfoLogARB (GLhandleARB, GLsizei, GLsizei *, GLcharARB *);
VGLEXTAPI void VGLEXTAPIENTRY glGetAttachedObjectsARB (GLhandleARB, GLsizei, GLsizei *, GLhandleARB *);
VGLEXTAPI GLint VGLEXTAPIENTRY glGetUniformLocationARB (GLhandleARB, const GLcharARB *);
VGLEXTAPI void VGLEXTAPIENTRY glGetActiveUniformARB (GLhandleARB, GLuint, GLsizei, GLsizei *, GLint *, GLenum *, GLcharARB *);
VGLEXTAPI void VGLEXTAPIENTRY glGetUniformfvARB (GLhandleARB, GLint, GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glGetUniformivARB (GLhandleARB, GLint, GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glGetShaderSourceARB (GLhandleARB, GLsizei, GLsizei *, GLcharARB *);
/* ARB_vertex_shader */
VGLEXTAPI void VGLEXTAPIENTRY glBindAttribLocationARB (GLhandleARB, GLuint, const GLcharARB *);
VGLEXTAPI void VGLEXTAPIENTRY glGetActiveAttribARB (GLhandleARB, GLuint, GLsizei, GLsizei *, GLint *, GLenum *, GLcharARB *);
VGLEXTAPI GLint VGLEXTAPIENTRY glGetAttribLocationARB (GLhandleARB, const GLcharARB *);
/* ARB_vertex_program */
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib1dARB (GLuint, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib1dvARB (GLuint, const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib1fARB (GLuint, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib1fvARB (GLuint, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib1sARB (GLuint, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib1svARB (GLuint, const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib2dARB (GLuint, GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib2dvARB (GLuint, const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib2fARB (GLuint, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib2fvARB (GLuint, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib2sARB (GLuint, GLshort, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib2svARB (GLuint, const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib3dARB (GLuint, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib3dvARB (GLuint, const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib3fARB (GLuint, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib3fvARB (GLuint, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib3sARB (GLuint, GLshort, GLshort, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib3svARB (GLuint, const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4NbvARB (GLuint, const GLbyte *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4NivARB (GLuint, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4NsvARB (GLuint, const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4NubARB (GLuint, GLubyte, GLubyte, GLubyte, GLubyte);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4NubvARB (GLuint, const GLubyte *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4NuivARB (GLuint, const GLuint *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4NusvARB (GLuint, const GLushort *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4bvARB (GLuint, const GLbyte *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4dARB (GLuint, GLdouble, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4dvARB (GLuint, const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4fARB (GLuint, GLfloat, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4fvARB (GLuint, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4ivARB (GLuint, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4sARB (GLuint, GLshort, GLshort, GLshort, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4svARB (GLuint, const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4ubvARB (GLuint, const GLubyte *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4uivARB (GLuint, const GLuint *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4usvARB (GLuint, const GLushort *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttribPointerARB (GLuint, GLint, GLenum, GLboolean, GLsizei, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glEnableVertexAttribArrayARB (GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glDisableVertexAttribArrayARB (GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glProgramStringARB (GLenum, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glBindProgramARB (GLenum, GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glDeleteProgramsARB (GLsizei, const GLuint *);
VGLEXTAPI void VGLEXTAPIENTRY glGenProgramsARB (GLsizei, GLuint *);
VGLEXTAPI void VGLEXTAPIENTRY glProgramEnvParameter4dARB (GLenum, GLuint, GLdouble, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glProgramEnvParameter4dvARB (GLenum, GLuint, const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glProgramEnvParameter4fARB (GLenum, GLuint, GLfloat, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glProgramEnvParameter4fvARB (GLenum, GLuint, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glProgramLocalParameter4dARB (GLenum, GLuint, GLdouble, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glProgramLocalParameter4dvARB (GLenum, GLuint, const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glProgramLocalParameter4fARB (GLenum, GLuint, GLfloat, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glProgramLocalParameter4fvARB (GLenum, GLuint, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glGetProgramEnvParameterdvARB (GLenum, GLuint, GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glGetProgramEnvParameterfvARB (GLenum, GLuint, GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glGetProgramLocalParameterdvARB (GLenum, GLuint, GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glGetProgramLocalParameterfvARB (GLenum, GLuint, GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glGetProgramivARB (GLenum, GLenum, GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glGetProgramStringARB (GLenum, GLenum, GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glGetVertexAttribdvARB (GLuint, GLenum, GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glGetVertexAttribfvARB (GLuint, GLenum, GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glGetVertexAttribivARB (GLuint, GLenum, GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glGetVertexAttribPointervARB (GLuint, GLenum, GLvoid* *);
VGLEXTAPI GLboolean VGLEXTAPIENTRY glIsProgramARB (GLuint);
/* ARB_vertex_buffer_object */
VGLEXTAPI void VGLEXTAPIENTRY glBindBufferARB (GLenum, GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glDeleteBuffersARB (GLsizei, const GLuint *);
VGLEXTAPI void VGLEXTAPIENTRY glGenBuffersARB (GLsizei, GLuint *);
VGLEXTAPI GLboolean VGLEXTAPIENTRY glIsBufferARB (GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glBufferDataARB (GLenum, GLsizeiptrARB, const GLvoid *, GLenum);
VGLEXTAPI void VGLEXTAPIENTRY glBufferSubDataARB (GLenum, GLintptrARB, GLsizeiptrARB, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glGetBufferSubDataARB (GLenum, GLintptrARB, GLsizeiptrARB, GLvoid *);
VGLEXTAPI GLvoid* VGLEXTAPIENTRY glMapBufferARB (GLenum, GLenum);
VGLEXTAPI GLboolean VGLEXTAPIENTRY glUnmapBufferARB (GLenum);
VGLEXTAPI void VGLEXTAPIENTRY glGetBufferParameterivARB (GLenum, GLenum, GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glGetBufferPointervARB (GLenum, GLenum, GLvoid* *);
/* ARB_occlusion_query */
VGLEXTAPI void VGLEXTAPIENTRY glGenQueriesARB (GLsizei, GLuint *);
VGLEXTAPI void VGLEXTAPIENTRY glDeleteQueriesARB (GLsizei, const GLuint *);
VGLEXTAPI GLboolean VGLEXTAPIENTRY glIsQueryARB (GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glBeginQueryARB (GLenum, GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glEndQueryARB (GLenum);
VGLEXTAPI void VGLEXTAPIENTRY glGetQueryivARB (GLenum, GLenum, GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glGetQueryObjectivARB (GLuint, GLenum, GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glGetQueryObjectuivARB (GLuint, GLenum, GLuint *);
/* ARB_draw_buffers */
VGLEXTAPI void VGLEXTAPIENTRY glDrawBuffersARB (GLsizei, const GLenum *);


#ifdef	__cplusplus
}
#endif /* __cplusplus */

#else	/* !DLL_TARGET_DCC */

/*
 * Pointers
 */

/* GL_ARB_multitexture */

VGLEXTAPI void (VGLEXTAPIENTRY *glActiveTextureARB) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glClientActiveTextureARB) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1dARB) (GLenum, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1dvARB) (GLenum, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1fARB) (GLenum, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1fvARB) (GLenum, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1iARB) (GLenum, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1ivARB) (GLenum, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1sARB) (GLenum, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1svARB) (GLenum, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2dARB) (GLenum, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2dvARB) (GLenum, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2fARB) (GLenum, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2fvARB) (GLenum, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2iARB) (GLenum, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2ivARB) (GLenum, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2sARB) (GLenum, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2svARB) (GLenum, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3dARB) (GLenum, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3dvARB) (GLenum, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3fARB) (GLenum, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3fvARB) (GLenum, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3iARB) (GLenum, GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3ivARB) (GLenum, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3sARB) (GLenum, GLshort, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3svARB) (GLenum, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4dARB) (GLenum, GLdouble, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4dvARB) (GLenum, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4fARB) (GLenum, GLfloat, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4fvARB) (GLenum, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4iARB) (GLenum, GLint, GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4ivARB) (GLenum, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4sARB) (GLenum, GLshort, GLshort, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4svARB) (GLenum, const GLshort *);
/* ARB_texture_compression */
VGLEXTAPI void (VGLEXTAPIENTRY *glCompressedTexImage3DARB) (GLenum, GLint, GLenum, GLsizei, GLsizei, GLsizei, GLint, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompressedTexImage2DARB) (GLenum, GLint, GLenum, GLsizei, GLsizei, GLint, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompressedTexImage1DARB) (GLenum, GLint, GLenum, GLsizei, GLint, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompressedTexSubImage3DARB) (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompressedTexSubImage2DARB) (GLenum, GLint, GLint, GLint, GLsizei, GLsizei, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompressedTexSubImage1DARB) (GLenum, GLint, GLint, GLsizei, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetCompressedTexImageARB) (GLenum, GLint, GLvoid *);
/* ARB_multisample */
VGLEXTAPI void (VGLEXTAPIENTRY *glSampleCoverageARB) (GLclampf, GLboolean);
/* ARB_transpose_matrix */
VGLEXTAPI void (VGLEXTAPIENTRY *glLoadTransposeMatrixfARB) (const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glLoadTransposeMatrixdARB) (const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultTransposeMatrixfARB) (const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultTransposeMatrixdARB) (const GLdouble *);
/* ARB_point_parameters */
VGLEXTAPI void (VGLEXTAPIENTRY *glPointParameterfARB) (GLenum, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glPointParameterfvARB) (GLenum, const GLfloat *);
/* ARB_window_pos */
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2dARB) (GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2dvARB) (const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2fARB) (GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2fvARB) (const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2iARB) (GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2ivARB) (const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2sARB) (GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2svARB) (const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3dARB) (GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3dvARB) (const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3fARB) (GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3fvARB) (const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3iARB) (GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3ivARB) (const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3sARB) (GLshort, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3svARB) (const GLshort *);
/* ARB_shader_objects */
VGLEXTAPI void (VGLEXTAPIENTRY *glDeleteObjectARB) (GLhandleARB);
VGLEXTAPI GLhandleARB (VGLEXTAPIENTRY *glGetHandleARB) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glDetachObjectARB) (GLhandleARB, GLhandleARB);
VGLEXTAPI GLhandleARB (VGLEXTAPIENTRY *glCreateShaderObjectARB) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glShaderSourceARB) (GLhandleARB, GLsizei, const GLcharARB* *, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompileShaderARB) (GLhandleARB);
VGLEXTAPI GLhandleARB (VGLEXTAPIENTRY *glCreateProgramObjectARB) (void);
VGLEXTAPI void (VGLEXTAPIENTRY *glAttachObjectARB) (GLhandleARB, GLhandleARB);
VGLEXTAPI void (VGLEXTAPIENTRY *glLinkProgramARB) (GLhandleARB);
VGLEXTAPI void (VGLEXTAPIENTRY *glUseProgramObjectARB) (GLhandleARB);
VGLEXTAPI void (VGLEXTAPIENTRY *glValidateProgramARB) (GLhandleARB);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform1fARB) (GLint, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform2fARB) (GLint, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform3fARB) (GLint, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform4fARB) (GLint, GLfloat, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform1iARB) (GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform2iARB) (GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform3iARB) (GLint, GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform4iARB) (GLint, GLint, GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform1fvARB) (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform2fvARB) (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform3fvARB) (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform4fvARB) (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform1ivARB) (GLint, GLsizei, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform2ivARB) (GLint, GLsizei, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform3ivARB) (GLint, GLsizei, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform4ivARB) (GLint, GLsizei, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniformMatrix2fvARB) (GLint, GLsizei, GLboolean, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniformMatrix3fvARB) (GLint, GLsizei, GLboolean, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniformMatrix4fvARB) (GLint, GLsizei, GLboolean, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetObjectParameterfvARB) (GLhandleARB, GLenum, GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetObjectParameterivARB) (GLhandleARB, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetInfoLogARB) (GLhandleARB, GLsizei, GLsizei *, GLcharARB *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetAttachedObjectsARB) (GLhandleARB, GLsizei, GLsizei *, GLhandleARB *);
VGLEXTAPI GLint (VGLEXTAPIENTRY *glGetUniformLocationARB) (GLhandleARB, const GLcharARB *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetActiveUniformARB) (GLhandleARB, GLuint, GLsizei, GLsizei *, GLint *, GLenum *, GLcharARB *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetUniformfvARB) (GLhandleARB, GLint, GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetUniformivARB) (GLhandleARB, GLint, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetShaderSourceARB) (GLhandleARB, GLsizei, GLsizei *, GLcharARB *);
/* ARB_vertex_shader */
VGLEXTAPI void (VGLEXTAPIENTRY *glBindAttribLocationARB) (GLhandleARB, GLuint, const GLcharARB *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetActiveAttribARB) (GLhandleARB, GLuint, GLsizei, GLsizei *, GLint *, GLenum *, GLcharARB *);
VGLEXTAPI GLint (VGLEXTAPIENTRY *glGetAttribLocationARB) (GLhandleARB, const GLcharARB *);
/* ARB_vertex_program */
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib1dARB) (GLuint, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib1dvARB) (GLuint, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib1fARB) (GLuint, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib1fvARB) (GLuint, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib1sARB) (GLuint, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib1svARB) (GLuint, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib2dARB) (GLuint, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib2dvARB) (GLuint, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib2fARB) (GLuint, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib2fvARB) (GLuint, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib2sARB) (GLuint, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib2svARB) (GLuint, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib3dARB) (GLuint, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib3dvARB) (GLuint, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib3fARB) (GLuint, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib3fvARB) (GLuint, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib3sARB) (GLuint, GLshort, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib3svARB) (GLuint, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4NbvARB) (GLuint, const GLbyte *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4NivARB) (GLuint, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4NsvARB) (GLuint, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4NubARB) (GLuint, GLubyte, GLubyte, GLubyte, GLubyte);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4NubvARB) (GLuint, const GLubyte *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4NuivARB) (GLuint, const GLuint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4NusvARB) (GLuint, const GLushort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4bvARB) (GLuint, const GLbyte *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4dARB) (GLuint, GLdouble, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4dvARB) (GLuint, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4fARB) (GLuint, GLfloat, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4fvARB) (GLuint, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4ivARB) (GLuint, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4sARB) (GLuint, GLshort, GLshort, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4svARB) (GLuint, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4ubvARB) (GLuint, const GLubyte *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4uivARB) (GLuint, const GLuint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4usvARB) (GLuint, const GLushort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttribPointerARB) (GLuint, GLint, GLenum, GLboolean, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glEnableVertexAttribArrayARB) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glDisableVertexAttribArrayARB) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glProgramStringARB) (GLenum, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glBindProgramARB) (GLenum, GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glDeleteProgramsARB) (GLsizei, const GLuint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGenProgramsARB) (GLsizei, GLuint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glProgramEnvParameter4dARB) (GLenum, GLuint, GLdouble, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glProgramEnvParameter4dvARB) (GLenum, GLuint, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glProgramEnvParameter4fARB) (GLenum, GLuint, GLfloat, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glProgramEnvParameter4fvARB) (GLenum, GLuint, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glProgramLocalParameter4dARB) (GLenum, GLuint, GLdouble, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glProgramLocalParameter4dvARB) (GLenum, GLuint, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glProgramLocalParameter4fARB) (GLenum, GLuint, GLfloat, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glProgramLocalParameter4fvARB) (GLenum, GLuint, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetProgramEnvParameterdvARB) (GLenum, GLuint, GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetProgramEnvParameterfvARB) (GLenum, GLuint, GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetProgramLocalParameterdvARB) (GLenum, GLuint, GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetProgramLocalParameterfvARB) (GLenum, GLuint, GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetProgramivARB) (GLenum, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetProgramStringARB) (GLenum, GLenum, GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetVertexAttribdvARB) (GLuint, GLenum, GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetVertexAttribfvARB) (GLuint, GLenum, GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetVertexAttribivARB) (GLuint, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetVertexAttribPointervARB) (GLuint, GLenum, GLvoid* *);
VGLEXTAPI GLboolean (VGLEXTAPIENTRY *glIsProgramARB) (GLuint);
/* ARB_vertex_buffer_object */
VGLEXTAPI void (VGLEXTAPIENTRY *glBindBufferARB) (GLenum, GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glDeleteBuffersARB) (GLsizei, const GLuint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGenBuffersARB) (GLsizei, GLuint *);
VGLEXTAPI GLboolean (VGLEXTAPIENTRY *glIsBufferARB) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glBufferDataARB) (GLenum, GLsizeiptrARB, const GLvoid *, GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glBufferSubDataARB) (GLenum, GLintptrARB, GLsizeiptrARB, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetBufferSubDataARB) (GLenum, GLintptrARB, GLsizeiptrARB, GLvoid *);
VGLEXTAPI GLvoid* (VGLEXTAPIENTRY *glMapBufferARB) (GLenum, GLenum);
VGLEXTAPI GLboolean (VGLEXTAPIENTRY *glUnmapBufferARB) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetBufferParameterivARB) (GLenum, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetBufferPointervARB) (GLenum, GLenum, GLvoid* *);
/* ARB_occlusion_query */
VGLEXTAPI void (VGLEXTAPIENTRY *glGenQueriesARB) (GLsizei, GLuint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glDeleteQueriesARB) (GLsizei, const GLuint *);
VGLEXTAPI GLboolean (VGLEXTAPIENTRY *glIsQueryARB) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glBeginQueryARB) (GLenum, GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glEndQueryARB) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetQueryivARB) (GLenum, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetQueryObjectivARB) (GLuint, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetQueryObjectuivARB) (GLuint, GLenum, GLuint *);
/* ARB_draw_buffers */
VGLEXTAPI void (VGLEXTAPIENTRY *glDrawBuffersARB) (GLsizei, const GLenum *);

#endif	/* !DLL_TARGET_DCC */

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Miscellaneous extensions.
 * Promoted to GL 1.2:
 *  - EXT_texture3D;
 *  - EXT_bgra, EXT_abgr;
 *  - EXT_packed_pixels;
 *  - EXT_rescale_normal;
 *  - EXT_separate_specular_color;
 *  - EXT_draw_range_elements;
 *  - SGIS_texture_lod.
 *  - Imaging: EXT_blend_color, EXT_blend_minmax, EXT_blend_subtract.
 * Promoted to GL 1.4:
 *  - SGIS_generate_mipmap;
 *  - NV_blend_square;
 *  - EXT_fog_coord;
 *  - EXT_multi_draw_arrays;
 *  - EXT_secondary_color;
 *  - EXT_blend_func_separate;
 *  - EXT_stencil_wrap;
 *  - EXT_texture_lod_bias.
 * Promoted to GL 1.4:
 *  - EXT_shadow_funcs.
 * Promoted to GL 2.0:
 *  - EXT_stencil_two_side.
 * Other:
 */

/*
 * Tokens
 */

/* EXT_texture3D */
#define GL_PACK_SKIP_IMAGES_EXT           0x806B
#define GL_PACK_IMAGE_HEIGHT_EXT          0x806C
#define GL_UNPACK_SKIP_IMAGES_EXT         0x806D
#define GL_UNPACK_IMAGE_HEIGHT_EXT        0x806E
#define GL_TEXTURE_3D_EXT                 0x806F
#define GL_PROXY_TEXTURE_3D_EXT           0x8070
#define GL_TEXTURE_DEPTH_EXT              0x8071
#define GL_TEXTURE_WRAP_R_EXT             0x8072
#define GL_MAX_3D_TEXTURE_SIZE_EXT        0x8073
/* EXT_bgra, EXT_abgr */
#define GL_BGR_EXT                        0x80E0
#define GL_BGRA_EXT                       0x80E1
#define GL_ABGR_EXT                       0x8000
/* EXT_packed_pixels */
#define GL_UNSIGNED_BYTE_3_3_2_EXT        0x8032
#define GL_UNSIGNED_SHORT_4_4_4_4_EXT     0x8033
#define GL_UNSIGNED_SHORT_5_5_5_1_EXT     0x8034
#define GL_UNSIGNED_INT_8_8_8_8_EXT       0x8035
#define GL_UNSIGNED_INT_10_10_10_2_EXT    0x8036
/* EXT_rescale_normal */
#define GL_RESCALE_NORMAL_EXT             0x803A
/* EXT_separate_specular_color */
#define GL_LIGHT_MODEL_COLOR_CONTROL_EXT  0x81F8
#define GL_SINGLE_COLOR_EXT               0x81F9
#define GL_SEPARATE_SPECULAR_COLOR_EXT    0x81FA
/* EXT_draw_range_elements */
#define GL_MAX_ELEMENTS_VERTICES_EXT      0x80E8
#define GL_MAX_ELEMENTS_INDICES_EXT       0x80E9
/* SGIS_texture_lod */
#define GL_TEXTURE_MIN_LOD_SGIS           0x813A
#define GL_TEXTURE_MAX_LOD_SGIS           0x813B
#define GL_TEXTURE_BASE_LEVEL_SGIS        0x813C
#define GL_TEXTURE_MAX_LEVEL_SGIS         0x813D
/* EXT_blend_color */
#define GL_CONSTANT_COLOR_EXT             0x8001
#define GL_ONE_MINUS_CONSTANT_COLOR_EXT   0x8002
#define GL_CONSTANT_ALPHA_EXT             0x8003
#define GL_ONE_MINUS_CONSTANT_ALPHA_EXT   0x8004
#define GL_BLEND_COLOR_EXT                0x8005
/* EXT_blend_minmax */
#define GL_FUNC_ADD_EXT                   0x8006
#define GL_MIN_EXT                        0x8007
#define GL_MAX_EXT                        0x8008
#define GL_BLEND_EQUATION_EXT             0x8009
/* EXT_blend_subtract */
#define GL_FUNC_SUBTRACT_EXT              0x800A
#define GL_FUNC_REVERSE_SUBTRACT_EXT      0x800B
/* SGIS_generate_mipmap */
#define GL_GENERATE_MIPMAP_SGIS           0x8191
#define GL_GENERATE_MIPMAP_HINT_SGIS      0x8192
/* NV_blend_square */
/* EXT_fog_coord */
#define GL_FOG_COORDINATE_SOURCE_EXT      0x8450
#define GL_FOG_COORDINATE_EXT             0x8451
#define GL_FRAGMENT_DEPTH_EXT             0x8452
#define GL_CURRENT_FOG_COORDINATE_EXT     0x8453
#define GL_FOG_COORDINATE_ARRAY_TYPE_EXT  0x8454
#define GL_FOG_COORDINATE_ARRAY_STRIDE_EXT 0x8455
#define GL_FOG_COORDINATE_ARRAY_POINTER_EXT 0x8456
#define GL_FOG_COORDINATE_ARRAY_EXT       0x8457
/* EXT_multi_draw_arrays */
/* EXT_secondary_color */
#define GL_COLOR_SUM_EXT                  0x8458
#define GL_CURRENT_SECONDARY_COLOR_EXT    0x8459
#define GL_SECONDARY_COLOR_ARRAY_SIZE_EXT 0x845A
#define GL_SECONDARY_COLOR_ARRAY_TYPE_EXT 0x845B
#define GL_SECONDARY_COLOR_ARRAY_STRIDE_EXT 0x845C
#define GL_SECONDARY_COLOR_ARRAY_POINTER_EXT 0x845D
#define GL_SECONDARY_COLOR_ARRAY_EXT      0x845E
/* EXT_blend_func_separate */
#define GL_BLEND_DST_RGB_EXT              0x80C8
#define GL_BLEND_SRC_RGB_EXT              0x80C9
#define GL_BLEND_DST_ALPHA_EXT            0x80CA
#define GL_BLEND_SRC_ALPHA_EXT            0x80CB
/* EXT_stencil_wrap */
#define GL_INCR_WRAP_EXT                  0x8507
#define GL_DECR_WRAP_EXT                  0x8508
/* EXT_texture_lod_bias */
#define GL_MAX_TEXTURE_LOD_BIAS_EXT       0x84FD
#define GL_TEXTURE_FILTER_CONTROL_EXT     0x8500
#define GL_TEXTURE_LOD_BIAS_EXT           0x8501
/* EXT_shadow_funcs */
/* EXT_stencil_two_side */
#define GL_STENCIL_TEST_TWO_SIDE_EXT      0x8910
#define GL_ACTIVE_STENCIL_FACE_EXT        0x8911

#ifdef	DLL_TARGET_DCC

#ifdef	__cplusplus
extern "C" {
#endif /* __cplusplus */

/*
 * Prototypes
 */

/* EXT_texture3D */
VGLEXTAPI void VGLEXTAPIENTRY glTexImage3DEXT (GLenum, GLint, GLint, GLsizei, GLsizei, GLsizei, GLint, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glTexSubImage3DEXT (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *);
/* EXT_draw_range_elements */
VGLEXTAPI void VGLEXTAPIENTRY glDrawRangeElementsEXT (GLenum, GLuint, GLuint, GLsizei, GLenum, const GLvoid *);
/* EXT_blend_color */
VGLEXTAPI void VGLEXTAPIENTRY glBlendColorEXT (GLclampf, GLclampf, GLclampf, GLclampf);
/* EXT_blend_minmax */
VGLEXTAPI void VGLEXTAPIENTRY glBlendEquationEXT (GLenum);
/* EXT_fog_coord */
VGLEXTAPI void VGLEXTAPIENTRY glFogCoordfEXT (GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glFogCoordfvEXT (const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glFogCoorddEXT (GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glFogCoorddvEXT (const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glFogCoordPointerEXT (GLenum, GLsizei, const GLvoid *);
/* EXT_multi_draw_arrays */
VGLEXTAPI void VGLEXTAPIENTRY glMultiDrawArraysEXT (GLenum, GLint *, GLsizei *, GLsizei);
VGLEXTAPI void VGLEXTAPIENTRY glMultiDrawElementsEXT (GLenum, const GLsizei *, GLenum, const GLvoid* *, GLsizei);
/* EXT_secondary_color */
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3bEXT (GLbyte, GLbyte, GLbyte);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3bvEXT (const GLbyte *);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3dEXT (GLdouble, GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3dvEXT (const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3fEXT (GLfloat, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3fvEXT (const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3iEXT (GLint, GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3ivEXT (const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3sEXT (GLshort, GLshort, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3svEXT (const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3ubEXT (GLubyte, GLubyte, GLubyte);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3ubvEXT (const GLubyte *);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3uiEXT (GLuint, GLuint, GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3uivEXT (const GLuint *);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3usEXT (GLushort, GLushort, GLushort);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3usvEXT (const GLushort *);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColorPointerEXT (GLint, GLenum, GLsizei, const GLvoid *);
/* EXT_blend_func_separate */
VGLEXTAPI void VGLEXTAPIENTRY glBlendFuncSeparateEXT (GLenum, GLenum, GLenum, GLenum);
/* EXT_stencil_two_side */
VGLEXTAPI void VGLEXTAPIENTRY glActiveStencilFaceEXT (GLenum);

#ifdef	__cplusplus
}
#endif /* __cplusplus */

#else	/* !DLL_TARGET_DCC */

/*
 * Pointers
 */

/* EXT_texture3D */
VGLEXTAPI void (VGLEXTAPIENTRY *glTexImage3DEXT) (GLenum, GLint, GLint, GLsizei, GLsizei, GLsizei, GLint, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexSubImage3DEXT) (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *);
/* EXT_draw_range_elements */
VGLEXTAPI void (VGLEXTAPIENTRY *glDrawRangeElementsEXT) (GLenum, GLuint, GLuint, GLsizei, GLenum, const GLvoid *);
/* EXT_blend_color */
VGLEXTAPI void (VGLEXTAPIENTRY *glBlendColorEXT) (GLclampf, GLclampf, GLclampf, GLclampf);
/* EXT_blend_minmax */
VGLEXTAPI void (VGLEXTAPIENTRY *glBlendEquationEXT) (GLenum);
/* EXT_fog_coord */
VGLEXTAPI void (VGLEXTAPIENTRY *glFogCoordfEXT) (GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glFogCoordfvEXT) (const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glFogCoorddEXT) (GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glFogCoorddvEXT) (const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glFogCoordPointerEXT) (GLenum, GLsizei, const GLvoid *);
/* EXT_multi_draw_arrays */
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiDrawArraysEXT) (GLenum, GLint *, GLsizei *, GLsizei);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiDrawElementsEXT) (GLenum, const GLsizei *, GLenum, const GLvoid* *, GLsizei);
/* EXT_secondary_color */
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3bEXT) (GLbyte, GLbyte, GLbyte);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3bvEXT) (const GLbyte *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3dEXT) (GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3dvEXT) (const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3fEXT) (GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3fvEXT) (const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3iEXT) (GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3ivEXT) (const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3sEXT) (GLshort, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3svEXT) (const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3ubEXT) (GLubyte, GLubyte, GLubyte);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3ubvEXT) (const GLubyte *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3uiEXT) (GLuint, GLuint, GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3uivEXT) (const GLuint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3usEXT) (GLushort, GLushort, GLushort);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3usvEXT) (const GLushort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColorPointerEXT) (GLint, GLenum, GLsizei, const GLvoid *);
/* EXT_blend_func_separate */
VGLEXTAPI void (VGLEXTAPIENTRY *glBlendFuncSeparateEXT) (GLenum, GLenum, GLenum, GLenum);
/* EXT_stencil_two_side */
VGLEXTAPI void (VGLEXTAPIENTRY *glActiveStencilFaceEXT) (GLenum);

#endif	/* !DLL_TARGET_DCC */


/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Version 1.2
 * New stuff:
 *  - Three-Dimensional Texturing (replaces EXT_texture3D);
 *  - BGRA Pixel Formats (replaces EXT_bgra);
 *  - Packed Pixel Formats (replaces EXT_packed_pixels);
 *  - Normal Rescaling (replaces EXT_rescale_normal);
 *  - Separate Specular Color (EXT_separate_specular_color);
 *  - Texture Coordinate Edge Clamping (replaces SGIS_texture_edge_clamp);
 *  - Texture Level of Detail Control (replaces SGIS_texture_lod);
 *  - Vertex Array Draw Element Range (replaces EXT_draw_range_elements);
 *  - Imaging Subset (former ARB_imaging):
 *    - Color Tables (replaces EXT_color_table);
 *    - Convolution (replaces EXT_convolution and HP_convolution_border_modes);
 *    - Color Matrix (replaces SGI_color_matrix)
 *    - Pixel Pipeline Statistics (replaces EXT_histogram);
 *    - Constant Blend Color (replaces EXT_blend_color)
 *    - New Blending Equations (replaces EXT_blend_minmax and EXT_blend_subtract).
 */

/*
 * Tokens
 */

#define GL_UNSIGNED_BYTE_3_3_2            0x8032
#define GL_UNSIGNED_SHORT_4_4_4_4         0x8033
#define GL_UNSIGNED_SHORT_5_5_5_1         0x8034
#define GL_UNSIGNED_INT_8_8_8_8           0x8035
#define GL_UNSIGNED_INT_10_10_10_2        0x8036
#define GL_RESCALE_NORMAL                 0x803A
#define GL_TEXTURE_BINDING_3D             0x806A
#define GL_PACK_SKIP_IMAGES               0x806B
#define GL_PACK_IMAGE_HEIGHT              0x806C
#define GL_UNPACK_SKIP_IMAGES             0x806D
#define GL_UNPACK_IMAGE_HEIGHT            0x806E
#define GL_TEXTURE_3D                     0x806F
#define GL_PROXY_TEXTURE_3D               0x8070
#define GL_TEXTURE_DEPTH                  0x8071
#define GL_TEXTURE_WRAP_R                 0x8072
#define GL_MAX_3D_TEXTURE_SIZE            0x8073
#define GL_UNSIGNED_BYTE_2_3_3_REV        0x8362
#define GL_UNSIGNED_SHORT_5_6_5           0x8363
#define GL_UNSIGNED_SHORT_5_6_5_REV       0x8364
#define GL_UNSIGNED_SHORT_4_4_4_4_REV     0x8365
#define GL_UNSIGNED_SHORT_1_5_5_5_REV     0x8366
#define GL_UNSIGNED_INT_8_8_8_8_REV       0x8367
#define GL_UNSIGNED_INT_2_10_10_10_REV    0x8368
#define GL_BGR                            0x80E0
#define GL_BGRA                           0x80E1
#define GL_MAX_ELEMENTS_VERTICES          0x80E8
#define GL_MAX_ELEMENTS_INDICES           0x80E9
#define GL_CLAMP_TO_EDGE                  0x812F
#define GL_TEXTURE_MIN_LOD                0x813A
#define GL_TEXTURE_MAX_LOD                0x813B
#define GL_TEXTURE_BASE_LEVEL             0x813C
#define GL_TEXTURE_MAX_LEVEL              0x813D
#define GL_LIGHT_MODEL_COLOR_CONTROL      0x81F8
#define GL_SINGLE_COLOR                   0x81F9
#define GL_SEPARATE_SPECULAR_COLOR        0x81FA
#define GL_SMOOTH_POINT_SIZE_RANGE        0x0B12
#define GL_SMOOTH_POINT_SIZE_GRANULARITY  0x0B13
#define GL_SMOOTH_LINE_WIDTH_RANGE        0x0B22
#define GL_SMOOTH_LINE_WIDTH_GRANULARITY  0x0B23
#define GL_ALIASED_POINT_SIZE_RANGE       0x846D
#define GL_ALIASED_LINE_WIDTH_RANGE       0x846E

#ifdef	DLL_TARGET_DCC

#ifdef	__cplusplus
extern "C" {
#endif /* __cplusplus */

/*
 * Prototypes
 */
VGLEXTAPI void VGLEXTAPIENTRY glBlendColor (GLclampf, GLclampf, GLclampf, GLclampf);
VGLEXTAPI void VGLEXTAPIENTRY glBlendEquation (GLenum);
VGLEXTAPI void VGLEXTAPIENTRY glDrawRangeElements (GLenum, GLuint, GLuint, GLsizei, GLenum, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glColorTable (GLenum, GLenum, GLsizei, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glColorTableParameterfv (GLenum, GLenum, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glColorTableParameteriv (GLenum, GLenum, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glCopyColorTable (GLenum, GLenum, GLint, GLint, GLsizei);
VGLEXTAPI void VGLEXTAPIENTRY glGetColorTable (GLenum, GLenum, GLenum, GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glGetColorTableParameterfv (GLenum, GLenum, GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glGetColorTableParameteriv (GLenum, GLenum, GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glColorSubTable (GLenum, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glCopyColorSubTable (GLenum, GLsizei, GLint, GLint, GLsizei);
VGLEXTAPI void VGLEXTAPIENTRY glConvolutionFilter1D (GLenum, GLenum, GLsizei, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glConvolutionFilter2D (GLenum, GLenum, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glConvolutionParameterf (GLenum, GLenum, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glConvolutionParameterfv (GLenum, GLenum, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glConvolutionParameteri (GLenum, GLenum, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glConvolutionParameteriv (GLenum, GLenum, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glCopyConvolutionFilter1D (GLenum, GLenum, GLint, GLint, GLsizei);
VGLEXTAPI void VGLEXTAPIENTRY glCopyConvolutionFilter2D (GLenum, GLenum, GLint, GLint, GLsizei, GLsizei);
VGLEXTAPI void VGLEXTAPIENTRY glGetConvolutionFilter (GLenum, GLenum, GLenum, GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glGetConvolutionParameterfv (GLenum, GLenum, GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glGetConvolutionParameteriv (GLenum, GLenum, GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glGetSeparableFilter (GLenum, GLenum, GLenum, GLvoid *, GLvoid *, GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glSeparableFilter2D (GLenum, GLenum, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glGetHistogram (GLenum, GLboolean, GLenum, GLenum, GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glGetHistogramParameterfv (GLenum, GLenum, GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glGetHistogramParameteriv (GLenum, GLenum, GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glGetMinmax (GLenum, GLboolean, GLenum, GLenum, GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glGetMinmaxParameterfv (GLenum, GLenum, GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glGetMinmaxParameteriv (GLenum, GLenum, GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glHistogram (GLenum, GLsizei, GLenum, GLboolean);
VGLEXTAPI void VGLEXTAPIENTRY glMinmax (GLenum, GLenum, GLboolean);
VGLEXTAPI void VGLEXTAPIENTRY glResetHistogram (GLenum);
VGLEXTAPI void VGLEXTAPIENTRY glResetMinmax (GLenum);
VGLEXTAPI void VGLEXTAPIENTRY glTexImage3D (GLenum, GLint, GLint, GLsizei, GLsizei, GLsizei, GLint, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glTexSubImage3D (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glCopyTexSubImage3D (GLenum, GLint, GLint, GLint, GLint, GLint, GLint, GLsizei, GLsizei);

#ifdef	__cplusplus
}
#endif /* __cplusplus */

#else	/* !DLL_TARGET_DCC */

/*
 * Pointers
 */
VGLEXTAPI void (VGLEXTAPIENTRY *glBlendColor) (GLclampf, GLclampf, GLclampf, GLclampf);
VGLEXTAPI void (VGLEXTAPIENTRY *glBlendEquation) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glDrawRangeElements) (GLenum, GLuint, GLuint, GLsizei, GLenum, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glColorTable) (GLenum, GLenum, GLsizei, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glColorTableParameterfv) (GLenum, GLenum, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glColorTableParameteriv) (GLenum, GLenum, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCopyColorTable) (GLenum, GLenum, GLint, GLint, GLsizei);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetColorTable) (GLenum, GLenum, GLenum, GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetColorTableParameterfv) (GLenum, GLenum, GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetColorTableParameteriv) (GLenum, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glColorSubTable) (GLenum, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCopyColorSubTable) (GLenum, GLsizei, GLint, GLint, GLsizei);
VGLEXTAPI void (VGLEXTAPIENTRY *glConvolutionFilter1D) (GLenum, GLenum, GLsizei, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glConvolutionFilter2D) (GLenum, GLenum, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glConvolutionParameterf) (GLenum, GLenum, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glConvolutionParameterfv) (GLenum, GLenum, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glConvolutionParameteri) (GLenum, GLenum, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glConvolutionParameteriv) (GLenum, GLenum, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCopyConvolutionFilter1D) (GLenum, GLenum, GLint, GLint, GLsizei);
VGLEXTAPI void (VGLEXTAPIENTRY *glCopyConvolutionFilter2D) (GLenum, GLenum, GLint, GLint, GLsizei, GLsizei);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetConvolutionFilter) (GLenum, GLenum, GLenum, GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetConvolutionParameterfv) (GLenum, GLenum, GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetConvolutionParameteriv) (GLenum, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetSeparableFilter) (GLenum, GLenum, GLenum, GLvoid *, GLvoid *, GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSeparableFilter2D) (GLenum, GLenum, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetHistogram) (GLenum, GLboolean, GLenum, GLenum, GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetHistogramParameterfv) (GLenum, GLenum, GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetHistogramParameteriv) (GLenum, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetMinmax) (GLenum, GLboolean, GLenum, GLenum, GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetMinmaxParameterfv) (GLenum, GLenum, GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetMinmaxParameteriv) (GLenum, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glHistogram) (GLenum, GLsizei, GLenum, GLboolean);
VGLEXTAPI void (VGLEXTAPIENTRY *glMinmax) (GLenum, GLenum, GLboolean);
VGLEXTAPI void (VGLEXTAPIENTRY *glResetHistogram) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glResetMinmax) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexImage3D) (GLenum, GLint, GLint, GLsizei, GLsizei, GLsizei, GLint, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glTexSubImage3D) (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLenum, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCopyTexSubImage3D) (GLenum, GLint, GLint, GLint, GLint, GLint, GLint, GLsizei, GLsizei);

#endif	/* !DLL_TARGET_DCC */

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Version 1.3
 * New stuff:
 *  - Compressed Textures (replaces ARB_texture_compression)
 *  - Cube Map Textures (replaces ARB_texture_cube_map)
 *  - Multisample (replaces ARB_multisample)
 *  - Multitexture (replaces GL 1.2.1 ARB_multitexture)
 *  - Texture Add Environment Mode (replaces ARB_texture_env_add)
 *  - Texture Combine Environment Mode (replaces ARB_texture_env_combine)
 *  - Texture Dot3 Environment Mode (replaces ARB_texture_env_dot3)
 *  - Texture Border Clamp (replaces ARB_texture_border_clamp)
 *  - Transpose Matrix (replaces ARB_transpose_matrix)
 */

/*
 * Tokens
 */
#define GL_TEXTURE0                       0x84C0
#define GL_TEXTURE1                       0x84C1
#define GL_TEXTURE2                       0x84C2
#define GL_TEXTURE3                       0x84C3
#define GL_TEXTURE4                       0x84C4
#define GL_TEXTURE5                       0x84C5
#define GL_TEXTURE6                       0x84C6
#define GL_TEXTURE7                       0x84C7
#define GL_TEXTURE8                       0x84C8
#define GL_TEXTURE9                       0x84C9
#define GL_TEXTURE10                      0x84CA
#define GL_TEXTURE11                      0x84CB
#define GL_TEXTURE12                      0x84CC
#define GL_TEXTURE13                      0x84CD
#define GL_TEXTURE14                      0x84CE
#define GL_TEXTURE15                      0x84CF
#define GL_TEXTURE16                      0x84D0
#define GL_TEXTURE17                      0x84D1
#define GL_TEXTURE18                      0x84D2
#define GL_TEXTURE19                      0x84D3
#define GL_TEXTURE20                      0x84D4
#define GL_TEXTURE21                      0x84D5
#define GL_TEXTURE22                      0x84D6
#define GL_TEXTURE23                      0x84D7
#define GL_TEXTURE24                      0x84D8
#define GL_TEXTURE25                      0x84D9
#define GL_TEXTURE26                      0x84DA
#define GL_TEXTURE27                      0x84DB
#define GL_TEXTURE28                      0x84DC
#define GL_TEXTURE29                      0x84DD
#define GL_TEXTURE30                      0x84DE
#define GL_TEXTURE31                      0x84DF
#define GL_ACTIVE_TEXTURE                 0x84E0
#define GL_CLIENT_ACTIVE_TEXTURE          0x84E1
#define GL_MAX_TEXTURE_UNITS              0x84E2
#define GL_TRANSPOSE_MODELVIEW_MATRIX     0x84E3
#define GL_TRANSPOSE_PROJECTION_MATRIX    0x84E4
#define GL_TRANSPOSE_TEXTURE_MATRIX       0x84E5
#define GL_TRANSPOSE_COLOR_MATRIX         0x84E6
#define GL_MULTISAMPLE                    0x809D
#define GL_SAMPLE_ALPHA_TO_COVERAGE       0x809E
#define GL_SAMPLE_ALPHA_TO_ONE            0x809F
#define GL_SAMPLE_COVERAGE                0x80A0
#define GL_SAMPLE_BUFFERS                 0x80A8
#define GL_SAMPLES                        0x80A9
#define GL_SAMPLE_COVERAGE_VALUE          0x80AA
#define GL_SAMPLE_COVERAGE_INVERT         0x80AB
#define GL_MULTISAMPLE_BIT                0x20000000
#define GL_NORMAL_MAP                     0x8511
#define GL_REFLECTION_MAP                 0x8512
#define GL_TEXTURE_CUBE_MAP               0x8513
#define GL_TEXTURE_BINDING_CUBE_MAP       0x8514
#define GL_TEXTURE_CUBE_MAP_POSITIVE_X    0x8515
#define GL_TEXTURE_CUBE_MAP_NEGATIVE_X    0x8516
#define GL_TEXTURE_CUBE_MAP_POSITIVE_Y    0x8517
#define GL_TEXTURE_CUBE_MAP_NEGATIVE_Y    0x8518
#define GL_TEXTURE_CUBE_MAP_POSITIVE_Z    0x8519
#define GL_TEXTURE_CUBE_MAP_NEGATIVE_Z    0x851A
#define GL_PROXY_TEXTURE_CUBE_MAP         0x851B
#define GL_MAX_CUBE_MAP_TEXTURE_SIZE      0x851C
#define GL_COMPRESSED_ALPHA               0x84E9
#define GL_COMPRESSED_LUMINANCE           0x84EA
#define GL_COMPRESSED_LUMINANCE_ALPHA     0x84EB
#define GL_COMPRESSED_INTENSITY           0x84EC
#define GL_COMPRESSED_RGB                 0x84ED
#define GL_COMPRESSED_RGBA                0x84EE
#define GL_TEXTURE_COMPRESSION_HINT       0x84EF
#define GL_TEXTURE_COMPRESSED_IMAGE_SIZE  0x86A0
#define GL_TEXTURE_COMPRESSED             0x86A1
#define GL_NUM_COMPRESSED_TEXTURE_FORMATS 0x86A2
#define GL_COMPRESSED_TEXTURE_FORMATS     0x86A3
#define GL_CLAMP_TO_BORDER                0x812D
#define GL_COMBINE                        0x8570
#define GL_COMBINE_RGB                    0x8571
#define GL_COMBINE_ALPHA                  0x8572
#define GL_SOURCE0_RGB                    0x8580
#define GL_SOURCE1_RGB                    0x8581
#define GL_SOURCE2_RGB                    0x8582
#define GL_SOURCE0_ALPHA                  0x8588
#define GL_SOURCE1_ALPHA                  0x8589
#define GL_SOURCE2_ALPHA                  0x858A
#define GL_OPERAND0_RGB                   0x8590
#define GL_OPERAND1_RGB                   0x8591
#define GL_OPERAND2_RGB                   0x8592
#define GL_OPERAND0_ALPHA                 0x8598
#define GL_OPERAND1_ALPHA                 0x8599
#define GL_OPERAND2_ALPHA                 0x859A
#define GL_RGB_SCALE                      0x8573
#define GL_ADD_SIGNED                     0x8574
#define GL_INTERPOLATE                    0x8575
#define GL_SUBTRACT                       0x84E7
#define GL_CONSTANT                       0x8576
#define GL_PRIMARY_COLOR                  0x8577
#define GL_PREVIOUS                       0x8578
#define GL_DOT3_RGB                       0x86AE
#define GL_DOT3_RGBA                      0x86AF

#ifdef	DLL_TARGET_DCC

#ifdef	__cplusplus
extern "C" {
#endif /* __cplusplus */

/*
 * Prototypes
 */
VGLEXTAPI void VGLEXTAPIENTRY glActiveTexture (GLenum);
VGLEXTAPI void VGLEXTAPIENTRY glClientActiveTexture (GLenum);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord1d (GLenum, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord1dv (GLenum, const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord1f (GLenum, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord1fv (GLenum, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord1i (GLenum, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord1iv (GLenum, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord1s (GLenum, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord1sv (GLenum, const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord2d (GLenum, GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord2dv (GLenum, const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord2f (GLenum, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord2fv (GLenum, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord2i (GLenum, GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord2iv (GLenum, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord2s (GLenum, GLshort, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord2sv (GLenum, const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord3d (GLenum, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord3dv (GLenum, const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord3f (GLenum, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord3fv (GLenum, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord3i (GLenum, GLint, GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord3iv (GLenum, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord3s (GLenum, GLshort, GLshort, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord3sv (GLenum, const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord4d (GLenum, GLdouble, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord4dv (GLenum, const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord4f (GLenum, GLfloat, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord4fv (GLenum, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord4i (GLenum, GLint, GLint, GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord4iv (GLenum, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord4s (GLenum, GLshort, GLshort, GLshort, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glMultiTexCoord4sv (GLenum, const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glLoadTransposeMatrixf (const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glLoadTransposeMatrixd (const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glMultTransposeMatrixf (const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glMultTransposeMatrixd (const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glSampleCoverage (GLclampf, GLboolean);
VGLEXTAPI void VGLEXTAPIENTRY glCompressedTexImage3D (GLenum, GLint, GLenum, GLsizei, GLsizei, GLsizei, GLint, GLsizei, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glCompressedTexImage2D (GLenum, GLint, GLenum, GLsizei, GLsizei, GLint, GLsizei, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glCompressedTexImage1D (GLenum, GLint, GLenum, GLsizei, GLint, GLsizei, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glCompressedTexSubImage3D (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glCompressedTexSubImage2D (GLenum, GLint, GLint, GLint, GLsizei, GLsizei, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glCompressedTexSubImage1D (GLenum, GLint, GLint, GLsizei, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glGetCompressedTexImage (GLenum, GLint, GLvoid *);

#ifdef	__cplusplus
}
#endif /* __cplusplus */

#else	/* !DLL_TARGET_DCC */

/*
 * Pointers
 */
VGLEXTAPI void (VGLEXTAPIENTRY *glActiveTexture) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glClientActiveTexture) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1d) (GLenum, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1dv) (GLenum, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1f) (GLenum, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1fv) (GLenum, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1i) (GLenum, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1iv) (GLenum, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1s) (GLenum, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord1sv) (GLenum, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2d) (GLenum, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2dv) (GLenum, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2f) (GLenum, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2fv) (GLenum, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2i) (GLenum, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2iv) (GLenum, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2s) (GLenum, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord2sv) (GLenum, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3d) (GLenum, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3dv) (GLenum, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3f) (GLenum, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3fv) (GLenum, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3i) (GLenum, GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3iv) (GLenum, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3s) (GLenum, GLshort, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord3sv) (GLenum, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4d) (GLenum, GLdouble, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4dv) (GLenum, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4f) (GLenum, GLfloat, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4fv) (GLenum, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4i) (GLenum, GLint, GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4iv) (GLenum, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4s) (GLenum, GLshort, GLshort, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiTexCoord4sv) (GLenum, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glLoadTransposeMatrixf) (const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glLoadTransposeMatrixd) (const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultTransposeMatrixf) (const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultTransposeMatrixd) (const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSampleCoverage) (GLclampf, GLboolean);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompressedTexImage3D) (GLenum, GLint, GLenum, GLsizei, GLsizei, GLsizei, GLint, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompressedTexImage2D) (GLenum, GLint, GLenum, GLsizei, GLsizei, GLint, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompressedTexImage1D) (GLenum, GLint, GLenum, GLsizei, GLint, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompressedTexSubImage3D) (GLenum, GLint, GLint, GLint, GLint, GLsizei, GLsizei, GLsizei, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompressedTexSubImage2D) (GLenum, GLint, GLint, GLint, GLsizei, GLsizei, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompressedTexSubImage1D) (GLenum, GLint, GLint, GLsizei, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetCompressedTexImage) (GLenum, GLint, GLvoid *);

#endif	/* !DLL_TARGET_DCC */

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Version 1.4
 * New stuff:
 *  - Vertex programs (replaces ARB_vertex_program)
 *  - Fragment programs (replaces ARB_fragment_program and ARB_fragment_program_shadow)
 *  - Automatic Mipmap Generation (replaces SGIS_generate_mipmap)
 *  - Blend Squaring (replaces NV_blend_square)
 *  - Changes to the Imaging Subset (replaces EXT_blend_color)
 *  - Depth Textures and Shadows (replaces ARB_depth_texture and ARB_shadow)
 *  - Fog Coordinate (replaces EXT_fog_coord)
 *  - Multiple Draw Arrays (replaces EXT_multi_draw_arrays)
 *  - Point Parameters (replaces ARB_point_parameters)
 *  - Secondary Color (replaces EXT_secondary_color)
 *  - Separate Blend Functions (replaces EXT_blend_func_separate)
 *  - Stencil Wrap (replaces EXT_stencil_wrap)
 *  - Texture Crossbar Environment Mode (replaces ARB_texture_env_crossbar)
 *  - Texture LOD Bias (replaces EXT_texture_lod_bias)
 *  - Texture Mirrored Repeat (replaces ARB_texture_mirrored_repeat)
 *  - Window Raster Position (replaces ARB_window_pos)
 */

/*
 * Tokens
 */
#define GL_BLEND_DST_RGB                  0x80C8
#define GL_BLEND_SRC_RGB                  0x80C9
#define GL_BLEND_DST_ALPHA                0x80CA
#define GL_BLEND_SRC_ALPHA                0x80CB
#define GL_POINT_SIZE_MIN                 0x8126
#define GL_POINT_SIZE_MAX                 0x8127
#define GL_POINT_FADE_THRESHOLD_SIZE      0x8128
#define GL_POINT_DISTANCE_ATTENUATION     0x8129
#define GL_GENERATE_MIPMAP                0x8191
#define GL_GENERATE_MIPMAP_HINT           0x8192
#define GL_DEPTH_COMPONENT16              0x81A5
#define GL_DEPTH_COMPONENT24              0x81A6
#define GL_DEPTH_COMPONENT32              0x81A7
#define GL_MIRRORED_REPEAT                0x8370
#define GL_FOG_COORDINATE_SOURCE          0x8450
#define GL_FOG_COORDINATE                 0x8451
#define GL_FRAGMENT_DEPTH                 0x8452
#define GL_CURRENT_FOG_COORDINATE         0x8453
#define GL_FOG_COORDINATE_ARRAY_TYPE      0x8454
#define GL_FOG_COORDINATE_ARRAY_STRIDE    0x8455
#define GL_FOG_COORDINATE_ARRAY_POINTER   0x8456
#define GL_FOG_COORDINATE_ARRAY           0x8457
#define GL_COLOR_SUM                      0x8458
#define GL_CURRENT_SECONDARY_COLOR        0x8459
#define GL_SECONDARY_COLOR_ARRAY_SIZE     0x845A
#define GL_SECONDARY_COLOR_ARRAY_TYPE     0x845B
#define GL_SECONDARY_COLOR_ARRAY_STRIDE   0x845C
#define GL_SECONDARY_COLOR_ARRAY_POINTER  0x845D
#define GL_SECONDARY_COLOR_ARRAY          0x845E
#define GL_MAX_TEXTURE_LOD_BIAS           0x84FD
#define GL_TEXTURE_FILTER_CONTROL         0x8500
#define GL_TEXTURE_LOD_BIAS               0x8501
#define GL_INCR_WRAP                      0x8507
#define GL_DECR_WRAP                      0x8508
#define GL_TEXTURE_DEPTH_SIZE             0x884A
#define GL_DEPTH_TEXTURE_MODE             0x884B
#define GL_TEXTURE_COMPARE_MODE           0x884C
#define GL_TEXTURE_COMPARE_FUNC           0x884D
#define GL_COMPARE_R_TO_TEXTURE           0x884E

#ifdef	DLL_TARGET_DCC

#ifdef	__cplusplus
extern "C" {
#endif /* __cplusplus */

/*
 * Prototypes
 */
VGLEXTAPI void VGLEXTAPIENTRY glBlendFuncSeparate (GLenum, GLenum, GLenum, GLenum);
VGLEXTAPI void VGLEXTAPIENTRY glFogCoordf (GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glFogCoordfv (const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glFogCoordd (GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glFogCoorddv (const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glFogCoordPointer (GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glMultiDrawArrays (GLenum, GLint *, GLsizei *, GLsizei);
VGLEXTAPI void VGLEXTAPIENTRY glMultiDrawElements (GLenum, const GLsizei *, GLenum, const GLvoid* *, GLsizei);
VGLEXTAPI void VGLEXTAPIENTRY glPointParameterf (GLenum, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glPointParameterfv (GLenum, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glPointParameteri (GLenum, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glPointParameteriv (GLenum, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3b (GLbyte, GLbyte, GLbyte);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3bv (const GLbyte *);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3d (GLdouble, GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3dv (const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3f (GLfloat, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3fv (const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3i (GLint, GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3iv (const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3s (GLshort, GLshort, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3sv (const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3ub (GLubyte, GLubyte, GLubyte);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3ubv (const GLubyte *);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3ui (GLuint, GLuint, GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3uiv (const GLuint *);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3us (GLushort, GLushort, GLushort);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColor3usv (const GLushort *);
VGLEXTAPI void VGLEXTAPIENTRY glSecondaryColorPointer (GLint, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos2d (GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos2dv (const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos2f (GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos2fv (const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos2i (GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos2iv (const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos2s (GLshort, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos2sv (const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos3d (GLdouble, GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos3dv (const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos3f (GLfloat, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos3fv (const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos3i (GLint, GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos3iv (const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos3s (GLshort, GLshort, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glWindowPos3sv (const GLshort *);

#ifdef	__cplusplus
}
#endif /* __cplusplus */

#else	/* !DLL_TARGET_DCC */

/*
 * Pointers
 */
VGLEXTAPI void (VGLEXTAPIENTRY *glBlendFuncSeparate) (GLenum, GLenum, GLenum, GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glFogCoordf) (GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glFogCoordfv) (const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glFogCoordd) (GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glFogCoorddv) (const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glFogCoordPointer) (GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiDrawArrays) (GLenum, GLint *, GLsizei *, GLsizei);
VGLEXTAPI void (VGLEXTAPIENTRY *glMultiDrawElements) (GLenum, const GLsizei *, GLenum, const GLvoid* *, GLsizei);
VGLEXTAPI void (VGLEXTAPIENTRY *glPointParameterf) (GLenum, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glPointParameterfv) (GLenum, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glPointParameteri) (GLenum, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glPointParameteriv) (GLenum, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3b) (GLbyte, GLbyte, GLbyte);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3bv) (const GLbyte *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3d) (GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3dv) (const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3f) (GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3fv) (const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3i) (GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3iv) (const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3s) (GLshort, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3sv) (const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3ub) (GLubyte, GLubyte, GLubyte);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3ubv) (const GLubyte *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3ui) (GLuint, GLuint, GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3uiv) (const GLuint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3us) (GLushort, GLushort, GLushort);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColor3usv) (const GLushort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glSecondaryColorPointer) (GLint, GLenum, GLsizei, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2d) (GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2dv) (const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2f) (GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2fv) (const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2i) (GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2iv) (const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2s) (GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos2sv) (const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3d) (GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3dv) (const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3f) (GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3fv) (const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3i) (GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3iv) (const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3s) (GLshort, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glWindowPos3sv) (const GLshort *);

#endif	/* !DLL_TARGET_DCC */

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Version 1.5
 * New stuff:
 *  - Vertex Buffer Objects (replaces ARB_vertex_buffer_object)
 *  - Occlusion Queries (replaces ARB_occlusion_query)
 *  - Shadow Functions (replaces EXT_shadow_funcs)
 */

/*
 * Tokens
 */
#define GL_BUFFER_SIZE                    0x8764
#define GL_BUFFER_USAGE                   0x8765
#define GL_QUERY_COUNTER_BITS             0x8864
#define GL_CURRENT_QUERY                  0x8865
#define GL_QUERY_RESULT                   0x8866
#define GL_QUERY_RESULT_AVAILABLE         0x8867
#define GL_ARRAY_BUFFER                   0x8892
#define GL_ELEMENT_ARRAY_BUFFER           0x8893
#define GL_ARRAY_BUFFER_BINDING           0x8894
#define GL_ELEMENT_ARRAY_BUFFER_BINDING   0x8895
#define GL_VERTEX_ARRAY_BUFFER_BINDING    0x8896
#define GL_NORMAL_ARRAY_BUFFER_BINDING    0x8897
#define GL_COLOR_ARRAY_BUFFER_BINDING     0x8898
#define GL_INDEX_ARRAY_BUFFER_BINDING     0x8899
#define GL_TEXTURE_COORD_ARRAY_BUFFER_BINDING 0x889A
#define GL_EDGE_FLAG_ARRAY_BUFFER_BINDING 0x889B
#define GL_SECONDARY_COLOR_ARRAY_BUFFER_BINDING 0x889C
#define GL_FOG_COORDINATE_ARRAY_BUFFER_BINDING 0x889D
#define GL_WEIGHT_ARRAY_BUFFER_BINDING    0x889E
#define GL_VERTEX_ATTRIB_ARRAY_BUFFER_BINDING 0x889F
#define GL_READ_ONLY                      0x88B8
#define GL_WRITE_ONLY                     0x88B9
#define GL_READ_WRITE                     0x88BA
#define GL_BUFFER_ACCESS                  0x88BB
#define GL_BUFFER_MAPPED                  0x88BC
#define GL_BUFFER_MAP_POINTER             0x88BD
#define GL_STREAM_DRAW                    0x88E0
#define GL_STREAM_READ                    0x88E1
#define GL_STREAM_COPY                    0x88E2
#define GL_STATIC_DRAW                    0x88E4
#define GL_STATIC_READ                    0x88E5
#define GL_STATIC_COPY                    0x88E6
#define GL_DYNAMIC_DRAW                   0x88E8
#define GL_DYNAMIC_READ                   0x88E9
#define GL_DYNAMIC_COPY                   0x88EA
#define GL_SAMPLES_PASSED                 0x8914
/*
 * GL 1.5 changes of token names.
 */
#define	FOG_COORD_SRC		FOG_COORDINATE_SOURCE
#define	FOG_COORD			FOG_COORDINATE
#define	CURRENT_FOG_COORD	CURRENT_FOG_COORDINATE
#define	FOG_COORD_ARRAY_TYPE	FOG_COORDINATE_ARRAY TYPE
#define	FOG_COORD_ARRAY_STRIDE	FOG_COORDINATE_ARRAY_STRIDE
#define	FOG_COORD_ARRAY_POINTER	FOG_COORDINATE_ARRAY_POINTER
#define	FOG_COORD_ARRAY		FOG_COORDINATE_ARRAY
#define	FOG_COORD_ARRAY_BUFFER_BINDING	FOG_COORDINATE_ARRAY_BUFFER_BINDING
#define	SRC0_RGB			SOURCE0_RGB
#define	SRC1_RGB			SOURCE1_RGB
#define	SRC2_RGB			SOURCE2_RGB
#define	SRC0_ALPHA			SOURCE0_ALPHA
#define	SRC1_ALPHA			SOURCE1_ALPHA
#define	SRC2_ALPHA			SOURCE2_ALPHA

typedef ptrdiff_t GLintptr;
typedef ptrdiff_t GLsizeiptr;

#ifdef	DLL_TARGET_DCC

#ifdef	__cplusplus
extern "C" {
#endif /* __cplusplus */

/*
 * Prototypes
 */
VGLEXTAPI void VGLEXTAPIENTRY glGenQueries (GLsizei, GLuint *);
VGLEXTAPI void VGLEXTAPIENTRY glDeleteQueries (GLsizei, const GLuint *);
VGLEXTAPI GLboolean VGLEXTAPIENTRY glIsQuery (GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glBeginQuery (GLenum, GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glEndQuery (GLenum);
VGLEXTAPI void VGLEXTAPIENTRY glGetQueryiv (GLenum, GLenum, GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glGetQueryObjectiv (GLuint, GLenum, GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glGetQueryObjectuiv (GLuint, GLenum, GLuint *);
VGLEXTAPI void VGLEXTAPIENTRY glBindBuffer (GLenum, GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glDeleteBuffers (GLsizei, const GLuint *);
VGLEXTAPI void VGLEXTAPIENTRY glGenBuffers (GLsizei, GLuint *);
VGLEXTAPI GLboolean VGLEXTAPIENTRY glIsBuffer (GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glBufferData (GLenum, GLsizeiptr, const GLvoid *, GLenum);
VGLEXTAPI void VGLEXTAPIENTRY glBufferSubData (GLenum, GLintptr, GLsizeiptr, const GLvoid *);
VGLEXTAPI void VGLEXTAPIENTRY glGetBufferSubData (GLenum, GLintptr, GLsizeiptr, GLvoid *);
VGLEXTAPI GLvoid* VGLEXTAPIENTRY glMapBuffer (GLenum, GLenum);
VGLEXTAPI GLboolean VGLEXTAPIENTRY glUnmapBuffer (GLenum);
VGLEXTAPI void VGLEXTAPIENTRY glGetBufferParameteriv (GLenum, GLenum, GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glGetBufferPointerv (GLenum, GLenum, GLvoid* *);

#ifdef	__cplusplus
}
#endif /* __cplusplus */

#else	/* !DLL_TARGET_DCC */

/*
 * Pointers
 */
VGLEXTAPI void (VGLEXTAPIENTRY *glGenQueries) (GLsizei, GLuint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glDeleteQueries) (GLsizei, const GLuint *);
VGLEXTAPI GLboolean (VGLEXTAPIENTRY *glIsQuery) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glBeginQuery) (GLenum, GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glEndQuery) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetQueryiv) (GLenum, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetQueryObjectiv) (GLuint, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetQueryObjectuiv) (GLuint, GLenum, GLuint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glBindBuffer) (GLenum, GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glDeleteBuffers) (GLsizei, const GLuint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGenBuffers) (GLsizei, GLuint *);
VGLEXTAPI GLboolean (VGLEXTAPIENTRY *glIsBuffer) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glBufferData) (GLenum, GLsizeiptr, const GLvoid *, GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glBufferSubData) (GLenum, GLintptr, GLsizeiptr, const GLvoid *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetBufferSubData) (GLenum, GLintptr, GLsizeiptr, GLvoid *);
VGLEXTAPI GLvoid* (VGLEXTAPIENTRY *glMapBuffer) (GLenum, GLenum);
VGLEXTAPI GLboolean (VGLEXTAPIENTRY *glUnmapBuffer) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetBufferParameteriv) (GLenum, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetBufferPointerv) (GLenum, GLenum, GLvoid* *);

#endif	/* !DLL_TARGET_DCC */

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Version 2.0
 * New stuff:
 *  - Programmable Shading (replaces GL 1.5 ARB_shader_objects,
 *    ARB_vertex_shader, ARB_fragment_shader, ARB_shading_language_100);
 *  - Multiple Render Targets (replaces ARB_draw_buffers);
 *  - Non-Power-Of-Two Textures (replaces ARB_texture_non_power_of_two);
 *  - Point Sprites (replaces ARB_point_sprite);
 *  - Separate Stencil (replaces EXT_stencil_two_side).
 */

/*
 * Tokens
 */
#define GL_BLEND_EQUATION_RGB             GL_BLEND_EQUATION
#define GL_VERTEX_ATTRIB_ARRAY_ENABLED    0x8622
#define GL_VERTEX_ATTRIB_ARRAY_SIZE       0x8623
#define GL_VERTEX_ATTRIB_ARRAY_STRIDE     0x8624
#define GL_VERTEX_ATTRIB_ARRAY_TYPE       0x8625
#define GL_CURRENT_VERTEX_ATTRIB          0x8626
#define GL_VERTEX_PROGRAM_POINT_SIZE      0x8642
#define GL_VERTEX_PROGRAM_TWO_SIDE        0x8643
#define GL_VERTEX_ATTRIB_ARRAY_POINTER    0x8645
#define GL_STENCIL_BACK_FUNC              0x8800
#define GL_STENCIL_BACK_FAIL              0x8801
#define GL_STENCIL_BACK_PASS_DEPTH_FAIL   0x8802
#define GL_STENCIL_BACK_PASS_DEPTH_PASS   0x8803
#define GL_MAX_DRAW_BUFFERS               0x8824
#define GL_DRAW_BUFFER0                   0x8825
#define GL_DRAW_BUFFER1                   0x8826
#define GL_DRAW_BUFFER2                   0x8827
#define GL_DRAW_BUFFER3                   0x8828
#define GL_DRAW_BUFFER4                   0x8829
#define GL_DRAW_BUFFER5                   0x882A
#define GL_DRAW_BUFFER6                   0x882B
#define GL_DRAW_BUFFER7                   0x882C
#define GL_DRAW_BUFFER8                   0x882D
#define GL_DRAW_BUFFER9                   0x882E
#define GL_DRAW_BUFFER10                  0x882F
#define GL_DRAW_BUFFER11                  0x8830
#define GL_DRAW_BUFFER12                  0x8831
#define GL_DRAW_BUFFER13                  0x8832
#define GL_DRAW_BUFFER14                  0x8833
#define GL_DRAW_BUFFER15                  0x8834
#define GL_BLEND_EQUATION_ALPHA           0x883D
#define GL_POINT_SPRITE                   0x8861
#define GL_COORD_REPLACE                  0x8862
#define GL_MAX_VERTEX_ATTRIBS             0x8869
#define GL_VERTEX_ATTRIB_ARRAY_NORMALIZED 0x886A
#define GL_MAX_TEXTURE_COORDS             0x8871
#define GL_MAX_TEXTURE_IMAGE_UNITS        0x8872
#define GL_FRAGMENT_SHADER                0x8B30
#define GL_VERTEX_SHADER                  0x8B31
#define GL_MAX_FRAGMENT_UNIFORM_COMPONENTS 0x8B49
#define GL_MAX_VERTEX_UNIFORM_COMPONENTS  0x8B4A
#define GL_MAX_VARYING_FLOATS             0x8B4B
#define GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS 0x8B4C
#define GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS 0x8B4D
#define GL_SHADER_TYPE                    0x8B4F
#define GL_FLOAT_VEC2                     0x8B50
#define GL_FLOAT_VEC3                     0x8B51
#define GL_FLOAT_VEC4                     0x8B52
#define GL_INT_VEC2                       0x8B53
#define GL_INT_VEC3                       0x8B54
#define GL_INT_VEC4                       0x8B55
#define GL_BOOL                           0x8B56
#define GL_BOOL_VEC2                      0x8B57
#define GL_BOOL_VEC3                      0x8B58
#define GL_BOOL_VEC4                      0x8B59
#define GL_FLOAT_MAT2                     0x8B5A
#define GL_FLOAT_MAT3                     0x8B5B
#define GL_FLOAT_MAT4                     0x8B5C
#define GL_SAMPLER_1D                     0x8B5D
#define GL_SAMPLER_2D                     0x8B5E
#define GL_SAMPLER_3D                     0x8B5F
#define GL_SAMPLER_CUBE                   0x8B60
#define GL_SAMPLER_1D_SHADOW              0x8B61
#define GL_SAMPLER_2D_SHADOW              0x8B62
#define GL_DELETE_STATUS                  0x8B80
#define GL_COMPILE_STATUS                 0x8B81
#define GL_LINK_STATUS                    0x8B82
#define GL_VALIDATE_STATUS                0x8B83
#define GL_INFO_LOG_LENGTH                0x8B84
#define GL_ATTACHED_SHADERS               0x8B85
#define GL_ACTIVE_UNIFORMS                0x8B86
#define GL_ACTIVE_UNIFORM_MAX_LENGTH      0x8B87
#define GL_SHADER_SOURCE_LENGTH           0x8B88
#define GL_ACTIVE_ATTRIBUTES              0x8B89
#define GL_ACTIVE_ATTRIBUTE_MAX_LENGTH    0x8B8A
#define GL_FRAGMENT_SHADER_DERIVATIVE_HINT 0x8B8B
#define GL_SHADING_LANGUAGE_VERSION       0x8B8C
#define GL_CURRENT_PROGRAM                0x8B8D
#define GL_POINT_SPRITE_COORD_ORIGIN      0x8CA0
#define GL_LOWER_LEFT                     0x8CA1
#define GL_UPPER_LEFT                     0x8CA2
#define GL_STENCIL_BACK_REF               0x8CA3
#define GL_STENCIL_BACK_VALUE_MASK        0x8CA4
#define GL_STENCIL_BACK_WRITEMASK         0x8CA5

typedef char GLchar;

#ifdef	DLL_TARGET_DCC

#ifdef	__cplusplus
extern "C" {
#endif /* __cplusplus */

/*
 * Prototypes
 */
VGLEXTAPI void VGLEXTAPIENTRY glBlendEquationSeparate (GLenum, GLenum);
VGLEXTAPI void VGLEXTAPIENTRY glDrawBuffers (GLsizei, const GLenum *);
VGLEXTAPI void VGLEXTAPIENTRY glStencilOpSeparate (GLenum, GLenum, GLenum, GLenum);
VGLEXTAPI void VGLEXTAPIENTRY glStencilFuncSeparate (GLenum, GLenum, GLint, GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glStencilMaskSeparate (GLenum, GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glAttachShader (GLuint, GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glBindAttribLocation (GLuint, GLuint, const GLchar *);
VGLEXTAPI void VGLEXTAPIENTRY glCompileShader (GLuint);
VGLEXTAPI GLuint VGLEXTAPIENTRY glCreateProgram (void);
VGLEXTAPI GLuint VGLEXTAPIENTRY glCreateShader (GLenum);
VGLEXTAPI void VGLEXTAPIENTRY glDeleteProgram (GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glDeleteShader (GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glDetachShader (GLuint, GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glDisableVertexAttribArray (GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glEnableVertexAttribArray (GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glGetActiveAttrib (GLuint, GLuint, GLsizei, GLsizei *, GLint *, GLenum *, GLchar *);
VGLEXTAPI void VGLEXTAPIENTRY glGetActiveUniform (GLuint, GLuint, GLsizei, GLsizei *, GLint *, GLenum *, GLchar *);
VGLEXTAPI void VGLEXTAPIENTRY glGetAttachedShaders (GLuint, GLsizei, GLsizei *, GLuint *);
VGLEXTAPI GLint VGLEXTAPIENTRY glGetAttribLocation (GLuint, const GLchar *);
VGLEXTAPI void VGLEXTAPIENTRY glGetProgramiv (GLuint, GLenum, GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glGetProgramInfoLog (GLuint, GLsizei, GLsizei *, GLchar *);
VGLEXTAPI void VGLEXTAPIENTRY glGetShaderiv (GLuint, GLenum, GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glGetShaderInfoLog (GLuint, GLsizei, GLsizei *, GLchar *);
VGLEXTAPI void VGLEXTAPIENTRY glGetShaderSource (GLuint, GLsizei, GLsizei *, GLchar *);
VGLEXTAPI GLint VGLEXTAPIENTRY glGetUniformLocation (GLuint, const GLchar *);
VGLEXTAPI void VGLEXTAPIENTRY glGetUniformfv (GLuint, GLint, GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glGetUniformiv (GLuint, GLint, GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glGetVertexAttribdv (GLuint, GLenum, GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glGetVertexAttribfv (GLuint, GLenum, GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glGetVertexAttribiv (GLuint, GLenum, GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glGetVertexAttribPointerv (GLuint, GLenum, GLvoid* *);
VGLEXTAPI GLboolean VGLEXTAPIENTRY glIsProgram (GLuint);
VGLEXTAPI GLboolean VGLEXTAPIENTRY glIsShader (GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glLinkProgram (GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glShaderSource (GLuint, GLsizei, const GLchar* *, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glUseProgram (GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glUniform1f (GLint, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glUniform2f (GLint, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glUniform3f (GLint, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glUniform4f (GLint, GLfloat, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glUniform1i (GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glUniform2i (GLint, GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glUniform3i (GLint, GLint, GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glUniform4i (GLint, GLint, GLint, GLint, GLint);
VGLEXTAPI void VGLEXTAPIENTRY glUniform1fv (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glUniform2fv (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glUniform3fv (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glUniform4fv (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glUniform1iv (GLint, GLsizei, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glUniform2iv (GLint, GLsizei, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glUniform3iv (GLint, GLsizei, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glUniform4iv (GLint, GLsizei, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glUniformMatrix2fv (GLint, GLsizei, GLboolean, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glUniformMatrix3fv (GLint, GLsizei, GLboolean, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glUniformMatrix4fv (GLint, GLsizei, GLboolean, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glValidateProgram (GLuint);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib1d (GLuint, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib1dv (GLuint, const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib1f (GLuint, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib1fv (GLuint, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib1s (GLuint, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib1sv (GLuint, const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib2d (GLuint, GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib2dv (GLuint, const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib2f (GLuint, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib2fv (GLuint, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib2s (GLuint, GLshort, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib2sv (GLuint, const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib3d (GLuint, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib3dv (GLuint, const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib3f (GLuint, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib3fv (GLuint, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib3s (GLuint, GLshort, GLshort, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib3sv (GLuint, const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4Nbv (GLuint, const GLbyte *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4Niv (GLuint, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4Nsv (GLuint, const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4Nub (GLuint, GLubyte, GLubyte, GLubyte, GLubyte);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4Nubv (GLuint, const GLubyte *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4Nuiv (GLuint, const GLuint *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4Nusv (GLuint, const GLushort *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4bv (GLuint, const GLbyte *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4d (GLuint, GLdouble, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4dv (GLuint, const GLdouble *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4f (GLuint, GLfloat, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4fv (GLuint, const GLfloat *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4iv (GLuint, const GLint *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4s (GLuint, GLshort, GLshort, GLshort, GLshort);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4sv (GLuint, const GLshort *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4ubv (GLuint, const GLubyte *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4uiv (GLuint, const GLuint *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttrib4usv (GLuint, const GLushort *);
VGLEXTAPI void VGLEXTAPIENTRY glVertexAttribPointer (GLuint, GLint, GLenum, GLboolean, GLsizei, const GLvoid *);

#ifdef	__cplusplus
}
#endif /* __cplusplus */

#endif	/* !DLL_TARGET_DCC */

/*
 * Pointers
 */
VGLEXTAPI void (VGLEXTAPIENTRY *glBlendEquationSeparate) (GLenum, GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glDrawBuffers) (GLsizei, const GLenum *);
VGLEXTAPI void (VGLEXTAPIENTRY *glStencilOpSeparate) (GLenum, GLenum, GLenum, GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glStencilFuncSeparate) (GLenum, GLenum, GLint, GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glStencilMaskSeparate) (GLenum, GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glAttachShader) (GLuint, GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glBindAttribLocation) (GLuint, GLuint, const GLchar *);
VGLEXTAPI void (VGLEXTAPIENTRY *glCompileShader) (GLuint);
VGLEXTAPI GLuint (VGLEXTAPIENTRY *glCreateProgram) (void);
VGLEXTAPI GLuint (VGLEXTAPIENTRY *glCreateShader) (GLenum);
VGLEXTAPI void (VGLEXTAPIENTRY *glDeleteProgram) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glDeleteShader) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glDetachShader) (GLuint, GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glDisableVertexAttribArray) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glEnableVertexAttribArray) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetActiveAttrib) (GLuint, GLuint, GLsizei, GLsizei *, GLint *, GLenum *, GLchar *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetActiveUniform) (GLuint, GLuint, GLsizei, GLsizei *, GLint *, GLenum *, GLchar *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetAttachedShaders) (GLuint, GLsizei, GLsizei *, GLuint *);
VGLEXTAPI GLint (VGLEXTAPIENTRY *glGetAttribLocation) (GLuint, const GLchar *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetProgramiv) (GLuint, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetProgramInfoLog) (GLuint, GLsizei, GLsizei *, GLchar *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetShaderiv) (GLuint, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetShaderInfoLog) (GLuint, GLsizei, GLsizei *, GLchar *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetShaderSource) (GLuint, GLsizei, GLsizei *, GLchar *);
VGLEXTAPI GLint (VGLEXTAPIENTRY *glGetUniformLocation) (GLuint, const GLchar *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetUniformfv) (GLuint, GLint, GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetUniformiv) (GLuint, GLint, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetVertexAttribdv) (GLuint, GLenum, GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetVertexAttribfv) (GLuint, GLenum, GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetVertexAttribiv) (GLuint, GLenum, GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glGetVertexAttribPointerv) (GLuint, GLenum, GLvoid* *);
VGLEXTAPI GLboolean (VGLEXTAPIENTRY *glIsProgram) (GLuint);
VGLEXTAPI GLboolean (VGLEXTAPIENTRY *glIsShader) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glLinkProgram) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glShaderSource) (GLuint, GLsizei, const GLchar* *, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUseProgram) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform1f) (GLint, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform2f) (GLint, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform3f) (GLint, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform4f) (GLint, GLfloat, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform1i) (GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform2i) (GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform3i) (GLint, GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform4i) (GLint, GLint, GLint, GLint, GLint);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform1fv) (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform2fv) (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform3fv) (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform4fv) (GLint, GLsizei, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform1iv) (GLint, GLsizei, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform2iv) (GLint, GLsizei, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform3iv) (GLint, GLsizei, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniform4iv) (GLint, GLsizei, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniformMatrix2fv) (GLint, GLsizei, GLboolean, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniformMatrix3fv) (GLint, GLsizei, GLboolean, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glUniformMatrix4fv) (GLint, GLsizei, GLboolean, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glValidateProgram) (GLuint);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib1d) (GLuint, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib1dv) (GLuint, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib1f) (GLuint, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib1fv) (GLuint, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib1s) (GLuint, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib1sv) (GLuint, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib2d) (GLuint, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib2dv) (GLuint, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib2f) (GLuint, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib2fv) (GLuint, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib2s) (GLuint, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib2sv) (GLuint, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib3d) (GLuint, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib3dv) (GLuint, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib3f) (GLuint, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib3fv) (GLuint, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib3s) (GLuint, GLshort, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib3sv) (GLuint, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4Nbv) (GLuint, const GLbyte *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4Niv) (GLuint, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4Nsv) (GLuint, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4Nub) (GLuint, GLubyte, GLubyte, GLubyte, GLubyte);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4Nubv) (GLuint, const GLubyte *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4Nuiv) (GLuint, const GLuint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4Nusv) (GLuint, const GLushort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4bv) (GLuint, const GLbyte *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4d) (GLuint, GLdouble, GLdouble, GLdouble, GLdouble);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4dv) (GLuint, const GLdouble *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4f) (GLuint, GLfloat, GLfloat, GLfloat, GLfloat);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4fv) (GLuint, const GLfloat *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4iv) (GLuint, const GLint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4s) (GLuint, GLshort, GLshort, GLshort, GLshort);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4sv) (GLuint, const GLshort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4ubv) (GLuint, const GLubyte *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4uiv) (GLuint, const GLuint *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttrib4usv) (GLuint, const GLushort *);
VGLEXTAPI void (VGLEXTAPIENTRY *glVertexAttribPointer) (GLuint, GLint, GLenum, GLboolean, GLsizei, const GLvoid *);

#endif	/* __Vgl_h_ */
